// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//
// Copyright (c) 2000 University of Mannheim. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for educational, research and non-profit purposes,
// without fee, and without a written agreement is hereby granted,
// provided that the above copyright notice, this paragraph and the
// following four paragraphs appear in all copies.
//
// Permission to incorporate this software into commercial products may
// be obtained from the following address:
// University of Mannheim
// Lehrstuhl fuer Praktische Informatik IV
// L15, 16
// 68161 Mannheim, Germany
//
// IN NO EVENT SHALL THE UNIVERSITY OF MANNHEIM BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
// ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF MANNHEIM HAS BEEN ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//
// THE UNIVERSITY OF MANNHEIM SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER
// IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF MANNHEIM HAS NO OBLIGATIONS
// TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The University of Mannheim will appreciate receiving any enhancements and
// modifications of this software made by others, and reserves the right to
// include them in future releases. The University also recommends that this
// software not be redistributed by third parties, but be obtained by all
// directly from the University's server, so that all users of the software
// may receive bug notifications and other relevant announcements.
//
// Authors:
//   Volker Hilt, Juergen Vogel, Martin Mauve
//   University of Mannheim
//   {hilt,vogel,mauve}@informatik.uni-mannheim.de
//
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/**********************************************************************

 program : rtcpiPacket.cpp
 authors : Juergen Vogel
 date    : 05.04.2000
 desc    : defines RCTP/I Packets and classes to handle them

 **********************************************************************/

#ifdef WIN32
#include <winsock.h>
#include <util/win32.h>
#else
#include <arpa/inet.h>
#endif

#include <iostream>
#include <string>

#include <rtpi/rtcpiPacket.h>


int RTPI::RtcpiCompoundPacket::max_packet_size = RTPI::RtcpiCompoundPacket::MAX_PKT_SIZE;

//
// Functions of class RtcpiPacket
//
int RTPI::RtcpiPacket::ParseType(char* pos) {
  uint8* p8 = (uint8*)pos;
  int type = (int)(p8[1] & RTCPI_MASK_TYPE);

  // RTCP/I header validy check
  if (((p8[0] & RTCPI_MASK_V) >> 6) != RtcpiPacket::VERSION)
    throw RtpiException(RtpiException::ERR_RTCPI_WRONG_VERSION, "RtcpiPacket::ParseType - wrong Version of RTCP/I Header");

  if ((type != PKT_SUBREP) && (type != PKT_SDES) && (type != PKT_BYE) && (type != PKT_APP))
    throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RtcpiPacket::RtcpiPacket - wrong type");

  return type;
}

int RTPI::RtcpiPacket::ParseLength(char* pos) {
  uint8*  p8  = (uint8*)pos;
  uint16* p16 = (uint16*)pos;

  // RTCP/I header validy check
  if (((p8[0] & RTCPI_MASK_V) >> 6) != RtcpiPacket::VERSION)
    throw RtpiException(RtpiException::ERR_RTCPI_WRONG_VERSION, "RtcpiPacket::ParseLength - wrong Version of RTCP/I Header");

  return (int)(ntohs(p16[1]));
}



//
// Functions of class RtcpiSubcomponentReportItem
//
RTPI::RtcpiSubcomponentReportItem::RtcpiSubcomponentReportItem(uint32 _subid, bool _active, const char* _name, int _length)
  : subid(_subid), length(_length), name((char*)_name), active(_active) {
  if (length > 255)
    throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RtcpiSubcomponentReportItem::RtcpiSubcomponentReportItem - name too long");
  if (length != 0) {
    name = new char[length + 1];
    memcpy(name, _name, length);
    name[length] = '\0';
  }
}


void RTPI::RtcpiSubcomponentReportItem::Print() {
  if (length != 0) {
    std::cout << "ITEM:(SUBID=" << (int)subid << ",LENGTH=" << (int)length << ",NAME=" ;
    for (int i = 0; i < length; i++)
      std::cout << ((name[i] > ' ' && name[i] < '~') ? name[i] : '*');
    std::cout << ",ACTIVE=" << (int)active << ")" << std::endl;
  } else
    std::cout << "ITEM:(SUBID=" << (int)subid << ",ACTIVE=" << (int)active << ")" << std::endl;
}


int RTPI::RtcpiSubcomponentReportItem::GetSize() {
  if (length == 0) {
    return 8;
  } else {
    int padd = ((length+2) % 4 == 0) ? 0 : 4 - ((length+2) % 4);
    return 6 + length + padd;
  }
}


void RTPI::RtcpiSubcomponentReportItem::SetApplicationName(const char* _name, int _length) {
  if (length > 255)
    throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RtcpiSubcomponentReportItem::SetApplicationName - name too long");
  if (name != NULL) delete[] name;
  length = _length;
  name = new char[length + 1];
  memcpy(name, _name, length);
  name[length] = '\0';
}


int RTPI::RtcpiSubcomponentReportItem::Parse(char* pos) {
  uint32* p32 = (uint32*)pos;
  uint8*  p8  = (uint8*)pos;

  subid       = ntohl(p32[0]);

  active = (p8[4] & RTCPI_MASK_A) >> 7;

  length = p8[5];

  if (length != 0) {
    if (name != NULL) delete[] name;
    name = new char[length + 1];
    memcpy(name, &pos[6], length);
    name[length] = '\0';
    int pad = ((length+2) % 4 == 0) ? 0 : 4 - ((length+2) % 4);
    return 6 + length + pad;  // ignore padding bits
  } else {
    if (name != NULL) delete[] name;
    name = NULL;
    return 8;
  }
}


int RTPI::RtcpiSubcomponentReportItem::Flush(char* pos) {
  uint32* p32 = (uint32*)pos;
  uint8* p8  = (uint8*)pos;

  p32[0] = htonl(subid);

  p32[1] = 0;
  if (active)
    p8[4] = p8[4] | (1 << 7);

  if ((length != 0) && (name != NULL)) {
    p8[5] = length;
    memcpy(pos + 6, name, length);

    // padding ?
    int padd = ((length+2) % 4 == 0) ? 0 : 4 - ((length+2) % 4);
    for (int i = 0; i < padd; i++)
      pos[6 + length + i] = 0;

    return 6 + length + padd;
  } else {
    // already set to 0
    return 8;
  }
}


//
// Functions of class RtcpiSubcomponentReportPacket
//
RTPI::RtcpiSubcomponentReportPacket::~RtcpiSubcomponentReportPacket() {
  for (std::vector<RtcpiSubcomponentReportItem*>::iterator i = items.begin(); i != items.end(); ++i)
    delete *i;
}


void RTPI::RtcpiSubcomponentReportPacket::Print() {
  std::cout << "SUBREP:(V=" << (int)v << ",SC=" << (int)sc << ",P=" << (int)p << ",TYPE=" << (int)type << ",LENGTH=" << (int)length
       << ",PID=" << (int)pid << ")" << std::endl;
  for (std::vector<RtcpiSubcomponentReportItem*>::iterator i = items.begin(); i != items.end(); ++i)
    (*i)->Print();
}


void RTPI::RtcpiSubcomponentReportPacket::AddItem(RtcpiSubcomponentReportItem* item, int max_size) {
  if ((items.size() == 32) ||
      ((max_size > 0) && ((GetSize() + item->GetSize()) > max_size)))
    throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RtcpiSubcomponentReportPacket::AddItem - maximum packet size reached");

  items.push_back(item);
}


void RTPI::RtcpiSubcomponentReportPacket::AddItem(uint32 subid, bool active, const char* name, int length, int max_size) {

  RtcpiSubcomponentReportItem* item = new RtcpiSubcomponentReportItem(subid,active,name,length);

  if ((items.size() == 32) ||
      ((max_size > 0) && (GetSize() + item->GetSize()) > max_size)) {
    delete item;
    throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RtcpiSubcomponentReportPacket::AddItem - maximum packet size reached");
  }

  items.push_back(item);
}


int RTPI::RtcpiSubcomponentReportPacket::GetSize() {
  int size = 8;
  for (std::vector<RtcpiSubcomponentReportItem*>::iterator i = items.begin(); i != items.end(); ++i)
    size += (*i)->GetSize();
  return size;
}


int RTPI::RtcpiSubcomponentReportPacket::Parse(char* pos) {
  uint8*  p8  = (uint8*)pos;
  uint16* p16 = (uint16*)pos;
  uint32* p32 = (uint32*)pos;

  v  = (p8[0] & RTCPI_MASK_V) >> 6;
  sc = (p8[0] & RTCPI_MASK_SC) + 1;

  p    = (p8[1] & RTCPI_MASK_P) >> 7;
  type = p8[1] & RTCPI_MASK_TYPE;

  if (v != RtcpiPacket::VERSION)
    throw RtpiException(RtpiException::ERR_RTCPI_WRONG_VERSION, "RTCPISubcomponentReportPacket::Parse - wrong version of RTCP/I Header");

  if (type != RtcpiPacket::PKT_SUBREP)
    throw RtpiException(RtpiException::ERR_RTCPI_PARSE, "RtcpiSubcomponentReportPacket::Parse - wrong packet type");

  if (sc < 1)
    throw RtpiException(RtpiException::ERR_RTCPI_PARSE, "RtcpiSubcomponentReportPacket::Parse - packet contains no items");

  length = ntohs(p16[1]);

  // we allow empty reports if requested and padding is zero!
  if ((mode == EMPTY_ALLOWED) && (p == 0) && (sc == 1) && (length == 0))
    sc = 0;

  pid = ntohl(p32[1]);

  char* ptr = pos + 8;
  int bytesParsed;
  RtcpiSubcomponentReportItem* item;
  for (int i = 0; i < sc; i++) {
    item = new RtcpiSubcomponentReportItem();
    items.push_back(item);
    bytesParsed = item->Parse(ptr);
    ptr += bytesParsed;
  }

  // ignore padding

  return 8 + length*4;
}


int RTPI::RtcpiSubcomponentReportPacket::Flush(char* pos, int space, int globalPadd) {
  if (space < GetSize() + globalPadd)
    throw RtpiException(RtpiException::ERR_RTCPI_FLUSH, "RTCPISubcomponentReportPacket::Flush - buffer size not big enough");

  uint8*  p8  = (uint8*)pos;
  uint16* p16 = (uint16*)pos;
  uint32* p32 = (uint32*)pos;

  sc = items.size();

  // we allow empty packets if requested (padding MUST be zero!)
  if ((sc < 1) && ((globalPadd != 0) || (mode == EMPTY_NOT_ALLOWED)))
    throw RtpiException(RtpiException::ERR_RTCPI_FLUSH, "RtcpiSubcomponentReportPacket::Flush - packet must contain at least one item");

  p8[0] = 0;
  p8[0] = p8[0] | (v << 6);
  if ((mode == EMPTY_ALLOWED) && (sc == 0))
    p8[0] = p8[0] | 0;
  else
    p8[0] = p8[0] | (sc - 1);

  p32[1] = htonl(pid);

  length = 0;
  char* ptr = pos + 8;
  int bytesFlushed;

  int j = 0;
  for (std::vector<RtcpiSubcomponentReportItem*>::iterator i = items.begin(); i != items.end(); ++i, ++j) {
    bytesFlushed = (*i)->Flush(ptr);
    length += bytesFlushed;
    ptr += bytesFlushed;
  }

  // global padding requested for this packet ?
  if (globalPadd != 0) {
    for (int i = 0; i < globalPadd - 1; ++i)
      p8[8 + length + i] = 0;

    length += globalPadd;
    p8[8 + length - 1] = globalPadd;

    p = 1;
  }

  p8[1] = 0;
  p8[1] = p8[1] | (p << 7);
  p8[1] = p8[1] | type;

  length /= 4;  // length gives packet size in 32 bit words
  p16[1] = htons(length);

  return 8 + length*4;
}




//
// Functions of class RtcpiSourceDescriptionItem
//
RTPI::RtcpiSourceDescriptionItem::RtcpiSourceDescriptionItem(int _type, const char* _data, unsigned short _length) : type(_type), length(_length) {
  if (length > 255)
    throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RtcpiSourceDescriptionItem::RtcpiSourceDescritpionItem - data too long");
  data = new char[length + 1];
  memcpy(data, _data, length);
  data[length]= '\0';
}


void RTPI::RtcpiSourceDescriptionItem::SetData(const char* _data, int _length) {
  if (length > 255)
    throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RtcpiSourceDescriptionItem::SetData - data too long");
  length = _length;
  if (data != NULL) delete[] data;
  data = new char[length + 1];
  memcpy(data, _data, length);
  data[length] = '\0';
}


void RTPI::RtcpiSourceDescriptionItem::Print() {
  if (length != 0)
    std::cout << "ITEM:(TYPE=" << (int)type << ",LENGTH=" << (int)length << ",DATA=" << data << ")" << std::endl;
  else
    std::cout << "ITEM:(TYPE=" << (int)type << ")" << std::endl;
}


int RTPI::RtcpiSourceDescriptionItem::ParseType(char* pos) {
  return ((uint8*)pos)[0];
}


int RTPI::RtcpiSourceDescriptionItem::Parse(char* pos) {
  uint8*  p8  = (uint8*)pos;

  type   = p8[0];
  length = p8[1];

  if (data != NULL) delete[] data;
  data = new char[length+1];
  memcpy(data, &p8[2], length);
  data[length] = '\0';

  return (2 + length);
}


int RTPI::RtcpiSourceDescriptionItem::Flush(char* pos) {
  uint8*  p8  = (uint8*)pos;

  p8[0] = type;
  p8[1] = length;

  memcpy(pos + 2, data, length);

  return (2 + length);
}



//
// Functions of class RtcpiSourceDescriptionPacket
//
RTPI::RtcpiSourceDescriptionPacket::~RtcpiSourceDescriptionPacket() {
  for (std::vector<RtcpiSourceDescriptionItem*>::iterator i = items.begin(); i != items.end(); ++i)
    delete *i;
}


void RTPI::RtcpiSourceDescriptionPacket::Print() {
  std::cout << "SDES:(V=" << (int)v << ",IC=" << (int)ic << ",P=" << (int)p << ",TYPE=" << (int)type << ",LENGTH=" << (int)length
       << ",PID=" << (int)pid << ")" << std::endl;
  for (std::vector<RtcpiSourceDescriptionItem*>::iterator i = items.begin(); i != items.end(); ++i)
    (*i)->Print();
}


void RTPI::RtcpiSourceDescriptionPacket::AddItem(RtcpiSourceDescriptionItem* item, int max_size) {
  if ((items.size() == 32) ||
      ((max_size > 0) && ((GetSize() + item->GetSize()) > max_size)))
    throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RtcpiSourceDescriptionPacket::AddItem - maximum packet size reached");

  items.push_back(item);
}


void RTPI::RtcpiSourceDescriptionPacket::AddItem(int type, const char* data, int length, int max_size) {

  RtcpiSourceDescriptionItem* item = new RtcpiSourceDescriptionItem(type,data,length);

  if ((items.size() == 32) ||
      ((max_size > 0) && ((GetSize() + item->GetSize()) > max_size))) {
    delete item;
    throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RtcpiSourceDescriptionPacket::AddItem - maximum packet size reached");
  }

  items.push_back(item);
}


int RTPI::RtcpiSourceDescriptionPacket::GetSize() {
  int size = 8;
  for (std::vector<RtcpiSourceDescriptionItem*>::iterator i = items.begin(); i != items.end(); ++i)
    size += (*i)->GetSize();
  int padd = (size % 4 == 0) ? 0 : 4 - (size % 4);
  return size + padd;
}


int RTPI::RtcpiSourceDescriptionPacket::Parse(char* pos) {
  uint8*  p8  = (uint8*)pos;
  uint16* p16 = (uint16*)pos;
  uint32* p32 = (uint32*)pos;

  v    = (p8[0] & RTCPI_MASK_V) >> 6;
  ic   = (p8[0] & RTCPI_MASK_IC) + 1;
  p    = (p8[1] & RTCPI_MASK_P) >> 7;
  type = p8[1] & RTCPI_MASK_TYPE;

  if (v != RtcpiPacket::VERSION)
    throw RtpiException(RtpiException::ERR_RTCPI_WRONG_VERSION, "RtcpiSourceDescriptionPacketPacket::Parse - wrong version of RTCP/I Header");

  if (type != RtcpiPacket::PKT_SDES)
    throw RtpiException(RtpiException::ERR_RTCPI_PARSE, "RtcpiSourceDescriptionPacket::Parse - wrong packet type");

  if (ic < 1)
    throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RtcpiSourceDescritpionPacket::Parse - packet must contain at least one CNAME item");

  length = ntohs(p16[1]);

  pid = ntohl(p32[1]);

  char* ptr = pos + 8;
  int bytesParsed;

  if (RtcpiSourceDescriptionItem::ParseType(ptr) != RtcpiSourceDescriptionItem::CNAME)
    throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RtcpiSourceDescriptionPacket::Parse - first SDES item must be CNAME");

  RtcpiSourceDescriptionItem* item;
  for (int i = 1; i <= ic; i++) {
    item = new RtcpiSourceDescriptionItem();
    bytesParsed = item->Parse(ptr);
    items.push_back(item);
    ptr += bytesParsed;
  }

  // ingnore padding

  return 8 + length*4;
}


int RTPI::RtcpiSourceDescriptionPacket::Flush(char* pos, int space, int globalPadd) {
  if (space < GetSize() + globalPadd)
    throw RtpiException(RtpiException::ERR_RTCPI_FLUSH, "RTCPISourceDescriptionPacket::Flush - buffer size not big enough");

  uint8*  p8  = (uint8*)pos;
  uint16* p16 = (uint16*)pos;
  uint32* p32 = (uint32*)pos;

  ic = items.size();

  if (ic < 1)
    throw RtpiException(RtpiException::ERR_RTCPI_FLUSH, "RtcpiSourceDescriptionPacket::Flush - packet must contain at least one CNAME item");

  if ((items.front())->GetType() != RtcpiSourceDescriptionItem::CNAME)
    throw RtpiException(RtpiException::ERR_RTCPI_FLUSH, "RtcpiSourceDescriptionPacket::Flush - first item must be CNAME");

  p8[0] = 0;
  p8[0] = p8[0] | (v << 6);
  p8[0] = p8[0] | (ic - 1);

  p32[1] = htonl(pid);

  length = 0;
  char* ptr = pos + 8;
  int bytesFlushed = 0;
  for (std::vector<RtcpiSourceDescriptionItem*>::iterator i = items.begin(); i != items.end(); ++i) {
    bytesFlushed = (*i)->Flush(ptr);
    length += bytesFlushed;
    ptr += bytesFlushed;
  }

  // local padding necessary?
  int padd = (length % 4 == 0) ? 0 : (4 - (length % 4));
  if (padd != 0) {
    for (int i = 0; i < padd; i++)
      pos[8 + length + i] = 0;
    length += padd;
  }

  // global padding requested for this packet ?
  if (globalPadd != 0) {
    for (int i = 0; i < globalPadd - 1; ++i)
      p8[8 + length + i] = 0;

    length += globalPadd;
    p8[8 + length - 1] = globalPadd;

    p = 1;
  }

  p8[1] = 0;
  p8[1] = p8[1] | (p << 7);
  p8[1] = p8[1] | type;

  length /= 4;  // length gives packet size in 32 bit words
  p16[1] = htons(length);

  return 8 + length*4;
}



//
// Functions of class RtcpiByePacket
//
RTPI::RtcpiByePacket::RtcpiByePacket(uint32 _pid, const char* _reason, int _rlength)
  : RtcpiPacket(RtcpiPacket::PKT_BYE, _pid) {
  if (_rlength > 255)
    throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RtcpiByePacket::RtcpiByePacket - reason too long");
  rlength = _rlength;
  reason = new char[rlength + 1];
  memcpy(reason, _reason, rlength);
  reason[rlength] = '\0';
}

void RTPI::RtcpiByePacket::Print() {
  if (rlength != 0)
    std::cout << "BYE:(V=" << (int)v << ",P=" << (int)p << ",TYPE=" << (int)type << ",LENGTH=" << (int)length
	 << ",PID=" << (int)pid << ",RLENGTH=" << (int)rlength << ",REASON=" << reason << ")" << std::endl;
  else
    std::cout << "BYE:(V=" << (int)v << ",P=" <<(int) p << ",TYPE=" << (int)type << ",LENGTH=" << (int)length
	 << ",PID=" << (int)pid << ")" << std::endl;
}


void RTPI::RtcpiByePacket::SetReason(const char* _reason, int _rlength) {
  if (_rlength > 255)
    throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RtcpiByePacket::RtcpiByePacket - reason too long");

  if (reason != NULL) delete[] reason;

  rlength =  _rlength;
  reason = new char[rlength];
  memcpy(reason, _reason, rlength);
  reason[rlength] = '\0';
}


int RTPI::RtcpiByePacket::GetSize() {
  int padd = ((rlength+1) % 4 == 0) ? 0 : 4 - ((rlength+1) % 4);
  return (rlength == 0) ? 8 : 9 + rlength + padd;
}


int RTPI::RtcpiByePacket::Parse(char* pos) {
  uint8*  p8  = (uint8*)pos;
  uint16* p16 = (uint16*)pos;
  uint32* p32 = (uint32*)pos;

  v    = (p8[0] & RTCPI_MASK_V) >> 6;
  p    = (p8[1] & RTCPI_MASK_P) >> 7;
  type = p8[1] & RTCPI_MASK_TYPE;

  if (v != RtcpiPacket::VERSION)
    throw RtpiException(RtpiException::ERR_RTCPI_WRONG_VERSION, "RtcpiByePacket::Parse - wrong version of RTCP/I Header");

  if (type != RtcpiPacket::PKT_BYE)
    throw RtpiException(RtpiException::ERR_RTCPI_PARSE, "RtcpiByePacket::Parse - wrong packet type");

  length = ntohs(p16[1]);
  int lengthInBytes = 4*length;

  pid = ntohl(p32[1]);

  int globalPadd = 0;
  if (p) {
    // global padding attached
    globalPadd = p8[8 + lengthInBytes - 1];
  }

  if (lengthInBytes - globalPadd > 0) {
    // reason included
    rlength = p8[8];

    if (reason != NULL) delete[] reason;
    reason   = new char[rlength + 1];
    memcpy(reason, pos + 9, rlength);
    reason[rlength] = '\0';
  }

  return (lengthInBytes + 8);
}


int RTPI::RtcpiByePacket::Flush(char* pos, int space, int globalPadd) {
  if (space < GetSize() + globalPadd)
    throw RtpiException(RtpiException::ERR_RTCPI_FLUSH, "RTCPIByePacket::Flush - buffer size not big enough");

  uint8*  p8  = (uint8*)pos;
  uint16* p16 = (uint16*)pos;
  uint32* p32 = (uint32*)pos;

  p8[0] = 0;
  p8[0] = p8[0] | (v << 6);

  p32[1] = htonl(pid);

  length = 0;
  if (rlength > 0) {
    p8[8] = rlength;

    memcpy(pos + 9, reason, rlength);
    length = 1 + rlength;

    // local padding
    int padd = ((rlength+1) % 4 == 0) ? 0 : 4 - ((rlength+1) % 4);
    if (padd != 0) {
      for (int i = 0; i < padd; i++)
	pos[9 + rlength + i] = 0;
      length += padd;
    }
  }

  // global padding requested for this packet ?
  if (globalPadd != 0) {
    for (int i = 0; i < globalPadd - 1; ++i)
      p8[8 + length + i] = 0;

    length += globalPadd;
    p8[8 + length - 1] = globalPadd;

    p = 1;
  }

  p8[1] = 0;
  p8[1] = p8[1] | (p << 7);
  p8[1] = p8[1] | type;

  length /= 4;  // length gives packet size in 32 bit words
  p16[1] = htons(length);

  return 8 + length*4;
}



//
// Functions of class RtcpiApplicationPacket
//
RTPI::RtcpiApplicationPacket::RtcpiApplicationPacket(uint32 _pid, const char* _name, int _subtype, const char* _data, int _dlength)
  : RtcpiPacket(RtcpiPacket::PKT_APP, _pid, 1), r(0), subtype(_subtype), dlength(_dlength) {
  if (dlength == 0) {
    data = NULL;
    return;
  }

  if ((dlength % 4) != 0)
    throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT,
			"RtcpiApplicationPacket::RtcpiApplicationPacket - data length is not a multiple of 32 bit");

  data = new char[dlength];
  memcpy(data, _data, dlength);
  memcpy(name, _name, 5);
}


void RTPI::RtcpiApplicationPacket::Print() {
  if (dlength != 0)
    std::cout << "APP:(V=" << (int)v << ",SUBTYPE=" << (int)subtype  << ",P=" << (int)p << ",TYPE=" << (int)type << ",LENGTH=" << (int)length
	 << ",PID=" <<(int) pid << ",NAME=" << name << ",DLENGTH=" << (int)dlength << ",DATA=" << data << ")" << std::endl;
  else
    std::cout << "APP:(V=" << (int)v << ",SUBTYPE=" << (int)subtype  << ",P=" << (int)p << ",TYPE=" << (int)type << ",LENGTH=" << (int)length
	 << ",PID=" << (int)pid << ",NAME=" << name << ")" << std::endl;
}


void RTPI::RtcpiApplicationPacket::SetData(const char* _data, int _dlength) {
  if ((dlength % 4) != 0)
    throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT,
			"RtcpiApplicationPacket::SetData - data length is not a multiple of 32 bit");

  if (data != NULL) delete[] data;
  dlength = _dlength;
  data = new char[dlength];
  memcpy(data, _data, dlength);
}


int RTPI::RtcpiApplicationPacket::GetSize() {
  return (dlength == 0) ? 12 : 12 + dlength;
}


int RTPI::RtcpiApplicationPacket::Parse(char* pos) {
  uint8*  p8  = (uint8*)pos;
  uint16* p16 = (uint16*)pos;
  uint32* p32 = (uint32*)pos;

  v       = (p8[0] & RTCPI_MASK_V) >> 6;
  subtype = p8[0] & RTCPI_MASK_SUBTYPE;
  p       = (p8[1] & RTCPI_MASK_P) >> 7;
  type    = p8[1] & RTCPI_MASK_TYPE;

  if (v != RtcpiPacket::VERSION)
    throw RtpiException(RtpiException::ERR_RTCPI_WRONG_VERSION, "RtcpiApplicationPacket::Parse - wrong version of RTCP/I Header");

  if (type != RtcpiPacket::PKT_APP)
    throw RtpiException(RtpiException::ERR_RTCPI_PARSE, "RtcpiApplicationPacket::Parse - wrong packet type");

  length = ntohs(p16[1]);
  int lengthInBytes = 4*length;

  pid = ntohl(p32[1]);

  memcpy(name, pos + 8, 4);
  name[4] = '\0';

  int globalPadd = 0;
  if (p) {
    // global padding attached
    globalPadd = p8[8 + lengthInBytes - 1];
  }

  dlength = 0;
  if (lengthInBytes - 4 - globalPadd > 0) { // name is always there
    // data
    dlength = lengthInBytes - 4 - globalPadd;

    if (data != NULL) delete[] data;
    data   = new char[dlength + 1];
    memcpy(data, pos + 12, dlength);
    data[dlength] = '\0';
  }

  return 8 + length*4;
}


int RTPI::RtcpiApplicationPacket::Flush(char* pos, int space, int globalPadd) {
  if (space < GetSize() + globalPadd)
    throw RtpiException(RtpiException::ERR_RTCPI_FLUSH, "RTCPIApplicationPacket::Flush - buffer size not big enough");

  uint8*  p8  = (uint8*)pos;
  uint16* p16 = (uint16*)pos;
  uint32* p32 = (uint32*)pos;

  p8[0] = 0;
  p8[0] = p8[0] | (v << 6);
  p8[0] = p8[0] | subtype;

  p32[1] = htonl(pid);

  memcpy(pos + 8, name, 4);

  length = 4;
  if (dlength > 0) {
    memcpy(pos + 12, data, dlength);
    length += dlength;
  }

  // global padding requested for this packet ?
  if (globalPadd != 0) {
    for (int i = 0; i < globalPadd - 1; ++i)
      p8[8 + length + i] = 0;

    length += globalPadd;
    p8[8 + length - 1] = globalPadd;

    p = 1;
  }

  p8[1] = 0;
  p8[1] = p8[1] | (p << 7);
  p8[1] = p8[1] | type;

  length /= 4;  // length gives packet size in 32 bit words
  p16[1] = htons(length);

  return 8 + length*4;
}


//
// Functions of class RtcpiCompoundPacket
//
void RTPI::RtcpiCompoundPacket::SetPaddingBorder(int border) {
  if ((border % 4) != 0) // packet mus be aligned to a multiple of 4 bytes (32 bits)
    throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT,
			"RtcpiCompoundPacket::SetPaddingBorder - packet must be aligned to a multiple of 4 bytes (32 bits)");

  paddingBorder = border;
}

void RTPI::RtcpiCompoundPacket::Print() {
  std::cout << "RtcpiCompoundPacket: " << packets.size() << " packets" << std::endl;
  for (RtcpiPacketVector::iterator i = packets.begin(); i != packets.end(); ++i) {
    (*i)->Print();
    std::cout << "----------" << std::endl;
  }
}


// returns overall size of compound packet including padding
int RTPI::RtcpiCompoundPacket::GetSize() {
  int size = 0;
  for (RtcpiPacketVector::iterator i = packets.begin(); i != packets.end(); ++i)
    size += (*i)->GetSize();

  // padding needed ?
  if ((size % paddingBorder) != 0) {
    // compute padding length in bytes
    size += paddingBorder - (size % paddingBorder);
  }

  return size;
}


void RTPI::RtcpiCompoundPacket::AddPacket(RtcpiPacket* _packet, int _max_size) {
  if (_packet != NULL) {
    // check if enough space
    if (_max_size != RtcpiCompoundPacket::KEEP_MAX_PKT_SIZE)
      max_size = _max_size;

    // check size?
    if (max_size > 0) {
      int size = 0;
      for (RtcpiPacketVector::iterator i = packets.begin(); i != packets.end(); ++i)
	size += (*i)->GetSize();

      if (size + _packet->GetSize() > max_size)
	throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RtcpiCompoundPacket::AddPacket - maximum packet size reached");
    }

    packets.push_back(_packet);
  }
}

int RTPI::RtcpiCompoundPacket::Flush() {
  if (buf == RefCount<Buffer>())
    throw RtpiException(RtpiException::ERR_RTPI_INTERNAL, "RtcpiCompoundPacket::Flush - no buffer.");

  // compute needed buffer space and check if padding bytes needs to be attached
  int sizeNeeded = 0;
  int padd = 0;
  for (RtcpiPacketVector::iterator i = packets.begin(); i != packets.end(); ++i)
    sizeNeeded += (*i)->GetSize();
  if ((sizeNeeded % paddingBorder) != 0) {
    // compute padding length in bytes
    padd = paddingBorder - (sizeNeeded % paddingBorder);
    sizeNeeded += padd;
  }

  if ((max_size > 0) && (sizeNeeded > max_size))
    throw RtpiException(RtpiException::ERR_RTCPI_FLUSH, "RtcpiCompoundPacket::Flush - size of compound packet exceeds allowed size");

  buf->resize(sizeNeeded);

  char* p = buf->getCont();

  RefCount<RtcpiPacket> packet;
  for (int j = 0; j < packets.size() - 1; j++) {
    packet = packets[j];
    p += packet->Flush(p, packet->GetSize());
  }

  packet = packets.back();
  packet->Flush(p, packet->GetSize() + padd, padd);

  return sizeNeeded;
}


//
// Functions of class RtcpiPacketFactory
//

/*********************************************************************
 Name:      RtcpiPacketFactory::CreateRtcpiPackets
 Param:     buf - buffer where RTCP/I packet are stored
 Result:    pointer to created RTCP/I packet
 Throws:    RTPIException
 Desc:      creates one or more RTCP/I packets from the bitstring stored in buf
 Calledby:
 Uses:
 *********************************************************************/
RTPI::RtcpiCompoundPacket* RTPI::RtcpiPacketFactory::CreateRtcpiPackets(RefCount<Buffer> buf, int parseOpt) {

  if (buf == NULL)    // Buffer is empty
    throw RtpiException(RtpiException::ERR_RTCPI_PARSE, "RtcpiPacketFactory::CreateRtcpiCompoundPacket - Empty packet");

  char* p = buf->getCont();       // retrieve header start position
  char* p_end = p + buf->getSize();
  int bytesRead = 0;

  // check if first packet is SDES
  // - during strict parsing of packets the payload type of the first packet
  //   in a compound packet must be equal to SDES!
  // - during relaxed parsing (for VCR OUTGOING RTCP packets) we don't have this
  //   restriction since we only record SDES, APP and BYE packets to keep the
  //   information
  if (parseOpt == RTCPI_PARSE_STRICT) {
    if (RtcpiPacket::ParseType(p) != RtcpiPacket::PKT_SDES) {
      throw RtpiException(RtpiException::ERR_RTPI_PARSE_HDR, "RtcpiPacketFactoy::CreateRtcpiPacket - RTCP/I compound packets must start with SDES.");
    };
  }

  RtcpiCompoundPacket* compound = new RtcpiCompoundPacket(buf);

  // cycle through packets
  while (p < p_end) {
    try {
      compound->packets.push_back(CreateRtcpiPacket(p, bytesRead, parseOpt)); // bytesRead is updated in CreateRtcpiPacket
      p += bytesRead;
    } catch (RtpiException e) {
      // delete all
      delete compound;
      throw e;
    }
  }

  return compound;
}


/*********************************************************************
 Name:      RtcpiPacketFactory::CreateRtcpiPacket
 Param:     buf - buffer where RTCP/I packet is stored
 Result:    pointer to created RTCP/I packet
 Throws:    RTPIException
 Desc:      creates a single RTCP/I packet from the bitstring stored in buf
 Calledby:
 Uses:
 *********************************************************************/
RefCount<RTPI::RtcpiPacket> RTPI::RtcpiPacketFactory::CreateRtcpiPacket(char* pos, int& bytesRead, int parseOpt) {
  int t = RtcpiPacket::ParseType(pos);


  RefCount<RtcpiPacket> p;
  switch (t) {
  case RtcpiPacket::PKT_SUBREP:    /* subcomponent report */
    if (parseOpt == RTCPI_PARSE_STRICT) {
      p = new RtcpiSubcomponentReportPacket();
    } else {
      p = new RtcpiSubcomponentReportPacket(0, RtcpiSubcomponentReportPacket::EMPTY_ALLOWED);
    }
    bytesRead = p->Parse(pos);
    break;
  case RtcpiPacket::PKT_SDES:      /* source description */
    p = new RtcpiSourceDescriptionPacket();
    bytesRead = p->Parse(pos);
    break;
  case RtcpiPacket::PKT_BYE:
    p = new RtcpiByePacket();
    bytesRead = p->Parse(pos);
    break;
  case RtcpiPacket::PKT_APP:
    p = new RtcpiApplicationPacket();
    bytesRead = p->Parse(pos);
    break;
  default:
    throw RtpiException(RtpiException::ERR_RTCPI_PARSE, "RtcpiPacketFactory::createRtcpiPacket - unknown RTCP/I payload type");
  }
  return p;
}

/*
std::vector<RTPI::RtcpiCompoundPacket*>* RTPI::RtcpiPacketFactory::CreateRtcpiCompoundPackets(RtcpiPacketVector& rtcpiPackets,
											 Refcount<Buffer> buf) {

  int mode = (buf == NULL) ? RtcpiCompoundPacket::INTERNAL_BUFFER_CTRL : RtcpiCompoundPacket::EXTERNAL_BUFFER_CTRL;
  int size = 0;

  if (mode == RtcpiCompoundPacket::EXTERNAL_BUFFER_CTRL) {
    // check if buffer is big enough
    for (RtcpiPacketVector::iterator i = rtcpiPackets.begin(); i != rtcpiPackets.end(); ++i)
      size += (*i)->GetSize();

    buf->resize(size);
  }

  std::vector<RtcpiCompoundPacket*>* packets = new std::vector<RtcpiCompoundPacket*>;

  if (mode == RtcpiCompoundPacket::INTERNAL_BUFFER_CTRL)
    buf = new Buffer(RtcpiCompoundPacket::MAX_PKT_SIZE);
  RtcpiCompoundPacket* compound = new RtcpiCompoundPacket(buf);
  packets->push_back(compound);
  size = 0;

  int bytesFlushed = 0;
  RtcpiPacketVector::iterator next;
  for (RtcpiPacketVector::iterator i = rtcpiPackets.begin(); i != rtcpiPackets.end(); ++i) {
    bytesFlushed = (*i)->Flush(buf->getCont(),(*i)->GetSize());
    size += bytesFlushed;
    buf += bytesFlushed;
    compound->packets.push_back(*i);

    next = i + 1;
    if (next != rtcpiPackets.end()) {
      if (size + (*next)->GetSize() > RtcpiCompoundPacket::MAX_PKT_SIZE) {
	compound->SetSize(size);
	// create new compound packet
	if (mode == RtcpiCompoundPacket::INTERNAL_BUFFER_CTRL)
	  buf = new char[RtcpiCompoundPacket::MAX_PKT_SIZE];
	compound = new RtcpiCompoundPacket(mode,buf);
	packets->push_back(compound);
	size = 0;
      }
    } else {
      compound->SetSize(size);
    }
  }

  return packets;
}
*/








