// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//
// Copyright (c) 2000 University of Mannheim. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for educational, research and non-profit purposes,
// without fee, and without a written agreement is hereby granted,
// provided that the above copyright notice, this paragraph and the
// following four paragraphs appear in all copies.
//
// Permission to incorporate this software into commercial products may
// be obtained from the following address:
// University of Mannheim
// Lehrstuhl fuer Praktische Informatik IV
// L15, 16
// 68161 Mannheim, Germany
//
// IN NO EVENT SHALL THE UNIVERSITY OF MANNHEIM BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
// ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF MANNHEIM HAS BEEN ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//
// THE UNIVERSITY OF MANNHEIM SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER
// IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF MANNHEIM HAS NO OBLIGATIONS
// TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The University of Mannheim will appreciate receiving any enhancements and
// modifications of this software made by others, and reserves the right to
// include them in future releases. The University also recommends that this
// software not be redistributed by third parties, but be obtained by all
// directly from the University's server, so that all users of the software
// may receive bug notifications and other relevant announcements.
//
// Authors:
//   Volker Hilt, Juergen Vogel, Martin Mauve
//   University of Mannheim
//   {hilt,vogel,mauve}@informatik.uni-mannheim.de
//
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/**********************************************************************

 program : rtcpiPacket.h
 authors : Juergen Vogel
 date    : 05.04.2000
 desc    : defines RTCP/I Packets and classes to handle them

 **********************************************************************/


#ifndef __RTCPI_PACKET__
#define __RTCPI_PACKET__

#ifdef WIN32
#  pragma warning (disable: 4786)
#endif


#define RTCPI_MASK_V 0xC0
#define RTCPI_MASK_N 0x20
#define RTCPI_MASK_SC 0x1F
#define RTCPI_MASK_IC 0x1F
#define RTCPI_MASK_SUBTYPE 0x1F
#define RTCPI_MASK_P 0x80
#define RTCPI_MASK_TYPE 0x7F
#define RTCPI_MASK_A 0x80

// parse options
#define RTCPI_PARSE_STRICT 0
#define RTCPI_PARSE_PREPARSED 1


#ifdef WIN32
#include <util/refcount.cpp>
#else
#include <util/refcount.cc>
#endif

#include <vector>
#include <memory.h>

#include <rtpi/rtpiException.h>
#include <util/Buffer.h>
#include <util/types.h>


namespace RTPI {

  class RtcpiPacket {
  public:
    enum types {
      // RTCP/I version
      VERSION = 0,
      // type of RTCP/I packet
      PKT_SUBREP = 70,
      PKT_BYE = 71,
      PKT_SDES = 72,
      PKT_APP = 73
    };

    enum compound_packet_types {
      UNLIMITED_PKT_SIZE = -1,
      KEEP_MAX_PKT_SIZE = 0,
    };

    RtcpiPacket(uint8 t) : v(RtcpiPacket::VERSION), p(0), type(t), length(0), pid(0) {
      if ((type != PKT_SUBREP) && (type != PKT_SDES) && (type != PKT_BYE) && (type != PKT_APP))
	throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RtcpiPacket::RtcpiPacket - wrong type");
    }
    RtcpiPacket(uint8 t, uint32 _pid) : v(RtcpiPacket::VERSION), p(0), type(t), length(0), pid(_pid) {
      if ((type != PKT_SUBREP) && (type != PKT_SDES) && (type != PKT_BYE) && (type != PKT_APP))
	throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RtcpiPacket::RtcpiPacket - wrong type");
    }
    RtcpiPacket(uint8 t, uint32 _pid, uint16 len) : v(RtcpiPacket::VERSION), p(0), type(t), length(len), pid(_pid) {
      if ((type != PKT_SUBREP) && (type != PKT_SDES) && (type != PKT_BYE) && (type != PKT_APP))
	throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RtcpiPacket::RtcpiPacket - wrong type");
    }

    virtual void Print() = 0;

    virtual int GetType() = 0;

    void SetParticipantID(uint32 _pid) {pid = _pid;}
    uint32 GetParticipantID() {return pid;}

    void SetPadding(bool mode = true) {
      p = ((mode) ? 1 : 0);
    }

    virtual int GetSize() = 0; // returns size of packet in bytes without padding

    virtual int Parse(char* pos) = 0;
    virtual int Flush(char* pos, int space, int padd = 0) = 0;

    // get infos from unparsed buffer
    static int ParseType(char* pos);
    static int ParseLength(char* pos);

    uint8  v;       // 2 b: version of RTCP/I
    uint8  p;       // 1 b: padding
    uint8  type;    // 7 b:  type
    uint16 length;  // 16 b: length
    uint32 pid;     // 32 b: participant identifier
  };


  class RtcpiSubcomponentReportPacket;
  class RtpiHandler;
  class RtpiParticipant;

  class RtcpiSubcomponentReportItem {
  public:
    RtcpiSubcomponentReportItem() : subid(0), length(0), name(NULL), active(false) {}
    RtcpiSubcomponentReportItem(uint32 _subid, bool _active, const char* _name = NULL, int _length = 0);
    ~RtcpiSubcomponentReportItem() {if (name != NULL) delete[] name;}

    void Print();

    void SetSubID(uint32 _subid) {subid = _subid;}
    uint32 GetSubID() {return subid;}

    void SetApplicationName(const char* _name, int _length);
    char* GetApplicationName() {return name;}
    int GetApplicationNameLength() {return length;}

    void SetActive(bool _active) {active = _active;}
    bool GetActive() {return active;}

    int GetSize();

    int Parse(char* pos);
    int Flush(char* pos);

  private:
    friend class RtpiHandler;
    friend class RtpiParticipant;
    friend class RtcpiSubcomponentReportPacket;

    uint32 subid;   // 32b: subcomponentID
    uint8  length;  // 8 b: length of application level subcomponent name (opt.), wihout '\0'
    char* name;     // application level subcomponent name
    bool active;    // subcomponent active?
  };

#ifdef WIN32
  using std::vector<RtcpiSubcomponentReportItem*>;
#endif

  class RtcpiSubcomponentReportPacket : public RtcpiPacket {
  public:
    enum types {
      EMPTY_NOT_ALLOWED,
      EMPTY_ALLOWED
    };

    RtcpiSubcomponentReportPacket()
      : RtcpiPacket(RtcpiPacket::PKT_SUBREP), sc(0), mode(EMPTY_NOT_ALLOWED) {}
    RtcpiSubcomponentReportPacket(uint32 _pid, int _mode = EMPTY_NOT_ALLOWED)
      : RtcpiPacket(RtcpiPacket::PKT_SUBREP, _pid),  sc(0), mode(_mode) {}
    ~RtcpiSubcomponentReportPacket();

    void Print();

    int GetSize();

    int Parse(char* pos);
    int Flush(char* pos, int space, int padd = 0);

    int GetType() {return RtcpiPacket::PKT_SUBREP;}

    void AddItem(RtcpiSubcomponentReportItem* item, int max_size = UNLIMITED_PKT_SIZE);
    void AddItem(uint32 subid, bool active, const char* name = NULL, int length = 0, int max_size = UNLIMITED_PKT_SIZE);
    std::vector<RtcpiSubcomponentReportItem*>* GetItems() {return &items;}
    std::vector<RtcpiSubcomponentReportItem*>::iterator
      DeleteItem(std::vector<RtcpiSubcomponentReportItem*>::iterator pos) {
      delete *pos;
      return items.erase(pos);
    }
    bool ContainsItems() { return !items.empty();}

  private:
    friend class RtpiHandler;
    friend class RtpiParticipant;

    uint8  sc;      // 5 b: subcomponent count
    int mode;

    std::vector<RtcpiSubcomponentReportItem*> items;
  };


  class RtcpiSourceDescriptionPacket;

  class RtcpiSourceDescriptionItem {
  public:
    enum types {
      // type of SDES Item
      CNAME = 1,
      NAME = 2,
      EMAIL = 3,
      PHONE = 4,
      LOC = 5,
      TOOL = 6,
      NOTE = 7,
      PRIV = 8
    };

    RtcpiSourceDescriptionItem() : type(0), length(0), data(NULL) {}
    RtcpiSourceDescriptionItem(int _type, const char* _data, unsigned short _length);
    ~RtcpiSourceDescriptionItem() { if (data != NULL) delete[] data;}

    void Print();

    int GetSize() {return 2 + length;}

    void SetData(const char* _data, int length);
    char* GetData() {return data;}
    int GetDataLength() {return length;}

    int Parse(char* pos);
    int Flush(char* pos);

    static int ParseType(char* pos);

    int GetType() {return type;}

  private:
    friend class RtpiHandler;
    friend class RtpiParticipant;
    friend class RtcpiSourceDescriptionPacket;

    uint8 type;   // 8 b: type of SDES item
    uint8 length; // 8 b: length of SDES item in bytes, not counting a '/0' delimiter
    char* data;   // x bytes: source description data (cname, email, etc.)
  };

#ifdef WIN32
  using std::vector<RtcpiSourceDescriptionItem*>;
#endif

  class RtcpiSourceDescriptionPacket : public RtcpiPacket {
  public:
    RtcpiSourceDescriptionPacket()
      : RtcpiPacket(RtcpiPacket::PKT_SDES), ic(0) {}
    RtcpiSourceDescriptionPacket(uint32 _pid)
      : RtcpiPacket(RtcpiPacket::PKT_SDES, _pid), ic(0) {}
    ~RtcpiSourceDescriptionPacket();

    void Print();

    int GetSize();

    int Parse(char* pos);
    int Flush(char* pos, int space, int padd = 0);

    int GetType() {return RtcpiPacket::PKT_SDES;}

    void AddItem(RtcpiSourceDescriptionItem* item, int max_size = UNLIMITED_PKT_SIZE);
    void AddItem(int type, const char* data, int length, int max_size = UNLIMITED_PKT_SIZE);
    std::vector<RtcpiSourceDescriptionItem*>::iterator DeleteItem(std::vector<RtcpiSourceDescriptionItem*>::iterator pos) {
      delete *pos;
      return items.erase(pos);
    }
    bool ContainsItems() {return !items.empty();}
    std::vector<RtcpiSourceDescriptionItem*>* GetItems() {return &items;}

  private:
    friend class RtpiHandler;
    friend class RtpiParticipant;

    // 1 b:  reserved
    uint8  ic;      // 5 b: item count

    std::vector<RtcpiSourceDescriptionItem*> items;
  };


  class RtcpiByePacket : public RtcpiPacket {
  public:
    RtcpiByePacket()
      : RtcpiPacket(RtcpiPacket::PKT_BYE), rlength(0), reason(NULL) {}
    RtcpiByePacket(uint32 _pid)
      : RtcpiPacket(RtcpiPacket::PKT_BYE, _pid), rlength(0), reason(NULL) {}
    RtcpiByePacket(uint32 _pid, const char* _reason, int _rlength);

    ~RtcpiByePacket() {if (reason != NULL) delete[] reason;}

    void Print();

    void SetReason(const char* _reason, int _rlength);
    char* GetReason() {return reason;}
    int GetReasonLength() {return rlength;}

    int GetSize();

    int Parse(char* pos);
    int Flush(char* pos, int spacem, int padd = 0);

    int GetType() {return RtcpiPacket::PKT_BYE;}

  private:
    friend class RtpiHandler;
    friend class RtpiParticipant;

    uint8  rlength; // 8 b:  reason length
    char* reason;   // x bytes: reason for leaving
  };


  class RtcpiApplicationPacket : public RtcpiPacket {
  public:
    RtcpiApplicationPacket()
      : RtcpiPacket(RtcpiPacket::PKT_APP, 0, 1), r(0), subtype(0), data(NULL), dlength(0) {}
    RtcpiApplicationPacket(uint32 _pid)
      : RtcpiPacket(RtcpiPacket::PKT_APP, _pid, 1), r(0), subtype(0), data(NULL), dlength(0) {}
    RtcpiApplicationPacket(uint32 _pid, const char* _name, int _subtype)
      : RtcpiPacket(RtcpiPacket::PKT_APP, _pid, 1), r(0), subtype(_subtype), data(NULL), dlength(0) {
      if ((dlength % 4) != 0)
	throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT,
			    "RtcpiApplicationPacket::RtcpiApplicationPacket - data length is not a multiple of 32 bit");
      memcpy(name, _name, 5);
    }
    RtcpiApplicationPacket(uint32 _pid, const char* _name, int _subtype, const char* _data, int _dlength);
    ~RtcpiApplicationPacket() { if (data != NULL) delete[] data;}

    void Print();

    void SetSubType(int _subtype) {subtype = _subtype;}
    int GetSubType() {return subtype;}

    void SetData(const char* _data, int _dlength);
    char* GetData() {return data;}
    int GetDataLength() {return dlength;}

    int GetSize();

    int Parse(char* pos);
    int Flush(char* pos, int space, int padd = 0);

    int GetType() {return RtcpiPacket::PKT_APP;}

  private:
    friend class RtpiHandler;
    friend class RtpiParticipant;

    uint8  r;       // 1 b:  reserved
    uint8  subtype; // 5 b:  subtype of application specific RTCP/I packet
    char name[5];   // 32 b: name to uniquely identfiy an application  - 1 bit for \0
    char* data;     // application dependent data
    int dlength;    // length of application data in bytes
  };

  typedef std::vector<RefCount<RtcpiPacket> > RtcpiPacketVector;

  class RtcpiPacketFactory;

  class RtcpiCompoundPacket {
  public:
    enum types {
      UNLIMITED_PKT_SIZE = -1,
      KEEP_MAX_PKT_SIZE = 0,
      MAX_PKT_SIZE = 1400      // this value has to be requested from the transport layer!
    };

    RtcpiCompoundPacket() : paddingBorder(4), max_size(GetMaxPacketSize()),
                            buf(RefCount<Buffer>()) {};
    RtcpiCompoundPacket(RefCount<Buffer> _buf) : paddingBorder(4), max_size(GetMaxPacketSize()),
                                                 buf(_buf) {};
    RtcpiCompoundPacket(RefCount<Buffer> _buf, int _max_size) : paddingBorder(4),
                                                                max_size(_max_size),
                                                                buf(_buf) {};

    void Print();

    void SetPaddingBorder(int border);
    int GetPaddingBorder() {return paddingBorder;}

    void AddPacket(RtcpiPacket* _packet, int _max_size = KEEP_MAX_PKT_SIZE);
    RtcpiPacketVector::iterator DeletePacket(RtcpiPacketVector::iterator _pos) {
      return packets.erase(_pos);
    }
    RtcpiPacketVector::iterator GetPacketsBegin() { return packets.begin(); }
    RtcpiPacketVector::iterator GetPacketsEnd() { return packets.end(); }
    bool ContainsPackets() { return !packets.empty(); }
    int GetNumberOfPackets() { return packets.size(); }

    int Flush(); // flush rtcp/i packets into internal buffer

    char* GetBufferStart() {
      if (buf == RefCount<Buffer>())
	throw RtpiException(RtpiException::ERR_RTPI_INTERNAL, "RtcpiCompoundPacket::GetStart - no buffer.");
      return buf->getCont();
    }
    int GetBufferSize() {  // returns size of buffer
      if (buf == RefCount<Buffer>())
	throw RtpiException(RtpiException::ERR_RTPI_INTERNAL, "RtcpiCompoundPacket::GetSize - no buffer.");
      return buf->getSize();
    }
    RefCount<Buffer> GetBuffer() {
      if (buf == RefCount<Buffer>())
	throw RtpiException(RtpiException::ERR_RTPI_INTERNAL, "RtcpiCompoundPacket::GetBuffer - no buffer.");
      return buf;
    }
    void SetBuffer(RefCount<Buffer> _buf) {
      buf = _buf;
    }
    int GetSize();  // calculates overall size of compound packet including padding

    int GetMaxSize() { return max_size; }
    void SetMaxSize(int _max_size) { max_size = _max_size; }
    static int GetMaxPacketSize() { return max_packet_size; }
    static int SetMaxPacketSize(int _m_s) { max_packet_size = _m_s; }

    RtcpiPacketVector packets;

  protected:
    friend class RtcpiPacketFactory;
    friend class RtpiHandler;
    friend class RtpiParticipant;

    void SetSize(int _size) {
      if (buf == RefCount<Buffer>())
	throw RtpiException(RtpiException::ERR_RTPI_INTERNAL, "RtcpiCompoundPacket::GetStart - no buffer.");
      buf->resize(_size);
    }

    int paddingBorder; // align RTCP/I packet to paddingBorder bytes; default is 4
    int max_size;      // maximum packet size

    static int max_packet_size;

    RefCount<Buffer> buf;
  };

  typedef std::vector<RtcpiCompoundPacket*> RtcpiCompoundPacketVector;

  class RtcpiPacketFactory {
  public:
    // parse RTCP/I packets contained in buf
    static RtcpiCompoundPacket* CreateRtcpiPackets(RefCount<Buffer> buf, int parse_opt = RTCPI_PARSE_STRICT);

  private:
    static RefCount<RtcpiPacket> CreateRtcpiPacket(char* pos, int& bytesRead, int parse_opt = RTCPI_PARSE_STRICT);
  };
}


#endif
