// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//
// Copyright (c) 2000 University of Mannheim. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for educational, research and non-profit purposes,
// without fee, and without a written agreement is hereby granted,
// provided that the above copyright notice, this paragraph and the
// following four paragraphs appear in all copies.
//
// Permission to incorporate this software into commercial products may
// be obtained from the following address:
// University of Mannheim
// Lehrstuhl fuer Praktische Informatik IV
// L15, 16
// 68161 Mannheim, Germany
//
// IN NO EVENT SHALL THE UNIVERSITY OF MANNHEIM BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
// ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF MANNHEIM HAS BEEN ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//
// THE UNIVERSITY OF MANNHEIM SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER
// IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF MANNHEIM HAS NO OBLIGATIONS
// TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The University of Mannheim will appreciate receiving any enhancements and
// modifications of this software made by others, and reserves the right to
// include them in future releases. The University also recommends that this
// software not be redistributed by third parties, but be obtained by all
// directly from the University's server, so that all users of the software
// may receive bug notifications and other relevant announcements.
//
// Authors:
//   Volker Hilt, Juergen Vogel, Martin Mauve
//   University of Mannheim
//   {hilt,vogel,mauve}@informatik.uni-mannheim.de
//
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

#include <rtpi/rtpi.h>


RTPI::Rtpi::Rtpi() 
  : rtcpiReport(NULL), rtpiFeedback(NULL), rtcpiPacketFactory(NULL), rtpiPacketFactory(NULL), transport(NULL), rtpiHandler(NULL), joined(false) {
}


RTPI::Rtpi::~Rtpi() {
  delete rtpiHandler; rtpiHandler = NULL;
}


// group membership
RTPI::RtpiParticipant* RTPI::Rtpi::join_group(std::string username, 
					      RtcpiReportInterface* _rtcpiReport, 
					      RtpiFeedbackInterface* _rtpiFeedback, 
					      Transport* rtpiTransport) {

  if (joined) return NULL;

  rtcpiReport = _rtcpiReport;
  rtpiFeedback = _rtpiFeedback;

  rtcpiPacketFactory = new RTPI::RtcpiPacketFactory();
  rtpiPacketFactory = new RTPI::RtpiPacketFactory();
  transport = rtpiTransport;
  
  if (!rtpiHandler) rtpiHandler = new RTPI::RtpiHandler(rtcpiReport, rtpiFeedback);
  rtpiHandler->createMasterParticipant(rtcpiReport->get_local_address(), username);

  joined = true;

  return (*rtpiHandler->getMasterParticipant());
}


void RTPI::Rtpi::leave_group() {
  if (!joined) return;
  rtpiHandler->removeAllParticipants();
  delete transport; transport = NULL;
  delete rtpiPacketFactory; rtpiPacketFactory = NULL;
  delete rtcpiPacketFactory; rtcpiPacketFactory = NULL;
  joined = false;
}


bool RTPI::Rtpi::group_joined() {
  return (joined);
}


// manage participants
RTPI::RtpiParticipant* RTPI::Rtpi::get_own_participant() {
  if (!joined) return NULL;
  return (*rtpiHandler->getMasterParticipant());
}


RTPI::RtpiParticipant* RTPI::Rtpi::get_participant(uint32 pid) {
  if (!joined) return NULL;
  RTPI::RtpiParticipantList::iterator participant = rtpiHandler->findParticipant(pid);
  return ((participant == rtpiHandler->getParticipantsEnd()) ? NULL : *participant);
}


bool RTPI::Rtpi::participant_exists(uint32 pid) {
  if (!joined) return false;
  RTPI::RtpiParticipantList::iterator participant = rtpiHandler->findParticipant(pid);
  return ((participant == rtpiHandler->getParticipantsEnd()) ? false : true);
}


RTPI::RtpiParticipantList::iterator RTPI::Rtpi::get_participants_begin() {
  if (!joined) return rtpiHandler->getParticipantsEnd();
  return rtpiHandler->getParticipantsBegin();
}


RTPI::RtpiParticipantList::iterator RTPI::Rtpi::get_participants_end() {
  return rtpiHandler->getParticipantsEnd();
}


RTPI::RtpiParticipantList::iterator RTPI::Rtpi::get_next_participant(RtpiParticipantList::iterator& i) {
  if (!joined) return rtpiHandler->getParticipantsEnd();
  return rtpiHandler->getNextParticipant(i);
}


// manage subcomponents
void RTPI::Rtpi::add_subcomponent(uint32 subid, std::string name, bool active) {
  if (!joined) return;

  SubComponentMap::iterator rtpiSub = rtpiHandler->get_seq_space()->find_component(subid);
  if (rtpiSub == rtpiHandler->get_seq_space()->get_components_end())
    rtpiSub = rtpiHandler->get_seq_space()->insert_component(subid);

  rtpiSub->second.add_component(active, (char*)name.c_str(), name.length());
  rtpiSub->second.set_active(active);
}


void RTPI::Rtpi::delete_subcomponent(uint32 subid) {
  if (!joined) return;

  SubComponentMap::iterator sub = rtpiHandler->get_seq_space()->find_component(subid);
  if (sub == rtpiHandler->get_seq_space()->get_components_end()) 
    return; // not found

  sub->second.set_active(false);
  rtpiHandler->get_seq_space()->delete_component(subid);
}


RTPI::RtpiSubComponent* RTPI::Rtpi::get_subcomponent(uint32 subid) {
  if (!joined) return NULL;
  SubComponentMap::iterator sub = rtpiHandler->get_seq_space()->find_component(subid);
  return ((sub == rtpiHandler->get_seq_space()->get_components_end()) ? NULL : &(sub->second));
}


bool RTPI::Rtpi::subcomponent_exists(uint32 subid) {
  if (!joined) return false;

  SubComponentMap::iterator sub = rtpiHandler->get_seq_space()->find_component(subid);
  return ((sub == rtpiHandler->get_seq_space()->get_components_end()) ? false : true);
}

    
RTPI::SubComponentMap::iterator RTPI::Rtpi::get_subcomponents_begin() {
  if (!joined) return rtpiHandler->get_seq_space()->get_components_end();
  return rtpiHandler->get_seq_space()->get_components_begin();
}


RTPI::SubComponentMap::iterator RTPI::Rtpi::get_subcomponents_end() {
  return rtpiHandler->get_seq_space()->get_components_end();
}


RTPI::SubComponentMap::iterator RTPI::Rtpi::get_next_subcomponent(SubComponentMap::iterator& i) {
  if (!joined) return rtpiHandler->get_seq_space()->get_components_end();
  return ++i;
}


void RTPI::Rtpi::activate_subcomponent(uint32 subid) {
  if (!joined) return;

  RTPI::SubComponentMap::iterator sub = rtpiHandler->get_seq_space()->find_component(subid);
  if (sub == rtpiHandler->get_seq_space()->get_components_end()) 
    return; // not found

  sub->second.set_active(true);
}


void RTPI::Rtpi::deactivate_subcomponent(uint32 subid) {
  if (!joined) return;

  RTPI::SubComponentMap::iterator sub = rtpiHandler->get_seq_space()->find_component(subid);
  if (sub == rtpiHandler->get_seq_space()->get_components_end()) 
    return; // not found

  sub->second.set_active(false);
}


bool RTPI::Rtpi::subcomponent_is_active(uint32 subid) {
  if (!joined) return false;

  SubComponentMap::iterator sub = rtpiHandler->get_seq_space()->find_component(subid);
  return ((sub == rtpiHandler->get_seq_space()->get_components_end()) ? false : sub->second.is_active());
}


void RTPI::Rtpi::deactivate_all_subcomponents() {
  if (!joined) return;

  for (RTPI::SubComponentMap::iterator i = rtpiHandler->get_seq_space()->get_components_begin();
       i != rtpiHandler->get_seq_space()->get_components_end(); ++i) {
    i->second.set_active(false);
  }
}


void RTPI::Rtpi::set_subcomponent_name(uint32 subid, std::string name) {
  if (!joined) return;
  SubComponentMap::iterator sub = rtpiHandler->get_seq_space()->find_component(subid);
  if (sub == rtpiHandler->get_seq_space()->get_components_end()) return;
  sub->second.add_component(sub->second.is_active(), (char*)name.c_str(), name.length());
}


std::string RTPI::Rtpi::get_subcomponent_name(uint32 subid) {
  if (!joined) return "";
  SubComponentMap::iterator sub = rtpiHandler->get_seq_space()->find_component(subid);
  return ((sub == rtpiHandler->get_seq_space()->get_components_end()) ? "" : string(sub->second.get_name()));
}


bool RTPI::Rtpi::subcomponent_has_name(uint32 subid) {
  if (!joined) return false;
  SubComponentMap::iterator sub = rtpiHandler->get_seq_space()->find_component(subid);
  return ((sub == rtpiHandler->get_seq_space()->get_components_end()) ? false : ((sub->second.get_name_length() == 0) ? false : true));
}


// helper functions
void RTPI::Rtpi::send_rtcpi_note(std::string note) {
  if (!joined) return;
  rtpiHandler->sendRtcpiNote();
}


uint32 RTPI::Rtpi::get_current_time() {
  return rtpitime();
}


// outgoing data
RefCount<RTPI::BufferVector> RTPI::Rtpi::create_rtpi_packets(RefCount<RtpiAdu> adu) {
  if (!joined || !adu) return NULL;
  
  RefCount<RtpiAduVector> adus = new RTPI::RtpiAduVector;
  adus->push_back(adu);

  rtpiHandler->handle_rtpi(adus, rtcpiReport->get_local_address(), RTPI_PACKET_OUTGOING);

  RefCount<RtpiPacketVector> rtpiPackets = transport->CreateRtpiPackets(adu);

  RefCount<BufferVector> bufVector = new BufferVector();
  for (RtpiPacketVector::iterator i = rtpiPackets->begin(); i != rtpiPackets->end(); ++i)
    bufVector->push_back((*i)->getBuffer());

  return bufVector;
}


RefCount<RTPI::BufferVector> RTPI::Rtpi::create_rtpi_packets(RefCount<RtpiAduVector> adus) {
  if (!joined || !adus || (adus->size() == 0)) return NULL;
  
  rtpiHandler->handle_rtpi(adus, rtcpiReport->get_local_address(), RTPI_PACKET_OUTGOING);
  
  RefCount<RtpiCompoundPacketVector> rtpiPackets = transport->CreateRtpiPackets(adus);
  
  RefCount<BufferVector> bufVector = new BufferVector();
  for (RtpiCompoundPacketVector::iterator i = rtpiPackets->begin(); i != rtpiPackets->end(); ++i)
    bufVector->push_back((*i)->GetBuffer());
  
  return bufVector;
}


// incoming data
RefCount<RTPI::RtpiAduVector> RTPI::Rtpi::create_rtpi_adus(RefCount<Buffer> buf) {
  if (!joined || !buf) return NULL;
  return (transport->CreateRtpiAdusFromPackets(rtpiPacketFactory->CreateRtpiPackets(buf)));
}


void RTPI::Rtpi::handle_received_rtpi(RefCount<RtpiAdu> adu, uint32 senderIpAddress) {
  if (!joined || !adu) return;
  RefCount<RtpiAduVector> adus = new RtpiAduVector;
  adus->push_back(adu);
  rtpiHandler->handle_rtpi(adus, senderIpAddress, RTPI_PACKET_INCOMING);
}


void RTPI::Rtpi::handle_received_rtpi(RefCount<RtpiAduVector> adus, uint32 senderIpAddress) {
  if (!joined || !adus) return;
  rtpiHandler->handle_rtpi(adus, senderIpAddress, RTPI_PACKET_INCOMING);
}


void RTPI::Rtpi::handle_received_rtcpi(RefCount<Buffer> buf, uint32 senderIpAddress) {
  if (!joined || !buf) return;
  RefCount<RtcpiCompoundPacket> rtcpiCompoundPacket = rtcpiPacketFactory->CreateRtcpiPackets(buf);
  rtpiHandler->handle_rtcpi(rtcpiCompoundPacket, senderIpAddress);
}


