// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//
// Copyright (c) 2000 University of Mannheim. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for educational, research and non-profit purposes,
// without fee, and without a written agreement is hereby granted,
// provided that the above copyright notice, this paragraph and the
// following four paragraphs appear in all copies.
//
// Permission to incorporate this software into commercial products may
// be obtained from the following address:
// University of Mannheim
// Lehrstuhl fuer Praktische Informatik IV
// L15, 16
// 68161 Mannheim, Germany
//
// IN NO EVENT SHALL THE UNIVERSITY OF MANNHEIM BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
// ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF MANNHEIM HAS BEEN ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//
// THE UNIVERSITY OF MANNHEIM SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER
// IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF MANNHEIM HAS NO OBLIGATIONS
// TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The University of Mannheim will appreciate receiving any enhancements and
// modifications of this software made by others, and reserves the right to
// include them in future releases. The University also recommends that this
// software not be redistributed by third parties, but be obtained by all
// directly from the University's server, so that all users of the software
// may receive bug notifications and other relevant announcements.
//
// Authors:
//   Volker Hilt, Juergen Vogel, Martin Mauve
//   University of Mannheim
//   {hilt,vogel,mauve}@informatik.uni-mannheim.de
//
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/**********************************************************************

 program : rtpiPacket.cpp
 authors : Juergen Vogel
 date    : 05.04.2000
 desc    : defines RTP/I Packets and classes to handle them

 **********************************************************************/

#ifdef WIN32
#include <winsock.h>
#else
#include <arpa/inet.h>
#endif

#include <math.h>

#include <rtpi/rtpiPacket.h>


//
// Functions from class RtpiHeader
//
RTPI::RtpiHeader* RTPI::RtpiHeader::Copy() {
  RtpiHeader* ret = new RtpiHeader();
  ret->v = v;
  ret->e = e;
  ret->x = x;
  ret->type = type;
  ret->p = p;
  ret->rt = rt;
  ret->length = length;
  ret->pid = pid;
  ret->subid = subid;
  ret->sn = sn;
  ret->fc = fc;
  return ret;
}


/*********************************************************************
 Name:     RtpiHeader::Parse
 Param:    pos  - pointer to packet to be parsed
 Result:   size of RTP/I Header
 Throws:   RtpiException
 Desc:     parses the packet stored in pos as RTP/I packet and stores the values in
           the corresponding attributes of the RtpiHeader object
 Calledby:
 Uses:
 *********************************************************************/
int RTPI::RtpiHeader::Parse(char* pos) {
  // see if there is something to do...
  if (pos == NULL)
    throw RtpiException(RtpiException::ERR_RTPI_PARSE_HDR, "RtpiHeader::Parse - Buffer empty");

  uint8*  p8  = (uint8*)  pos;
  uint16* p16 = (uint16*) pos;
  uint32* p32 = (uint32*) pos;

  v    = (p8[0] & RTPI_MASK_V) >> 6;
  e    = (p8[0] & RTPI_MASK_E) >> 5;
  x    = (p8[0] & RTPI_MASK_X) >> 4;
  type =  p8[0] & RTPI_MASK_TYPE;
  p    = (p8[1] & RTPI_MASK_P) >> 7;
  rt   =  p8[1] & RTPI_MASK_RT;
  length = ntohs(p16[1]);

  pid = ntohl(p32[1]);

  subid = ntohl(p32[2]);

  sn = ntohs(p16[6]);
  fc = ntohs(p16[7]);

  // RTP/I header validy check
  if (v != RtpiHeader::VERSION)
    throw RtpiException(RtpiException::ERR_RTPI_WRONG_VERSION, "RtpiHeader::Parse - wrong Version of RTP/I Header");

  if ((type != PKT_EVENT) && (type != PKT_CUE) && (type != PKT_STATE) && (type != PKT_DELTA_STATE) && (type != PKT_STATE_QUERY))
    throw RtpiException(RtpiException::ERR_RTPI_PACKET_FORMAT, "RtpiHeader::Parse - wrong type");

  return GetSize(); // return length of RTP/I header in uint8s
}


/*********************************************************************
 Name:      RtpiHeader::Flush
 Param:     pos - pointer to buffer where to write the RTP/I header
 Result:    number of bytes flushed
 Throws:    RtpiException
 Desc:      flushs RTP/I-Header starting at address pos
 Calledby:
 Uses:
 *********************************************************************/
int RTPI::RtpiHeader::Flush(char* pos) {
  // turning host uint8 order to network uint8 order and copy:
  uint8  *p8  = (uint8*)  pos;
  uint16 *p16 = (uint16*) pos;
  uint32 *p32 = (uint32*) pos;

  p8[0] = 0;
  p8[0] = p8[0] | (v << 6);
  p8[0] = p8[0] | (e << 5);
  p8[0] = p8[0] | (x << 4);
  p8[0] = p8[0] | type;

  p8[1] = 0;
  p8[1] = p8[1] | (p << 7);
  p8[1] = p8[1] | rt;

  p16[1] = htons(length);

  p32[1] = htonl(pid);

  p32[2] = htonl(subid);

  p16[6] = htons(sn);
  p16[7] = htons(fc);

  return GetSize();
}

void RTPI::RtpiHeader::Print() {
  std::cout << "Packet:(v=" << (int)v << ", e=" << (int)e << ", x=" << (int)x
       << ", type=" << (int)type << ", p=" << (int)p << ", rt=" << (int)rt
       << ", length=" << (int)length << ", pid=" << (int)pid << ", subid=" << (int)subid
       << ", sn=" << (int)sn << ", fc=" << (int)fc << ")" << std::endl;
}


/*********************************************************************
 Name:      RtpiHeader::ParseType
 Param:     pos - pointer to buffer where packet is kept
 Result:    type (EVENT, STATE, DELTA_STATE, STATE_QUERY, CUE)
 Throws:    RtpiException
 Desc:      interprets structure in pos as RTP/I Header and returns its type
 Calledby:
 Uses:
 *********************************************************************/
int RTPI::RtpiHeader::ParseType(char* pos) {
  uint8* p8 = (uint8*)pos;
  int type = (int)(p8[0] & RTPI_MASK_TYPE);

  // RTP/I header validy check
  if (((p8[0] & RTPI_MASK_V) >> 6) != RtpiHeader::VERSION)
    throw RtpiException(RtpiException::ERR_RTPI_WRONG_VERSION, "RtpiHeader::ParseType - wrong Version of RTP/I-Header");

  if ((type != PKT_EVENT) && (type != PKT_CUE) && (type != PKT_STATE) && (type != PKT_DELTA_STATE) && (type != PKT_STATE_QUERY))
    throw RtpiException(RtpiException::ERR_RTPI_PACKET_FORMAT, "RtpiHeader::ParseType - wrong type");

  return type;
}


int RTPI::RtpiHeader::ParseSize(char* pos) {
  uint8* p8 = (uint8*)pos;

  // RTP/I header validy check
  if (((p8[0] & RTPI_MASK_V) >> 6) != RtpiHeader::VERSION)
    throw RtpiException(RtpiException::ERR_RTPI_WRONG_VERSION, "RtpiHeader::ParseSize - wrong Version of RTP/I-Header");

  return (GetSize() + ntohs(((uint16*)pos)[1]));
}


//
// Functions of class RtpiHeaderExtension
//
int RTPI::RtpiHeaderExtension::size = 0;


int RTPI::RtpiHeaderExtension::Parse(char* pos) {
  uint8* p8 = (uint8*)pos;

  int length = 4 * p8[0] + 3;

  if ((length + 1) != size)
    throw RtpiException(RtpiException::ERR_RTPI_PARSE_HDR, "RtpiHeaderExtension::Parse - wrong size");

  data = new char[length];
  memcpy(data, pos + 1, length);

  return size;
}


int RTPI::RtpiHeaderExtension::Flush(char* pos) {
  uint8* p8 = (uint8*)pos;
  p8[0] = (size / 4) - 1 ; // length stored in p8[0] in 4 byte units - padding has already been done

  memcpy (pos + 1, data, size - 1);

  return size;
}


//
// Functions of class RtpiPacket
//
int RTPI::RtpiPacket::paddingBorder = 4;

std::ostream& operator<<(std::ostream& os, const RTPI::RtpiPacket& p) {
  return os;
}


int RTPI::RtpiPacket::GetSize() {
  if ((hdr == NULL) || (payloadSize == 0))
    return -1;

  int size = hdr->GetSize() + payloadSize;
  if (hdrX) size += hdrX->GetSize();

  if (size % paddingBorder != 0) // padding ?
    size += (paddingBorder - (size % paddingBorder));

  return size;
}


int RTPI::RtpiPacket::GetHeaderSize() {
  int size = hdr->GetSize();
  if (HeaderExtension()) size += hdrX->GetSize();
  return size;
}


void RTPI::RtpiPacket::SetPaddingBorder(int border) {
  if ((border % 4) != 0) // packet mus be aligned to a multiple of 4 bytes (32 bits)
    throw RtpiException(RtpiException::ERR_RTPI_PACKET_FORMAT,
			"RtpiPacket::SetPaddingBorder - packet must be aligned to a multiple of 4 bytes (32 bits)");

  paddingBorder = border;
}


void RTPI::RtpiPacket::Print() {
  std::cout << "RTP/I packet - size " << GetSize() << " Payload Size " << payloadSize << std::endl;
  if (hdr) hdr->Print();
  if (hdrX) hdrX->Print();
}


/*********************************************************************
 Name:      RtpiPacket::Parse
 Param:     pos - place where packet is stored
 Result:    total bytes parsed, including padding
 Throws:    RtpiException
 Desc:      Parses RTP/I-Header of packet stored in buffer
 Calledby:
 Uses:
 *********************************************************************/
int RTPI::RtpiPacket::Parse(RefCount<Buffer> _buf, char* pos) {
  // we discard an (probably existing) old buffer
  buf = _buf;
  headerStart = pos;

  int bytesParsed = 0;
  int totalBytesParsed = 0;
  char* p = headerStart;

  // create RTP/I Header
  hdr = new RtpiHeader();
  bytesParsed = hdr->Parse(p);
  totalBytesParsed += bytesParsed;
  p += bytesParsed;

  payloadStart = p;
  payloadSize = hdr->length;
  totalBytesParsed += payloadSize;

  // create RTP/I Header Extension (opt.)
  if (hdr->x) {
    hdrX = new RtpiHeaderExtension();
    bytesParsed = hdrX->Parse(p);
    totalBytesParsed += bytesParsed;
    payloadStart += bytesParsed;
    payloadSize -= bytesParsed;
  }

  // padding attached ?
  if (hdr->p) {
    int paddingSize = ((uint8*)pos)[totalBytesParsed - 1];  // padding size is stored in last byte of packet
    payloadSize -= paddingSize;  // padding field has one byte as well
  }

  return totalBytesParsed;
}


/*********************************************************************
 Name:      RtpiPacket::Flush
 Param:     pos - place to flush the RTP/I packet
 Result:    number of bytes flushed
 Throws:    RtpiException
 Desc:      flushes the RTP/I packet to the buffer, including RTP/I-Header,
            RTP/I-Header Extension (if avail.) and padding
            payload is already in place
 Calledby:
 Uses:
 *********************************************************************/
int RTPI::RtpiPacket::Flush() {
  // check if we have header and buffer
  if (hdr == NULL)
    throw RtpiException(RtpiException::ERR_RTPI_FLUSH_HDR, "RtpiPacket::Flush - Header is empty (=NULL)!");
  if (buf == RefCount<Buffer>())
    throw RtpiException(RtpiException::ERR_RTPI_FLUSH_HDR, "RtpiPacket::Flush - No Buffer to flush to (=NULL)!");
  if (headerStart == NULL)
    headerStart = buf->getCont();

  hdr->length = payloadSize;

  // reserve space for header extension
  if (RtpiHeaderExtension::GetSize() != 0) {
    hdr->length += hdrX->GetSize();
    hdr->x = 1;
  }

  // padding needed ?
  int padd;
  if ((hdr->GetSize() + hdr->length) % paddingBorder == 0) // incl. Header !
    // no padding needed
    padd = 0;
  else
    // compute padding length in bytes
    padd = paddingBorder - ((hdr->GetSize() + hdr->length) % paddingBorder);

  if (padd == 0) {
    hdr->p = 0;
  } else {
    hdr->p = 1;
    uint8* p8 = (uint8*)headerStart;
    p8[hdr->GetSize() + hdr->length + padd - 1] = padd; // write padding size in bytes into last byte of packet
    for (int i = 0; i < padd - 1; ++i) p8[hdr->GetSize() + hdr->length + i] = 0;
    hdr->length += padd;
  }

  // flush packet header
  int bytesFlushed = hdr->Flush(headerStart);

  // flush header extension if available; if not simulate
  // juergi: die laenge der header extension muss eigentlich noch ins length feld -> ist schon drin
  // damit wuerde dann der rueckgabewert aber nicht mehr stimmen.
  if (RtpiHeaderExtension::GetSize() != 0)
    if (hdr->x)
      bytesFlushed += hdrX->Flush(headerStart + bytesFlushed);
    else
      bytesFlushed += RtpiHeaderExtension::GetSize();

  // put payload in place
  // maybe we should take care about this case - juergi?
  if (((payloadStart + payloadSize) > headerStart) &&
      (payloadStart < (headerStart + bytesFlushed))) {
    throw RtpiException(RtpiException::ERR_RTPI_FLUSH_HDR, "RtpiPacket::Flush - misaligned payload - payload data was overwritten by header.");
  }
  // move payload ot the right place
  if (payloadStart != (headerStart + bytesFlushed)) {
    memmove((headerStart + bytesFlushed), payloadStart, payloadSize);
    payloadStart = headerStart + bytesFlushed;
  }

  return bytesFlushed + hdr->length;
}


//
// Functions from class RtpiPacketFactory
//
RTPI::RtpiPacketVector* RTPI::RtpiPacketFactory::CreateRtpiPackets(RefCount<Buffer> buf) {
  if (buf == RefCount<Buffer>())    // Buffer is empty
    throw RtpiException(RtpiException::ERR_RTPI_PARSE, "RtpiPacketFactory::CreateRtpiPacket - Empty packet");

  RtpiPacketVector* packets = new RtpiPacketVector;

  char* p = buf->getCont();
  char* p_end = p + buf->getSize();
  int bytesRead = 0;

  // cycle through packets
  while (p < p_end) {
    try {
      packets->push_back(CreateRtpiPacket(buf, p, bytesRead)); // bytesRead is updated in CreateRtpiPackets
      p += bytesRead;
    } catch (RtpiException e) {
      delete packets;
      throw e;
    }
  }

  return packets;
}


RefCount<RTPI::RtpiPacket> RTPI::RtpiPacketFactory::CreateRtpiPacket(RefCount<Buffer> _buf, char* pos, int& bytesRead) {
  RefCount<RtpiPacket> rtpi = new RtpiPacket();
  bytesRead = rtpi->Parse(_buf, pos);
  return rtpi;
}


//
// Functions of class RtpiEventHeader
//
RTPI::RtpiAduHeader* RTPI::RtpiEventHeader::Copy() {
  RtpiEventHeader* ret = new RtpiEventHeader();
  ret->pt = pt;
  ret->pi = pi;
  ret->ts = ts;
  ret->o = o;
  if (o) ret->objectSubid = objectSubid;
  return ret;
}


int RTPI::RtpiEventHeader::Parse(char* pos) {
  uint8  *p8  = (uint8*)  pos;
  uint32 *p32 = (uint32*) pos;

  pt = p8[0];
  pi = p8[1];

  o = (p8[2] & RTPI_MASK_O) >> 7;

  ts = ntohl(p32[1]);

  if (o)
    objectSubid  = ntohl(p32[2]);

  return GetSize();
}


int RTPI::RtpiEventHeader::Flush(char* pos) {
  uint8  *p8  = (uint8*)  pos;
  uint32 *p32 = (uint32*) pos;

  p32[0] = 0;

  p8[0] = pt;
  p8[1] = pi;

  p8[2] = p8[2] | (o << 7);

  p32[1] = htonl(ts);

  if (o)
    p32[2] = htonl(objectSubid);

  return GetSize();
}

void RTPI::RtpiEventHeader::Print() {
  std::cout << "EVENT:(pt=" << (int)pt << ", pi=" << (int)pi << ", o=" << (int)o << ", ts=" << (uint32)ts;
  if (o) std::cout << "pSubid=" << (int)objectSubid;
}

//
// Functions of class RtpiCueHeader
//
RTPI::RtpiAduHeader* RTPI::RtpiCueHeader::Copy() {
  RtpiCueHeader* ret = new RtpiCueHeader();
  ret->pt = pt;
  ret->pi = pi;
  ret->ts = ts;
  return ret;
}


int RTPI::RtpiCueHeader::Parse(char* pos) {
  uint8  *p8  = (uint8*)  pos;
  uint32 *p32 = (uint32*) pos;

  pt = p8[0];
  pi = p8[1];

  ts = ntohl(p32[1]);

  return GetSize();
}


int RTPI::RtpiCueHeader::Flush(char* pos) {
  uint8  *p8  = (uint8*)  pos;
  uint32 *p32 = (uint32*) pos;

  p32[0] = 0;
  p8[0] = pt;
  p8[1] = pi;
  p32[1] = htonl(ts);

  return GetSize();
}

void RTPI::RtpiCueHeader::Print() {
  std::cout << "CUE:(pt=" << (int)pt << ", pi=" << (int)pi << ", ts=" << (uint32)ts;
}


//
// Functions of class RtpiStateHeader
//
RTPI::RtpiAduHeader* RTPI::RtpiStateHeader::Copy() {
  RtpiStateHeader* ret = new RtpiStateHeader();
  ret->pt = pt;
  ret->pi = pi;
  ret->ts = ts;
  ret->o  = o;
  ret->snc = snc;
  ret->pri = pri;
  for (std::vector<uint32>::iterator i = pids.begin(); i != pids.end(); ++i)
    ret->pids.push_back(*i);
  for (std::vector<uint16>::iterator j = sns.begin(); j != sns.end(); ++j)
    ret->sns.push_back(*j);
  ret->objectSubid = objectSubid;
  return ret;
}


int RTPI::RtpiStateHeader::GetSize() {
  int n = ((snc % 2) == 0) ? snc : snc + 1;
  return (2*sizeof(uint32) + snc*sizeof(uint32) + n*sizeof(uint16)) + (o ? sizeof(uint32) : 0);
}

int RTPI::RtpiStateHeader::Parse(char* pos) {
  uint8  *p8  = (uint8*)  pos;
  uint16 *p16 = (uint16*) pos;
  uint32 *p32 = (uint32*) pos;

  pt  = p8[0];
  pi  = p8[1];
  snc = p8[2];
  o = (p8[3] & RTPI_MASK_O) >> 7;
  pri = p8[3] & RTPI_MASK_PRI;

  ts = ntohl(p32[1]);

  for (int i = 0; i < snc; ++i)
    pids.push_back(ntohl(p32[2+i]));

  int start = 4 + 2*snc;
  for (int j = 0; j < snc; ++j)
    sns.push_back(ntohs(p16[start+j]));

  objectSubid = ntohl(p32[2 + snc + (snc+1)/2]);

  return GetSize();
}


int RTPI::RtpiStateHeader::Flush(char* pos) {
  uint8  *p8  = (uint8*)  pos;
  uint16 *p16 = (uint16*) pos;
  uint32 *p32 = (uint32*) pos;

  p32[0] = 0;
  p8[0] = pt;
  p8[1] = pi;
  p8[2] = snc;
  p8[3] = p8[3] | (o << 7);
  p8[3] = p8[3] | pri;

  p32[1] = htonl(ts);

  for (int i = 0; i < snc; ++i)
    p32[2+i] = htonl(pids[i]);

  int start = 4 + 2*snc;
  for (int j = 0; j < snc; ++j)
    p16[start+j] = htons(sns[j]);

  if ((snc % 2) != 0)
    // padding
    p16[start+snc] = 0;

  if (o)
    p32[2 + snc + (snc+1)/2] = htonl(objectSubid);

  return GetSize();
}


void RTPI::RtpiStateHeader::Print() {
  std::cout << "STATE:(pt=" << (int)pt << ", pi=" << (int)pi << ", snc=" << (int)snc << ", h=" << int(o) << ", pri=" << (int)pri << ", ts=" << (uint32)ts;
  for (int i = 0; i < snc; ++i) std::cout << ", pid[" << i << "]=" << (int)pids[i];
  for (int j = 0; j < snc; ++j) std::cout << ", sns[" << j << "]=" << (int)sns[j];
  if (o) std::cout << ", pSubid=" << (int)objectSubid;
}

//
// Functions of class RtpiDeltaHeader
//
RTPI::RtpiAduHeader* RTPI::RtpiDeltaStateHeader::Copy() {
  RtpiDeltaStateHeader* ret = new RtpiDeltaStateHeader();
  ret->pt = pt;
  ret->pi = pi;
  ret->ts = ts;
  ret->o  = o;
  ret->snc = snc;
  ret->pri = pri;
  for (std::vector<uint32>::iterator i = pids.begin(); i != pids.end(); ++i)
    ret->pids.push_back(*i);
  for (std::vector<uint16>::iterator j = sns.begin(); j != sns.end(); ++j)
    ret->sns.push_back(*j);
  ret->objectSubid = objectSubid;
  ret->dsn = dsn;
  ret->dpid = dpid;
  return ret;
}


int RTPI::RtpiDeltaStateHeader::GetSize() {
  int n = ((snc % 2) == 0) ? snc : snc + 1;
  return (4*sizeof(uint32) + snc*sizeof(uint32) + n*sizeof(uint16)) + (o ? sizeof(uint32) : 0);
}


int RTPI::RtpiDeltaStateHeader::Parse(char* pos) {
  uint8  *p8  = (uint8*)  pos;
  uint16 *p16 = (uint16*) pos;
  uint32 *p32 = (uint32*) pos;

  pt = p8[0];
  pi = p8[1];
  snc = p8[2];
  o   = (p8[3] & RTPI_MASK_O) >> 7;
  pri = p8[3] & RTPI_MASK_PRI;

  ts = ntohl(p32[1]);

  for (int i = 0; i < snc; ++i)
    pids.push_back(ntohl(p32[2+i]));

  int start = 4 + 2*snc;
  for (int j = 0; j < snc; ++j)
    sns.push_back(ntohs(p16[start+j]));

  int offset = 2 + snc + (snc + 1)/2;
  dpid = ntohl(p32[offset]);

  dsn = ntohs(p16[offset*2 + 2]);

  if (o)
    objectSubid = ntohl(p32[offset + 2]);

  return GetSize();
}


int RTPI::RtpiDeltaStateHeader::Flush(char* pos) {
  uint8  *p8  = (uint8*)  pos;
  uint16 *p16 = (uint16*) pos;
  uint32 *p32 = (uint32*) pos;

  p32[0] = 0;
  p8[0] = pt;
  p8[1] = pi;
  p8[2] = snc;
  p8[3] = p8[3] | (o << 7);
  p8[3] = p8[3] | pri;

  p32[1] = htonl(ts);

  for (int i = 0; i < snc; ++i)
    p32[2+i] = htonl(pids[i]);

  int start = 4 + 2*snc;
  for (int j = 0; j < snc; ++j)
    p16[start+j] = htons(sns[j]);

  if ((snc % 2) != 0)
    // padding
    p16[start+snc] = 0;

  int offset = 2 + snc + (snc + 1)/2;
  p32[offset] = htonl(dpid);

  p32[offset+1] = 0;
  p16[offset*2 + 2] = htons(dsn);

  if (o)
    p32[offset + 2] = htonl(objectSubid);

  return GetSize();
}


void RTPI::RtpiDeltaStateHeader::Print() {
  std::cout << "DELTASTATE:(pt=" << (int)pt << ", pi=" << (int)pi << ", snc=" << (int)snc << ", h=" << int(o) << ", pri=" << (int)pri << ", ts=" << (uint32)ts;
  for (int i = 0; i < snc; ++i) std::cout << ", pid[" << i << "]=" << (int)pids[i];
  for (int j = 0; j < snc; ++j) std::cout << ", sns[" << j << "]=" << (int)sns[j];
  std::cout << ", dpid=" << (int)dpid << ", dsn=" << (int)dsn;
  if (o) std::cout << ", pSubid=" << (int)objectSubid;
}


//
// Functions of class RtpiStateQueryHeader
//
RTPI::RtpiAduHeader* RTPI::RtpiStateQueryHeader::Copy() {
  RtpiStateQueryHeader* ret = new RtpiStateQueryHeader();
  ret->pt = pt;
  ret->pi = pi;
  ret->ts = ts;
  ret->pri = pri;
  return ret;
}


int RTPI::RtpiStateQueryHeader::Parse(char* pos) {
  uint8  *p8  = (uint8*)  pos;
  uint32 *p32 = (uint32*) pos;

  pt = p8[0];
  pi = p8[1];
  pri = p8[3] & RTPI_MASK_PRI;
  ts = ntohl(p32[1]);

  return GetSize();
}


int RTPI::RtpiStateQueryHeader::Flush(char* pos) {
  uint8  *p8  = (uint8*)  pos;
  uint32 *p32 = (uint32*) pos;

  p32[0] = 0;
  p8[0] = pt;
  p8[1] = pi;
  p8[3] = pri;
  p32[1] = htonl(ts);

  return GetSize();
}


void RTPI::RtpiStateQueryHeader::Print() {
  std::cout << "STATEQUERY:(pt=" << (int)pt << ", pi=" << (int)pi << ", pri" << (int)pri << ", ts=" << (uint32)ts;
}


//
// Functions from class RtpiAdu
//
RTPI::RtpiAdu::~RtpiAdu() {
  if (aduHeader)
    delete aduHeader;
  if (packetHeader)
    delete packetHeader;
}


int RTPI::RtpiAdu::GetAduSize() {
  return payloadSize + aduHeader->GetSize();
}


int RTPI::RtpiAdu::Flush() {
  if (packetHeader == NULL || aduHeader == NULL)
    throw RtpiException(RtpiException::ERR_RTPI_FLUSH_ADU, "RtpiAdu::Flush - Header does not exist (nothing to flush)");

  if (buf == RefCount<Buffer>())
    throw RtpiException(RtpiException::ERR_RTPI_FLUSH_ADU, "RtpiAdu::Flush - No buffer to flush to");


  // check if adu has changed in size since parsing because of change in header size
  if (buf->getSize() < GetPacketSize()) {        // header size has increased

    // enlarge buffer and move payload to the right
    int offset = GetPacketSize() - buf->getSize();

    //buf->resize(GetPacketSize());
    buf->shiftCont(offset); // resize included

    // reset pointers
    packetHeaderStart = buf->getCont();
    aduHeaderStart = packetHeaderStart + packetHeader->GetSize();
    payload += offset; // payload = aduHeaderStart + aduHeader->GetSize()

  } else if (buf->getSize() > GetPacketSize()) { // header size has decreased

    // move headers to the right; payload stays in place
    int offset = buf->getSize() - GetPacketSize();

    buf->setBegin(offset);
    packetHeaderStart = buf->getCont();
    aduHeaderStart = packetHeaderStart + packetHeader->GetSize();
  }

  // flush headers; payload already in place
  packetHeader->Flush(packetHeaderStart);
  aduHeader->Flush(aduHeaderStart);

  buf->setSize(GetPacketSize());

  return GetPacketSize();
}


RTPI::RtpiAdu* RTPI::RtpiAdu::CreateAndParse(RefCount<Buffer> buf) {
  // create one  RTP/I ADU from buf  ( RTP/I packet header + ADU header + payload)
  RtpiAdu* adu = CreateRtpiAdu(RtpiHeader::ParseType(buf->getCont()));
  adu->Parse(buf);
  return adu;
}


RTPI::RtpiAdu* RTPI::RtpiAdu::CreateRtpiAdu(int type) {
  switch (type) {
  case RtpiHeader::PKT_EVENT:
    return (new RtpiEvent());
  case RtpiHeader::PKT_CUE:
    return (new RtpiCue());
  case RtpiHeader::PKT_STATE:
    return (new RtpiState());
  case RtpiHeader::PKT_STATE_QUERY:
    return (new RtpiStateQuery());
  case RtpiHeader::PKT_DELTA_STATE:
    return (new RtpiDeltaState());
  }
  return NULL;
}


int RTPI::RtpiAdu::Parse(RefCount<Buffer> _buf) {
  // clean up
  if (packetHeader) delete packetHeader;
  if (aduHeader) delete aduHeader;

  buf = _buf;
  char* pos = buf->getCont();

  if (RtpiHeader::ParseSize(pos) > buf->getSize())
    throw RtpiException(RtpiException::ERR_RTPI_PARSE_ADU, "RtpiAdu::Parse - given buffer too small to contain the whole ADU");

  packetHeader = new RtpiHeader();
  packetHeader->Parse(pos);
  aduHeader = CreateAduHeader(packetHeader->GetType());
  aduHeader->Parse(pos + packetHeader->GetSize());

  packetHeaderStart = pos;
  aduHeaderStart    = pos + packetHeader->GetSize();

  if (packetHeader->GetType() != RtpiHeader::PKT_STATE_QUERY) {
    payload         = aduHeaderStart + aduHeader->GetSize();
    payloadSize     = packetHeader->length - aduHeader->GetSize();
  }

  return GetPacketSize();
}


int RTPI::RtpiAdu::Fragment(char* pos, int start, int bytes) {
  if (aduHeader->GetType() == RtpiHeader::PKT_STATE_QUERY)
    // only Header
    return aduHeader->Flush(pos);

  if (start == 0) {
    // write header first
    int bytesFlushed = aduHeader->Flush(pos);
    pos += bytesFlushed;

    if (bytes - bytesFlushed> 0)
      // write payload
      memcpy(pos, payload, bytes - bytesFlushed);
  } else {
    // write only payload
    int offset = start - aduHeader->GetSize();
    memcpy(pos, payload + offset, bytes);
  }

  return bytes;
}


int RTPI::RtpiAdu::ParseHeader(char* pos, int type) {
  if (aduHeader != NULL) delete aduHeader;

  aduHeader = CreateAduHeader(type);
  return aduHeader->Parse(pos);
}


RTPI::RtpiAduHeader* RTPI::RtpiAdu::CreateAduHeader(int type) {
  switch (type) {
  case RtpiHeader::PKT_EVENT:
    return (new RtpiEventHeader());
  case RtpiHeader::PKT_CUE:
    return (new RtpiCueHeader());
  case RtpiHeader::PKT_STATE:
    return (new RtpiStateHeader());
  case RtpiHeader::PKT_STATE_QUERY:
    return (new RtpiStateQueryHeader());
  case RtpiHeader::PKT_DELTA_STATE:
    return (new RtpiDeltaStateHeader());
  }
  return NULL;
}


void RTPI::RtpiAdu::Print() {
  std::cout << "Packet: ";
  packetHeader->Print();
  std::cout << "ADU:";
  aduHeader->Print();
  if (aduHeader->GetType() != RtpiHeader::PKT_STATE_QUERY)
    std::cout << ", payloadStart=" << (int)payload << ", ADULength (incl. ADUHeader)=" << (int)payloadSize << ")" << std::endl;
  else
    std::cout << ")" << std::endl;
}


//
// class RtpiEvent
//
RTPI::RtpiAdu* RTPI::RtpiEvent::Copy() {
  RtpiEvent* ret = new RtpiEvent();
  ret->packetHeader = packetHeader->Copy();
  ret->aduHeader    = aduHeader->Copy();
  ret->payload      = payload;
  ret->payloadSize  = payloadSize;
  return ret;
}


//
// class RtpiCue
//
RTPI::RtpiAdu* RTPI::RtpiCue::Copy() {
  RtpiCue* ret = new RtpiCue();
  ret->packetHeader = packetHeader->Copy();
  ret->aduHeader    = aduHeader->Copy();
  ret->payload      = payload;
  ret->payloadSize  = payloadSize;
  return ret;
}


//
// class RtpiState
//
RTPI::RtpiAdu* RTPI::RtpiState::Copy() {
  RtpiState* ret = new RtpiState();
  ret->packetHeader = packetHeader->Copy();
  ret->aduHeader    = aduHeader->Copy();
  ret->payload      = payload;
  ret->payloadSize  = payloadSize;
  return ret;
}


//
// class RtpiDeltaState
//
RTPI::RtpiAdu* RTPI::RtpiDeltaState::Copy() {
  RtpiDeltaState* ret = new RtpiDeltaState();
  ret->packetHeader = packetHeader->Copy();
  ret->aduHeader    = aduHeader->Copy();
  ret->payload      = payload;
  ret->payloadSize  = payloadSize;
  return ret;
}


//
// class RtpiStateQuery
//
RTPI::RtpiAdu* RTPI::RtpiStateQuery::Copy() {
  RtpiStateQuery* ret = new RtpiStateQuery();
  ret->packetHeader = packetHeader->Copy();
  ret->aduHeader    = aduHeader->Copy();
  return ret;
}



//
// Functions from class RtpiCompoundPacket
//
void RTPI::RtpiCompoundPacket::AddPacket(RefCount<RtpiPacket> packet) {
  // compute overall size
  int size = 0;
  for (RtpiPacketVector::iterator i = packets.begin(); i != packets.end(); ++i)
    size += (*i)->GetSize();

  if (size > RtpiPacket::MAX_PKT_SIZE)
    throw RtpiException(RtpiException::ERR_RTPI_PACKET_FORMAT, "RtpiCompoundPacket::AddPacket - maximum packet size exceeded");

  packets.push_back(packet);
}


void RTPI::RtpiCompoundPacket::Print() {
  if (packets.size() == 0)
    std::cout << "RTPI/I compound packet empty" << std::endl;
  else {
    std::cout << "RTP/I compound packet - packets: " << packets.size() << " buffer size: " << buf->getSize() << std::endl;
    for (RtpiPacketVector::iterator i = packets.begin(); i != packets.end(); ++i)
      (*i)->Print();
  }
}


/*

//
// Functions from class RtpiAduFactory
//
RTPI::RtpiAduFactory::~RtpiAduFactory() {
  for (list<list<RefCount<RtpiPacket> >*>::iterator i = fragments.begin(); i != fragments.end(); ++i) {
    delete *i;
  }
}


// fragment a RTP/I ADU into one or more RTP/I packets; these packtes are flushed into the buffer given
// by the appliction; ! if this buffer is deleted, all RTP/I packets become obsolete and  must be destroyed
// if no header extension shall be attached, the packets are completely flushed;
//   however, the packet header can be changed and the packet flushed again (the payload stays in its assigned place
// if a header extension shall be attached, this has to be done manually by calling RtpiPacket::SetHeaderExtension for
//   each packet of the vector returend; afterwards all packets have to be flushed manually (call RtpiPacket::Flush)
//   the payload is already in it's correct position
RTPI::RtpiPacketVector* RTPI::RtpiAduFactory::CreateRtpiPackets(RtpiAdu* adu) {

  int aduSize = adu->GetAduSize();
  int reservedPacketSize = RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize();
  int maxRtpiPayloadSize = RtpiPacket::MAX_PKT_SIZE - reservedPacketSize;
  int rtpiPacketsNeeded  = (int)ceil((double)aduSize / maxRtpiPayloadSize);

  RtpiPacketVector* packets = new RtpiPacketVector();
  RefCount<RtpiPacket> packet;
  RtpiHeader* hdr;

  int totalAduBytesFlushed = 0;
  int bytesFlushed = 0;
  int fc = 0;

  while (totalAduBytesFlushed < aduSize) {
    // reserve a Buffer for each packet
    RefCount<Buffer> buf = new Buffer(RtpiPacket::MAX_PKT_SIZE);
    char* pos = buf->getCont();

    // create new RTP/I packet
    packet = new RtpiPacket(buf, pos);
    packets->push_back(packet);

    pos += (RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize());

    // write payload of RTP/I packet  = piece of RTP/I adu
    int bytesLeft = aduSize - totalAduBytesFlushed;
    int bytesToFlush = (bytesLeft <= maxRtpiPayloadSize) ? bytesLeft : maxRtpiPayloadSize;
    adu->Fragment(pos, totalAduBytesFlushed, bytesToFlush);

    // write RTP/I packet header
    packet->SetPayload(buf, pos, bytesToFlush);
    hdr = new RtpiHeader();
    hdr->e = (fc == rtpiPacketsNeeded - 1) ? 1 : 0;
    hdr->type = adu->GetType();
    hdr->rt = adu->packetHeader->rt;
    hdr->pid = adu->packetHeader->pid;
    hdr->subid = adu->packetHeader->subid;
    hdr->sn = adu->packetHeader->sn;
    hdr->fc = fc;
    packet->SetHeader(hdr);
    // if no header extension the packet is flushed completely; if extension, the packet has to be flushed again after
    // the extension has been attached
    bytesFlushed = packet->Flush();

    // resize buffer to number of bytes written; can be smaller than value above
    buf->setSize(bytesFlushed);

    // prepare for next round
    totalAduBytesFlushed += bytesToFlush;

    fc++;
  }

  return packets;
}


// fragment a RTP/I ADU into one or more RTP/I packets; these packtes are flushed into the buffer given
// by the appliction; ! if this buffer is deleted, all RTP/I packets become obsolete and  must be destroyed
// if no header extension shall be attached, the packets are completely flushed;
//   however, the packet header can be changed and the packet flushed again (the payload stays in its assigned place
// if a header extension shall be attached, this has to be done manually by calling RtpiPacket::SetHeaderExtension for
//   each packet of the vector returend; afterwards all packets have to be flushed manually (call RtpiPacket::Flush)
//   the payload is already in it's correct position
RTPI::RtpiPacketVector* RTPI::RtpiAduFactory::CreateRtpiPackets(RtpiAdu* adu, RefCount<Buffer> buf) {

  int aduSize = adu->GetAduSize();
  int reservedPacketSize = RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize() + RtpiPacket::GetPaddingBorder();
  // juergi: wozu ist hier GetPaddingBorder noetig? Stimmt das Ergebnis?
  int maxRtpiPayloadSize = RtpiPacket::MAX_PKT_SIZE - reservedPacketSize;
  int rtpiPacketsNeeded  = (int)ceil((double)aduSize / maxRtpiPayloadSize);

  // resize buffer
  buf->resize(aduSize + rtpiPacketsNeeded * reservedPacketSize);

  RtpiPacketVector* packets = new RtpiPacketVector();
  RefCount<RtpiPacket> packet;
  RtpiHeader* hdr;

  char* pos = buf->getCont();
  int totalBytesFlushed = 0;
  int totalAduBytesFlushed = 0;
  int bytesFlushed = 0;
  int fc = 0;

  while (totalAduBytesFlushed < aduSize) {
    // create new RTP/I packet
    packet = new RtpiPacket(buf, pos);
    packets->push_back(packet);

    pos += (RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize());

    // write payload of RTP/I packet  = piece of RTP/I adu
    int bytesLeft = aduSize - totalAduBytesFlushed;
    int bytesToFlush = (bytesLeft <= maxRtpiPayloadSize) ? bytesLeft : maxRtpiPayloadSize;
    adu->Fragment(pos, totalAduBytesFlushed, bytesToFlush);

    // write RTP/I packet header
    packet->SetPayload(buf, pos, bytesToFlush);
    hdr = new RtpiHeader();
    hdr->e = (fc == rtpiPacketsNeeded - 1) ? 1 : 0;
    hdr->type = adu->GetType();
    hdr->rt = adu->packetHeader->rt;
    hdr->pid = adu->packetHeader->pid;
    hdr->subid = adu->packetHeader->subid;
    hdr->sn = adu->packetHeader->sn;
    hdr->fc = fc;
    packet->SetHeader(hdr);
    // if no header extension the packet is flushed completely; if extension, the packet has to be flushed again after
    // the extension has been attached
    bytesFlushed = packet->Flush();
    totalBytesFlushed += bytesFlushed;

    // prepare for next round
    totalAduBytesFlushed += bytesToFlush;
    pos += (bytesFlushed - RtpiHeader::GetSize() - RtpiHeaderExtension::GetSize()); // payload plus padding ; rest has been taken care of above

    fc++;
  }

  // resize buffer to number of bytes written; can be smaller than value above
  buf->setSize(totalBytesFlushed);

  return packets;
}


// create RTP/I packets from a list of RTP/I ADUs: if a ADU is too large for one transport packet it is
// fragmented into several RTP/I packets; if not, only one RTP/I packet is created per ADU. If several
// RTP/I packets fit into one tranpsort packet, a compound packet is created containing a list of
// RTP/I packets which share the same buffer.
// ! if a buffer is deleted, all corresponding RTP/I packets become obsolete and  must be destroyed
// if no header extension shall be attached, the packets are completely flushed;
//   however, the packet header can be changed and the packet flushed again (the payload stays in its assigned place
// if a header extension shall be attached, this has to be done manually by calling RtpiPacket::SetHeaderExtension for
//   each packet of the vector returend; afterwards all packets have to be flushed manually (call RtpiPacket::Flush)
//   the payload is already in it's correct position
RTPI::RtpiCompoundPacketVector* RTPI::RtpiAduFactory::CreateRtpiPackets(RefCount<RtpiAduVector> adus) {
  if ((!adus) || (adus->size() == 0))
    return NULL;

  int totalSize = 0;
  int aduSize = 0;
  int reservedPacketSize = RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize() + RtpiPacket::GetPaddingBorder();
  int maxRtpiPayloadSize = RtpiPacket::MAX_PKT_SIZE - reservedPacketSize;

  RefCount<Buffer> buf = new Buffer(RtpiPacket::MAX_PKT_SIZE);
  int bytesFlushedIntoBuffer = 0;
  RtpiCompoundPacketVector* compoundPackets = new RtpiCompoundPacketVector();
  RefCount<RtpiCompoundPacket> compoundPacket = new RtpiCompoundPacket(buf);
  compoundPackets->push_back(compoundPacket);

  char* pos = buf->getCont();
  RefCount<RtpiPacket> packet;
  RtpiHeader* hdr = NULL;

  for (RtpiAduVector::iterator i = adus->begin(); i != adus->end(); ++i) {
    int totalAduBytesFlushed = 0;
    int bytesFlushed = 0;
    aduSize = (*i)->GetAduSize();
    int rtpiPacketsNeeded = (int)ceil((double)aduSize / maxRtpiPayloadSize);
    int fc = 0;

    if (rtpiPacketsNeeded == 1) {
      //
      // 1. only one RTP/I packet is needed for the complete ADU
      //
      // create new RTP/I packet
      packet = new RtpiPacket(buf, pos);
      compoundPacket->AddPacket(packet);

      pos += (RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize());

      // write payload of RTP/I packet  = complete RTP/I ADU
      (*i)->Fragment(pos, 0, aduSize);

      // write RTP/I packet header
      packet->SetPayload(buf, pos, aduSize);
      hdr = new RtpiHeader();
      hdr->e = 1;
      hdr->type = (*i)->GetType();
      hdr->rt = (*i)->packetHeader->rt;
      hdr->pid = (*i)->packetHeader->pid;
      hdr->subid = (*i)->packetHeader->subid;
      hdr->sn = (*i)->packetHeader->sn;
      hdr->fc = 0;
      packet->SetHeader(hdr);
      // if no header extension the packet is flushed completely; if extension, the packet has to be flushed again after
      // the extension has been attached
      bytesFlushed = packet->Flush();
      bytesFlushedIntoBuffer += bytesFlushed;

      if ((i + 1) != adus->end()) { // there is at least one more ADU
	// prepare for next round
	if ((buf->getSize() - bytesFlushedIntoBuffer) < ((*(i+1))->GetAduSize() + reservedPacketSize)) {
	  // next ADU does not fit into the current buffer -> create new RTP/I Compound packet and new buffer
	  buf->setSize(bytesFlushedIntoBuffer);
	  buf = new Buffer(RtpiPacket::MAX_PKT_SIZE);
	  bytesFlushedIntoBuffer = 0;
	  pos = buf->getCont();
	  compoundPacket = new RtpiCompoundPacket(buf);
	  compoundPackets->push_back(compoundPacket);
	} else {
	  // next ADU does fit into the current buffer
	  pos += (bytesFlushed - RtpiHeader::GetSize() - RtpiHeaderExtension::GetSize()); // payload plus padding ; rest has been taken care of above
	}
      } else {
	// no more ADUs
	buf->setSize(bytesFlushedIntoBuffer);
      }
    } else {
      //
      // 2. the ADU has to be fragmented into several RTP/I packets
      //
      while (totalAduBytesFlushed < aduSize) {
	// create new RTP/I packet
	packet = new RtpiPacket(buf, pos);
	compoundPacket->AddPacket(packet);

	pos += (RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize());

	// write payload of RTP/I packet  = piece of RTP/I adu
	int bytesLeft = aduSize - totalAduBytesFlushed;
	int bytesToFlush = (bytesLeft <= maxRtpiPayloadSize) ? bytesLeft : maxRtpiPayloadSize;
	(*i)->Fragment(pos, totalAduBytesFlushed, bytesToFlush);

	// write RTP/I packet header
	packet->SetPayload(buf, pos,bytesToFlush);
	hdr = new RtpiHeader();
	hdr->e = (fc == rtpiPacketsNeeded - 1) ? 1 : 0;
	hdr->type = (*i)->GetType();
	hdr->rt = (*i)->packetHeader->rt;
	hdr->pid = (*i)->packetHeader->pid;
	hdr->subid = (*i)->packetHeader->subid;
	hdr->sn = (*i)->packetHeader->sn;
	hdr->fc = fc;
	packet->SetHeader(hdr);
	// if no header extension the packet is flushed completely; if extension, the packet has to be flushed again after
	// the extension has been attached
	bytesFlushed = packet->Flush();
	buf->setSize(bytesFlushed);  // resize buffer to number of bytes written; can be smaller than value above

	// prepare for next round
	totalAduBytesFlushed += bytesToFlush;

	fc++;

	// create new compound packet and new buffer
	buf = new Buffer(RtpiPacket::MAX_PKT_SIZE);
	pos = buf->getCont();
	compoundPacket = new RtpiCompoundPacket(buf);
	compoundPackets->push_back(compoundPacket);
      }

      if ((i + 1) == adus->end()) { // there are no more ADUs
	// delete last compound packet
	compoundPackets->pop_back();
      }
    }
  }

  return compoundPackets;
}


// create RTP/I ADUs from RTP/I packets stored in buf
RTPI::RtpiAduVector* RTPI::RtpiAduFactory::CreateRtpiAdusFromPackets(RefCount<Buffer> buf) {
  RtpiPacketFactory factory;
  RtpiPacketVector* rtpiPackets = factory.CreateRtpiPackets(buf);

  if (rtpiPackets == NULL)
    throw RtpiException(RtpiException::ERR_RTPI_PARSE, "RtpiAduFactory::CreateRtpiAdusFromPackets - Buffer corrupt or empty");

  //std::cout << "RtpiAduFactory: Created the following packets: " << std::endl;
  //for (RtpiPacketVector::iterator ppos = rtpiPackets->begin(); ppos < rtpiPackets->end(); ++ppos)
  //  (*ppos)->Print();

  RtpiAduVector* rtpiAdus = CreateRtpiAdusFromPackets(rtpiPackets);

  // clean up
  delete rtpiPackets; // only vector is destroyed - the packets live until the corresponding adu is complete

  return rtpiAdus;
}


// create RTP/I ADUs from RTP/I packets (=ADU fragments); returns all ADUs that have been completed
RTPI::RtpiAduVector* RTPI::RtpiAduFactory::CreateRtpiAdusFromPackets(RtpiPacketVector* packets) {
  //std::cout << "RtpiAduFactory::CreateRtpiAdusFromPackets" << std::endl;

  if (packets->size() == 0) return NULL; // nothing to do

  RtpiAduVector* aduList = new RtpiAduVector;
  RefCount<RtpiPacket> packet;
  RtpiAdu* adu;

  for (RtpiPacketVector::iterator i = packets->begin(); i != packets->end(); ++i) {
    packet = *i;

    if ((packet->hdr->e == 1) && (packet->hdr->fc == 0)) {
      // packet contains one complete ADU -> create ADU
      adu = CreateRtpiAdu(packet->hdr->type);

      // create new buffer which contains RTP/I packet header  + RTP/I Adu Header + Payload
      adu->buf = new Buffer(packet->hdr->GetSize() + packet->GetPayloadSize());
      adu->packetHeaderStart = adu->buf->getCont(); // ! extension

      // create Adu Header
      adu->ParseHeader(packet->GetPayloadStart(), packet->hdr->type); // parse RTP/I ADU header
      adu->aduHeader->Flush(adu->packetHeaderStart + packet->hdr->GetSize()); // ! extension

      // set beginning of ADU Header in Buffer
      adu->aduHeaderStart = adu->packetHeaderStart + packet->hdr->GetSize(); // ! extension

      // set beginning of payload in Buffer
      if (adu->GetType() != RtpiHeader::PKT_STATE_QUERY) {
	adu->payload = adu->aduHeaderStart + adu->aduHeader->GetSize();
	adu->payloadSize = packet->GetPayloadSize() - adu->aduHeader->GetSize();
      }

      // copy RTP/I ADU Header and payload (if exists)
      memcpy(adu->aduHeaderStart, packet->GetPayloadStart(), packet->GetPayloadSize());

      // create single RTP/I packet header at beginning of RTP/I ADU
      adu->packetHeader = new RtpiHeader();
      adu->packetHeader->v = packet->hdr->v;
      adu->packetHeader->e = 1;
      adu->packetHeader->x = packet->hdr->x; // ! extension
      adu->packetHeader->type = packet->hdr->type;
      adu->packetHeader->p = 0;
      adu->packetHeader->rt = packet->hdr->rt;
      adu->packetHeader->length = adu->aduHeader->GetSize() + adu->payloadSize;
      adu->packetHeader->pid = packet->hdr->pid;
      adu->packetHeader->subid = packet->hdr->subid;
      adu->packetHeader->sn = packet->hdr->sn;
      adu->packetHeader->fc = 0;
      adu->packetHeader->Flush(adu->packetHeaderStart);

     // create RTP/I packet header extension ?

      aduList->push_back(adu);

      continue;
    }

    // packet contains one fragment of the ADU
    list<RefCount<RtpiPacket> >* fragmentList = FindFragmentList(packet->hdr->subid, packet->hdr->type, packet->hdr->pid, packet->hdr->sn);
    if (fragmentList == NULL) {
      // first fragment of ADU -> generate new list
      fragmentList = InsertFragmentList(packet->hdr->subid, packet->hdr->type, packet->hdr->pid, packet->hdr->sn);
      fragmentList->push_front(packet);
    } else {
      // some fragments already there -> insert new fragment
      list<RefCount<RtpiPacket> >::iterator pos = fragmentList->begin();
      while (**pos < *packet)
	if (++pos == fragmentList->end()) break;

      fragmentList->insert(pos,packet);

      // check if ADU complete
      if (AduComplete(fragmentList)) {
	// create ADU

	list<RefCount<RtpiPacket> >::iterator i = fragmentList->begin();
	RefCount<RtpiPacket> first = *i;
	adu = CreateRtpiAdu(first->hdr->type);
	int aduSize = GetSizeOfAdu(fragmentList);

	// create new buffer which contains RTP/I packet header  + RTP/I Adu Header + Payload
	adu->buf = new Buffer(first->hdr->GetSize() + aduSize); // ! extension
	adu->packetHeaderStart = adu->buf->getCont(); // ! extension

	// create Adu Header
	adu->ParseHeader(first->GetPayloadStart(), first->hdr->type); // parse RTP/I ADU header
	adu->aduHeader->Flush(adu->packetHeaderStart + first->hdr->GetSize()); // ! extension

	// set beginning of ADU Header in Buffer
	adu->aduHeaderStart = adu->packetHeaderStart + first->hdr->GetSize(); // ! extension

	// set beginning of payload in Buffer
	adu->payload = adu->aduHeaderStart + adu->aduHeader->GetSize();
	adu->payloadSize = aduSize - adu->aduHeader->GetSize();

	// create single RTP/I packet header at beginning of RTP/I ADU
	adu->packetHeader = new RtpiHeader();
	adu->packetHeader->v = first->hdr->v;
	adu->packetHeader->e = 1;
	adu->packetHeader->x = first->hdr->x; // ! extension
	adu->packetHeader->type = first->hdr->type;
	adu->packetHeader->p = 0;
	adu->packetHeader->rt = first->hdr->rt;
	adu->packetHeader->length = aduSize;
	adu->packetHeader->pid = first->hdr->pid;
	adu->packetHeader->subid = first->hdr->subid;
	adu->packetHeader->sn = first->hdr->sn;
	adu->packetHeader->fc = 0;
	adu->packetHeader->Flush(adu->packetHeaderStart);

	// create RTP/I packet header extension ?

	// copy RTP/I ADU Header and payload
	char* p = adu->aduHeaderStart;
	int bytesToCopy = first->GetPayloadSize();
	memcpy(p, first->GetPayloadStart(), bytesToCopy);
	p += bytesToCopy;

	++i;
	for ( ; i != fragmentList->end(); ++i) {
	  bytesToCopy = (*i)->GetPayloadSize();
	  memcpy(p, (*i)->GetPayloadStart(), bytesToCopy);
	  p += bytesToCopy;
	}

	aduList->push_back(adu);

	// clean up
	DeleteFragmentList(packet->hdr->subid, packet->hdr->type, packet->hdr->pid, packet->hdr->sn);

      }
    }
  }

  if (aduList->size() == 0) {
    // no comlete ADU's
    delete aduList;
    return NULL;
  } else
    return aduList;
}


RTPI::RtpiAdu* RTPI::RtpiAduFactory::CreateRtpiAduFromAdu(RefCount<Buffer> buf) {
  // create one  RTP/I ADU from buf  ( RTP/I packet header + ADU header + payload)
  RtpiAdu* adu = CreateRtpiAdu(RtpiHeader::ParseType(buf->getCont()));
  adu->Parse(buf);
  return adu;
}


RTPI::RtpiAdu* RTPI::RtpiAduFactory::CreateRtpiAdu(int type) {
  switch (type) {
  case RtpiHeader::PKT_EVENT:
    return (new RtpiEvent());
  case RtpiHeader::PKT_CUE:
    return (new RtpiCue());
  case RtpiHeader::PKT_STATE:
    return (new RtpiState());
  case RtpiHeader::PKT_STATE_QUERY:
    return (new RtpiStateQuery());
  case RtpiHeader::PKT_DELTA_STATE:
    return (new RtpiDeltaState());
  }
  return NULL;
}


bool RTPI::RtpiAduFactory::AduComplete(list<RefCount<RtpiPacket> >* fragmentList) {
  int i = 0;
  list<RefCount<RtpiPacket> >::iterator p;

  for (p = fragmentList->begin(); p != fragmentList->end(); ++p) {
    if ((*p)->hdr->fc != i) return false; // gap discovered
    ++i;
  }

  p = fragmentList->end();
  --p;
  if ((*p)->hdr->e != 1) return false; // tail not there

  return true;
}


int RTPI::RtpiAduFactory::GetSizeOfAdu(list<RefCount<RtpiPacket> >* fragmentList) {
  int size = 0;

  for (list<RefCount<RtpiPacket> >::iterator i = fragmentList->begin(); i != fragmentList->end(); ++i)
    size += (*i)->GetPayloadSize();

  return size;
}


void RTPI::RtpiAduFactory::Print(list<RefCount<RtpiPacket> >* fragmentList) {
  for (list<RefCount<RtpiPacket> >::iterator i = fragmentList->begin(); i != fragmentList->end(); ++i)
    (*i)->Print();
}


list<RefCount<RTPI::RtpiPacket> >* RTPI::RtpiAduFactory::FindFragmentList(uint64 subid, uint8 type, uint32 pid, uint16 sn) {
  RefCount<RtpiPacket> first;
  for (list<list<RefCount<RtpiPacket> >*>::iterator i = fragments.begin(); i != fragments.end(); ++i) {
    first = (*i)->front();
    if ((first->hdr->subid == subid) && (first->hdr->type == type) && (first->hdr->pid == pid) && (first->hdr->sn == sn))
      return *i;
  }

  return NULL; // not found
}


list<RefCount<RTPI::RtpiPacket> >* RTPI::RtpiAduFactory::InsertFragmentList(uint64 subid, uint8 type, uint32 pid, uint16 sn) {
  list<RefCount<RtpiPacket> >* fragmentList = new list<RefCount<RtpiPacket> >;
  fragments.push_back(fragmentList);
  return fragmentList;
}


void RTPI::RtpiAduFactory::DeleteFragmentList(uint64 subid, uint8 type, uint32 pid, uint16 sn) {
  RefCount<RtpiPacket> first;
  for (list<list<RefCount<RtpiPacket> >*>::iterator i = fragments.begin(); i != fragments.end(); ++i) {
    first = (*i)->front();
    if ((first->hdr->subid == subid) && (first->hdr->type == type) && (first->hdr->pid == pid) && (first->hdr->sn == sn)) {
      delete *i;
      fragments.erase(i);
      return;
    }
  }
}


*/



