// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//
// Copyright (c) 2000 University of Mannheim. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for educational, research and non-profit purposes,
// without fee, and without a written agreement is hereby granted,
// provided that the above copyright notice, this paragraph and the
// following four paragraphs appear in all copies.
//
// Permission to incorporate this software into commercial products may
// be obtained from the following address:
// University of Mannheim
// Lehrstuhl fuer Praktische Informatik IV
// L15, 16
// 68161 Mannheim, Germany
//
// IN NO EVENT SHALL THE UNIVERSITY OF MANNHEIM BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
// ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF MANNHEIM HAS BEEN ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//
// THE UNIVERSITY OF MANNHEIM SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER
// IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF MANNHEIM HAS NO OBLIGATIONS
// TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The University of Mannheim will appreciate receiving any enhancements and
// modifications of this software made by others, and reserves the right to
// include them in future releases. The University also recommends that this
// software not be redistributed by third parties, but be obtained by all
// directly from the University's server, so that all users of the software
// may receive bug notifications and other relevant announcements.
//
// Authors:
//   Volker Hilt, Juergen Vogel, Martin Mauve
//   University of Mannheim
//   {hilt,vogel,mauve}@informatik.uni-mannheim.de
//
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/**********************************************************************

 program : rtpiPacket.h
 authors : Juergen Vogel
 date    : 05.04.2000
 desc    : defines RTP/I Packets

 **********************************************************************/


#ifndef __RTPI_PACKET__
#define __RTPI_PACKET__

#ifdef WIN32
#  pragma warning (disable: 4786 4181)
#endif


#define RTPI_MASK_V 0xC0
#define RTPI_MASK_E 0x20
#define RTPI_MASK_X 0x10
#define RTPI_MASK_TYPE 0x0F
#define RTPI_MASK_P 0x80
#define RTPI_MASK_RT 0x7F
#define RTPI_MASK_O 0x80
#define RTPI_MASK_PRI 0x03

#ifdef WIN32
#include <util/refcount.cpp>
#else
#include <util/refcount.cc>
#endif


#include <list>
#include <vector>
#include <memory.h>
#include <iostream>

#include <util/types.h>
#include <util/Buffer.h>
#include <rtpi/rtpiException.h>
//#include <rtpi/rtpihandler.h>


namespace RTPI {

  class RtpiHeader {
  public:
    enum types {
      // RTP/I Version
      VERSION = 0,
      // type of ADU
      PKT_EVENT = 0,
      PKT_STATE = 1,
      PKT_DELTA_STATE = 2,
      PKT_STATE_QUERY = 3,
      PKT_CUE = 4
    };

    RtpiHeader()
      : v(VERSION), e(1), x(0), type(PKT_EVENT), p(0), rt(0), length(0), pid(0), subid(), sn(0), fc(0) {}
    RtpiHeader(uint32 _pid, uint32 _subid, uint8 _type, uint16 _sn, uint16 _fc, bool end)
      : v(VERSION), e(end), x(0), type(_type), p(0), rt(0), length(0), pid(_pid), subid(_subid), sn(_sn), fc(_fc) {
      if ((type != PKT_EVENT) && (type != PKT_CUE) && (type != PKT_STATE) && (type != PKT_DELTA_STATE) && (type != PKT_STATE_QUERY))
	throw RtpiException(RtpiException::ERR_RTPI_PACKET_FORMAT, "RtpiHeader::RtpiHeader - wrong type");
    }
    virtual ~RtpiHeader() { };

    virtual RtpiHeader* Copy();

    uint8 GetType() {return type;}
    void SetType(int _type) {
      if ((_type != PKT_EVENT) && (_type != PKT_CUE) && (_type != PKT_STATE) && (_type != PKT_DELTA_STATE) && (_type != PKT_STATE_QUERY))
	throw RtpiException(RtpiException::ERR_RTPI_PACKET_FORMAT, "RtpiHeader::SetType - wrong type");
      type = _type;
    }

    int GetReliabilityType() {return rt;}
    void SetReliabilityType(uint8 _rt) {
      rt = _rt;
    }

    void SetSubcomponentID(uint32 _subid) {subid = _subid;}
    uint32 GetSubcomponentID() {return subid;}

    void SetParticipantID(uint32 _pid) {pid = _pid;}
    uint32 GetParticipantID() {return pid;}

    int GetSequenceNumber() {return sn;}
    void SetSequenceNumber(uint16 _sn) {
      sn = _sn;
    }

    int GetFragmentCount() {return fc;}
    void SetFragmentCount(uint16 _fc) {
      fc = _fc;
    }

    virtual int Parse(char* pos);
    virtual int Flush(char* pos);

    virtual void Print();

    static int GetSize() { return (4 * sizeof(uint32)); } // get size of RTP/I Header in bytes

    static int ParseType(char* pos);   // return the TYPE flag
    static int ParseSize(char* pos);   // return the size of rtpi packet (header + payload)

    // private:  // should be ;)
    friend class TransparentReliableTransport;
    friend class UnreliableTransport;
    friend class RtpiAdu;
    friend class RtpiHeaderExtension;
    friend class RtpiPacket;
    friend class rtpihandler;
    friend class rtpiparticipant;

    // Header fields
    uint8  v;      // 2 b:  protocol version
    uint8  e;      // 1 b:  end bit
    uint8  x;      // 1 b:  reliability header extension
    uint8  type;   // 4 b:  ADU type (event, state, delta state, state query)
    uint8  p;      // 1 b:  padding bit
    uint8  rt;     // 7 b:  reliability type
    uint16 length; // 16 b: length of payload in bytes
    uint32 pid;    // 32 b: participant identifier
    uint32 subid;  // 32 b: subcomponent id
    uint16 sn;     // 16 b: sequence number
    uint16 fc;     // 16 b: fragment count
  };

  class RtpiHeaderExtension {
  public:
    RtpiHeaderExtension() : data(NULL) {}
    RtpiHeaderExtension(char* _data) {
      if (size == 0)
	throw RtpiException(RtpiException::ERR_RTPI_INTERNAL, "RtpiHeaderExtension::RtpiHeaderExtension - size of extension must be set first");
      data = new char[size-1];
      memcpy(data, _data, size-1);
    }
    ~RtpiHeaderExtension() { if (data != NULL) delete[] data; }

    RtpiHeaderExtension* Copy() {return new RtpiHeaderExtension(data);}

    void SetData(char* _data) {
      if (data != NULL) delete[] data;
      data = new char[size-1];
      memcpy(data,_data,size-1);
    }
    char* GetData() { return data;}
    int GetDataLength() {return size - 1;}

    int Parse(char* pos);
    int Flush(char* pos);

    void Print() {std::cout << "Header Extension - size " << size << " Data " << data << std::endl;}

    static void SetSize(int _size) {
      if ((size % 4) != 0)
	throw RtpiException(RtpiException::ERR_RTPI_INTERNAL, "RtpiHeaderExtension::SetSize - size must be multiple of 4");
      size = _size;
    }
    static int GetSize() {return size;}

  private:
    char* data;   // x bytes: reliabiltiy header extension data
    static int size; // size of complete header extension in bytes; if 0 (default), no extension attached to packets
  };


  class RtpiPacket {
  public:
    enum types {
      MAX_PKT_SIZE         = 1400,  // 1400 this value has to be requested from the transport layer!
    };

    friend std::ostream& operator<<(std::ostream& os, const RtpiPacket &p);

    RtpiPacket()
      : hdr(NULL), hdrX(NULL), buf(RefCount<Buffer>()), headerStart(NULL), payloadStart(NULL), payloadSize(0) {}
    RtpiPacket(RefCount<Buffer> _buf, char* pos)
      : hdr(NULL), hdrX(NULL), buf(_buf), headerStart(pos), payloadStart(NULL), payloadSize(0) {}

    RtpiPacket(RtpiHeader* _hdr)
      :hdr(_hdr), hdrX(NULL),  buf(RefCount<Buffer>()), headerStart(NULL), payloadStart(NULL), payloadSize(0) {}

    RtpiPacket(RtpiHeader* _hdr, RefCount<Buffer> _buf, char* _payloadStart, int _payloadSize)
      : hdr(_hdr), hdrX(NULL), buf(_buf), headerStart(NULL), payloadStart(_payloadStart), payloadSize(_payloadSize) {}

    RtpiPacket(RtpiHeader* _hdr, RtpiHeaderExtension* _hdrX, RefCount<Buffer> _buf, char* _payloadStart, int _payloadSize)
      : hdr(_hdr), hdrX(_hdrX), buf(_buf), headerStart(NULL), payloadStart(_payloadStart), payloadSize(_payloadSize) {}

    ~RtpiPacket() { if (hdr) delete hdr; if (hdrX) delete hdrX;}

    // function needed to manage adu fragments (=packets) in class RtpiAduFactory
    bool operator<(const RtpiPacket& packet) {return hdr->fc < packet.hdr->fc;}

    int Parse(RefCount<Buffer> _buf, char* pos); // create RTP/I packet from bitstring
    int Flush(); // create bitstring from RTP/I packet

    int GetSize();        // returns packet size in bytes including padding
    int GetHeaderSize();  // returns size of complete Header in bytes

    void SetHeader(RtpiHeader* _hdr) {
      if (hdr != NULL)	delete hdr;
      hdr = _hdr;
    }
    RtpiHeader* GetHeader() { return hdr; }
    char* GetStart() {return headerStart;}

    void SetHeaderExtension(RtpiHeaderExtension* _hdrX) {
      if (hdrX != NULL) delete hdrX;
      hdrX = _hdrX;
      hdr->x = 1;
    }
    void ClearHeaderExtension() {
      if (hdrX) delete hdrX;
      hdr->x = 0;
    }
    RtpiHeaderExtension* GetHeaderExtension() {return hdrX;}
    bool HeaderExtension() { if (hdrX && hdr->x) return true; return false;}

    void SetPayload(RefCount<Buffer> _buf, char* payload, int size) {
      if ((buf != RefCount<Buffer>()) && (buf != _buf))
	throw RtpiException(RtpiException::ERR_RTPI_INTERNAL, "RtpiPacket::SetPayload - Illegal Buffer change!");
      buf = _buf;
      payloadStart = payload; payloadSize = size;
    }
    void SetPayloadSize(int size) {payloadSize = size;}
    char* GetPayloadStart() {return payloadStart;}
    int GetPayloadSize() {return payloadSize;}

    RefCount<Buffer> getBuffer() { return buf; }

    static void SetPaddingBorder(int border);  // set nbr of bytes to which the packet should be aligned; default is 4
    static int GetPaddingBorder() {return paddingBorder;}

    void Print();

  private:
    friend class TransparentReliableTransport;
    friend class UnreliableTransport;

    // juergi: buf added to RtpiPacket- please check if ok
    RtpiHeader* hdr;            // Standard RTP/I Header
    RtpiHeaderExtension* hdrX;  // optional Header Extension
    RefCount<Buffer> buf;       // buffer containing the packet
    char* headerStart;          // pointer to addr in buffer where the RtpiPacket (Header) starts
    char* payloadStart;         // pointer to payload (points at buffer addr)
    int payloadSize;            // payload size in bytes

    static int paddingBorder;   // packet is aligned to paddinBorder bytes
  };

  typedef std::vector<RefCount<RtpiPacket> > RtpiPacketVector;


  class RtpiPacketFactory {
  public:
    static RtpiPacketVector* CreateRtpiPackets(RefCount<Buffer> buf);
  private:
    static RefCount<RtpiPacket> CreateRtpiPacket(RefCount<Buffer> _buf, char* pos, int& bytesRead);
  };


  class RtpiAduHeader {
  public:
    RtpiAduHeader()
      : pt(0), pi(0), ts(0) {}
    RtpiAduHeader(uint8 _pt, uint8 _pi, uint32 _ts)
      : pt(_pt), pi(_pi), ts(_ts) {}

    virtual RtpiAduHeader* Copy() = 0;

    virtual uint8 GetType() = 0;
    virtual int Parse(char* pos) = 0;
    virtual int Flush(char* pos) = 0;
    virtual int GetSize() = 0;

    int GetPayloadType() {return pt;}
    int GetProfileInformation() {return pi;}

    uint32 GetTimestamp() {return ts;}
    void SetTimestamp(uint32 _ts) {ts = _ts;}

    virtual void Print() = 0;

    // protected: (should be ;) )
    friend class TransparentReliableTransport;
    friend class UnreliableTransport;

    uint8  pt;     // 8 b:  payload type
    uint8  pi;     // 8 b:  profile information
    uint32 ts;     // 32 b: timestamp
  };


  class RtpiEventHeader : public RtpiAduHeader {
  public:
    RtpiEventHeader()
      : RtpiAduHeader(), o(0), objectSubid(0) {}
    RtpiEventHeader(uint8 pt, uint8 pi, uint32 ts)
      : RtpiAduHeader(pt,pi,ts), o(0), objectSubid(0) {}
    RtpiEventHeader(uint8 pt, uint8 pi, uint32 ts, uint32 _objectSubid)
      : RtpiAduHeader(pt,pi,ts), o(1), objectSubid(_objectSubid) {}
    ~RtpiEventHeader() {}

    RtpiAduHeader* Copy();

    uint8 GetType() {return RtpiHeader::PKT_EVENT;}

    bool ObjectIDIsSet() { return o; }
    void SetObjectID(uint32 subid) {objectSubid = subid; o = 1;}
    uint32 GetObjectID() {return objectSubid;}

    int Parse(char* pos);
    int Flush(char* pos);

    int GetSize() {return (o ? 3 * sizeof(uint32) : 2 * sizeof(uint32));}

    void Print();

    // protected:
    uint8  o;            // 1 b:  if 1, header includes the ID of another subcomponent to express
                         //       an object relationship
    uint32 objectSubid;  // 32 b: object subcomponent ID
  };


  class RtpiCueHeader : public RtpiAduHeader {
  public:
    RtpiCueHeader()
      : RtpiAduHeader() {}
    RtpiCueHeader(uint8 pt, uint8 pi, uint32 ts)
      : RtpiAduHeader(pt,pi,ts) {}
    ~RtpiCueHeader() {}

    RtpiAduHeader* Copy();

    uint8 GetType() {return RtpiHeader::PKT_CUE;}

    int Parse(char* pos);
    int Flush(char* pos);

    int GetSize() {return 2 * sizeof(uint32);}

    void Print();

    // 16 bits reserved
  };


  class RtpiStateQueryHeader : public RtpiAduHeader {
  public:
    RtpiStateQueryHeader()
      : RtpiAduHeader(), pri(0) {}
    RtpiStateQueryHeader(uint8 pt, uint8 pi, uint32 ts, uint8 _pri)
      : RtpiAduHeader(pt,pi,ts), pri(_pri) {}
    ~RtpiStateQueryHeader() {}

    virtual RtpiAduHeader* Copy();

    uint8 GetType() {return RtpiHeader::PKT_STATE_QUERY;}

    int  GetPriority() {return pri;}
    void SetPriority(int priority) {pri = priority;}

    int Parse(char* pos);
    int Flush(char* pos);

    int GetSize() {return 2*sizeof(uint32);}

    void Print();

  // private:   (should be ;) )
    // reserved 14 b
    uint8  pri;    // 2 b:  priority
  };


  class RtpiStateBaseHeader : public RtpiAduHeader {
  public:
    RtpiStateBaseHeader()
      : RtpiAduHeader(), o(0), snc(0), pri(0), pids(), sns(), objectSubid(0) {}
    RtpiStateBaseHeader(uint8 pt, uint8 pi, uint32 ts, uint8 _pri)
      : RtpiAduHeader(pt,pi,ts), o(0), snc(0), pri(_pri), pids(), sns(), objectSubid(0) {}
    RtpiStateBaseHeader(uint8 pt, uint8 pi, uint32 ts, uint8 _pri, uint32 _objectSubid)
      : RtpiAduHeader(pt,pi,ts), o(1), snc(0), pri(_pri), pids(), sns(), objectSubid(_objectSubid) {}
    virtual ~RtpiStateBaseHeader() {}

    virtual uint8 GetType() = 0;

    int  GetPriority() {return pri;}
    void SetPriority(int priority) {pri = priority;}

    bool ObjectIDIsSet() { return o; }
    void SetObjectID(uint32 subid) {objectSubid = subid; o = 1;}
    uint32 GetObjectID() {return objectSubid;}

    virtual int Parse(char* pos) = 0;
    virtual int Flush(char* pos) = 0;

    virtual int GetSize() = 0;

    virtual void Print() = 0;

    void AddSequenceNumber(uint32 _pid, uint16 _sn) {pids.push_back(_pid); sns.push_back(_sn); ++snc;}
    void ClearEventList() {pids.clear(); sns.clear(); snc = 0;}

  //protected:
    uint8  o;            // 1 b:  if 1, header includes the ID of another subcomponent to express
                         //       an object relationship
    uint8  snc;          // 8 b:  sequence number count
    uint8  pri;          // 2 b:  priority
    std::vector<uint32> pids; // snc * 32 b: list of participant id's
    std::vector<uint16> sns;  // snc * 16 b: list of sequence numbers, corresponding to the pid list
    uint32 objectSubid;  // 32 b: object subcomponent ID
  };


  class RtpiStateHeader : public RtpiStateBaseHeader {
  public:
    RtpiStateHeader()
      : RtpiStateBaseHeader() {}
    RtpiStateHeader(uint8 pt, uint8 pi, uint32 ts, uint8 pri)
      : RtpiStateBaseHeader(pt,pi,ts,pri) {}
    RtpiStateHeader(uint8 pt, uint8 pi, uint32 ts, uint8 pri, uint32 objectSubid)
      : RtpiStateBaseHeader(pt,pi,ts,pri,objectSubid) {}
    ~RtpiStateHeader() {}

    RtpiAduHeader* Copy();

    uint8 GetType() {return RtpiHeader::PKT_STATE;}

    int Parse(char* pos);
    int Flush(char* pos);

    int GetSize();

    void Print();
  };


  class RtpiDeltaStateHeader : public RtpiStateBaseHeader {
  public:
    RtpiDeltaStateHeader()
      : RtpiStateBaseHeader(), dsn(0), dpid(0) {}
    RtpiDeltaStateHeader(uint8 pt, uint8 pi, uint32 ts, uint8 pri, uint16 _dsn, uint32 _dpid)
      : RtpiStateBaseHeader(pt,pi,ts,pri), dsn(_dsn), dpid(_dpid) {}
    RtpiDeltaStateHeader(uint8 pt, uint8 pi, uint32 ts, uint8 pri, uint32 objectSubid, uint16 _dsn, uint32 _dpid)
      : RtpiStateBaseHeader(pt,pi,ts,pri,objectSubid), dsn(_dsn), dpid(_dpid) {}
    ~RtpiDeltaStateHeader() {}

    RtpiAduHeader* Copy();

    uint8 GetType() {return RtpiHeader::PKT_DELTA_STATE;}

    int Parse(char* pos);
    int Flush(char* pos);

    int GetSize();

    void Print();

    // private:   (should be ;) )
    uint16 dsn;     // 16 b: sequence number of referenced state
    uint32 dpid;    // 32 b; participant identifier of referenced state
  };


  class RtpiAdu {
  public:
    RtpiAdu()
      : packetHeader(NULL), aduHeader(NULL), packetHeaderStart(NULL), aduHeaderStart(NULL), payload(NULL), payloadSize(0) {}

    RtpiAdu(RtpiHeader* phdr, RtpiAduHeader* ahdr) // Version for State Query
      : packetHeader(phdr), aduHeader(ahdr), packetHeaderStart(NULL), aduHeaderStart(NULL), payload(NULL), payloadSize(0) {
      if (phdr->GetType() != ahdr->GetType())
	throw RtpiException(RtpiException::ERR_RTPI_PACKET_FORMAT, "RtpiAdu::RtpiAdu - Header types mismatch");
    }

    RtpiAdu(RtpiHeader* phdr, RtpiAduHeader* ahdr, RefCount<Buffer> _buf, char* _payload, int _payloadSize)
      : packetHeader(phdr), aduHeader(ahdr), buf(_buf), packetHeaderStart(NULL), aduHeaderStart(NULL), payload(_payload), payloadSize(_payloadSize) {
      if (phdr->GetType() != ahdr->GetType())
	throw RtpiException(RtpiException::ERR_RTPI_PACKET_FORMAT, "RtpiAdu::RtpiAdu - Header types mismatch");
    }

    virtual ~RtpiAdu();

    virtual RtpiAdu* Copy() = 0;

    uint32 GetSubcomponentID() {return packetHeader->GetSubcomponentID();}
    uint32 GetParticipantID() {return packetHeader->GetParticipantID();}

    RtpiHeader* GetPacketHeader() {return packetHeader;}
    void SetPacketHeader(RtpiHeader* hdr) {
      if (packetHeader)
	delete packetHeader;
      packetHeader = hdr;
    }

    RtpiAduHeader* GetAduHeader() {return aduHeader;}

    virtual char* GetPayload() {return payload;}
    virtual int GetPayloadSize() {return payloadSize;}
    virtual void SetPayload(RefCount<Buffer> _buf, char* _payload, int _payloadSize) {
      if ((buf != RefCount<Buffer>()) && (buf != _buf))
	throw RtpiException(RtpiException::ERR_RTPI_INTERNAL, "RtpiAdu::SetPayload - Illegal Buffer change!");
      buf = _buf;
      payload = _payload;
      payloadSize = _payloadSize;
    }

    static RtpiAdu* CreateAndParse(RefCount<Buffer> b);
    static RtpiAdu* CreateRtpiAdu(int type);
    int Parse(RefCount<Buffer> _buf);
    int Flush();                                   // flushes complete adu (payload already there) into internal buffer

    virtual uint8 GetType() = 0;

    int GetAduSize();
    int GetPacketSize() {
      if (packetHeader == NULL)
	return -1;
      return packetHeader->GetSize() + aduHeader->GetSize() + payloadSize;
    }
    int GetAduHeaderSize() {return aduHeader->GetSize();}
    int GetPacketHeaderSize() {return packetHeader->GetSize();}

    RefCount<Buffer> GetBuffer() {return buf;}

    void Print();

  protected:
    RtpiAduHeader* CreateAduHeader(int type);
    int ParseHeader(char* pos, int type);
    int Fragment(char* pos, int start, int bytes); // flushes chunks of adu-header and payload data into buffer of RTP/I packets

    friend class TransparentReliableTransport;
    friend class UnreliableTransport;
    friend class RtpiHandler;
    friend class RtpiParticipant;

    RtpiHeader*           packetHeader;
    RtpiAduHeader*        aduHeader;

    RefCount<Buffer> buf;

    // buffer pointers
    char* packetHeaderStart;
    char* aduHeaderStart;
    char* payload;
    int payloadSize;
  };


  class RtpiEvent : public RtpiAdu {
  public:
    RtpiEvent()
      : RtpiAdu() {}
    RtpiEvent(uint32 subid, uint8 pt, uint8 pi, uint32 ts, RefCount<Buffer> buf)
      : RtpiAdu(new RtpiHeader(0,subid,RtpiHeader::PKT_EVENT,0,0,1), new RtpiEventHeader(pt,pi,ts), buf, buf->getCont(), buf->getSize()) {}
    RtpiEvent(uint32 pid, uint32 subid, uint16 sn, uint8 pt, uint8 pi, uint32 ts, RefCount<Buffer> buf,  char* payload, int payloadSize)
      : RtpiAdu(new RtpiHeader(pid,subid,RtpiHeader::PKT_EVENT,sn,0,1), new RtpiEventHeader(pt,pi,ts), buf, payload, payloadSize) {}
    RtpiEvent(uint32 subid, uint8 pt, uint8 pi, uint32 ts, char* payload, int payloadSize)
      : RtpiAdu(new RtpiHeader(0,subid,RtpiHeader::PKT_EVENT,0,0,1), new RtpiEventHeader(pt,pi,ts), NULL, payload, payloadSize) {}
    ~RtpiEvent() {}

    RtpiAdu* Copy();

    uint8 GetType() {return RtpiHeader::PKT_EVENT;}

    RtpiEventHeader* GetAduHeader() {return (RtpiEventHeader*)aduHeader;}

    bool ObjectIDIsSet() { return ((RtpiEventHeader*)aduHeader)->ObjectIDIsSet(); }
    void SetObjectID(uint32 subid) {((RtpiEventHeader*)aduHeader)->SetObjectID(subid);}
    uint32 GetObjectID() {return ((RtpiEventHeader*)aduHeader)->GetObjectID();}

  private:
    friend class UnreliableTransport;
    friend class TransparentReliableTransport;
    friend class RtpiHandler;
    friend class RtpiParticipant;
  };

  class RtpiCue : public RtpiAdu {
  public:
    RtpiCue()
      : RtpiAdu() {}
    RtpiCue(uint32 subid, uint8 pt, uint8 pi, uint32 ts, RefCount<Buffer> buf)
      : RtpiAdu(new RtpiHeader(0,subid,RtpiHeader::PKT_CUE,0,0,1), new RtpiCueHeader(pt,pi,ts), buf, buf->getCont(), buf->getSize()) {}
    RtpiCue(uint32 pid, uint32 subid, uint16 sn, uint8 pt, uint8 pi, uint32 ts, RefCount<Buffer> buf, char* payload, int payloadSize)
      : RtpiAdu(new RtpiHeader(pid,subid,RtpiHeader::PKT_CUE,sn,0,1), new RtpiCueHeader(pt,pi,ts), buf, payload, payloadSize) {}
    RtpiCue(uint32 subid, uint8 pt, uint8 pi, uint32 ts, char* payload, int payloadSize)
      : RtpiAdu(new RtpiHeader(0,subid,RtpiHeader::PKT_CUE,0,0,1), new RtpiCueHeader(pt,pi,ts), NULL, payload, payloadSize) {}
    ~RtpiCue() {}

    RtpiAdu* Copy();

    uint8 GetType() {return RtpiHeader::PKT_CUE;}

    RtpiCueHeader* GetAduHeader() {return (RtpiCueHeader*)aduHeader;}

  private:
    friend class UnreliableTransport;
    friend class TransparentReliableTransport;
    friend class RtpiHandler;
    friend class RtpiParticipant;
  };


  class RtpiStateBase : public RtpiAdu {
  public:
    RtpiStateBase()
      : RtpiAdu() {}
    RtpiStateBase(RtpiHeader* phdr, RtpiAduHeader* ahdr, RefCount<Buffer> buf)
      : RtpiAdu(phdr, ahdr, buf, buf->getCont(), buf->getSize()) {}
    RtpiStateBase(RtpiHeader* phdr, RtpiAduHeader* ahdr, RefCount<Buffer> buf, char* payload, int payloadSize)
      : RtpiAdu(phdr, ahdr, buf, payload, payloadSize) {}
    virtual ~RtpiStateBase() {}

    virtual uint8 GetType() = 0;

    RtpiStateBaseHeader* GetAduHeader() {return (RtpiStateBaseHeader*)aduHeader;}

    bool ObjectIDIsSet() { return ((RtpiStateBaseHeader*)aduHeader)->ObjectIDIsSet(); }
    void SetObjectID(uint32 subid) {((RtpiStateBaseHeader*)aduHeader)->SetObjectID(subid);}
    uint32 GetObjectID() {return ((RtpiStateBaseHeader*)aduHeader)->GetObjectID();}

    void AddSequenceNumberToHeader(uint32 pid, uint16 sn) {((RtpiStateBaseHeader*)aduHeader)->AddSequenceNumber(pid,sn);}

  private:
    friend class UnreliableTransport;
    friend class TransparentReliableTransport;
    friend class RtpiHandler;
    friend class RtpiParticipant;
  };


  class RtpiState : public RtpiStateBase {
  public:
    RtpiState()
      : RtpiStateBase() {}
    RtpiState(uint32 subid, uint8 pt, uint8 pi, uint32 ts, uint8 pri, RefCount<Buffer> buf)
      : RtpiStateBase(new RtpiHeader(0,subid,RtpiHeader::PKT_STATE,0,0,1), new RtpiStateHeader(pt,pi,ts,pri), buf) {}
    RtpiState(uint32 pid, uint32 subid, uint16 sn, uint8 pt, uint8 pi, uint32 ts, uint8 pri, RefCount<Buffer> buf, char* payload, int payloadSize)
      : RtpiStateBase(new RtpiHeader(pid,subid,RtpiHeader::PKT_STATE,sn,0,1), new RtpiStateHeader(pt,pi,ts,pri), buf, payload, payloadSize) {}
    RtpiState(uint32 subid,uint8 pt, uint8 pi, uint32 ts, uint8 pri, char* payload, int payloadSize)
      : RtpiStateBase(new RtpiHeader(0,subid,RtpiHeader::PKT_STATE,0,0,1), new RtpiStateHeader(pt,pi,ts,pri), NULL, payload, payloadSize) {}
    virtual ~RtpiState() {}

    RtpiAdu* Copy();

    uint8 GetType() {return RtpiHeader::PKT_STATE;}

    RtpiStateHeader* GetAduHeader() {return (RtpiStateHeader*)aduHeader;}

  private:
    friend class UnreliableTransport;
    friend class TransparentReliableTransport;
    friend class RtpiHandler;
    friend class RtpiParticipant;
  };

  class RtpiDeltaState : public RtpiStateBase {
  public:
    RtpiDeltaState()
      : RtpiStateBase() {}
    RtpiDeltaState(uint32 pid, uint32 subid, uint16 sn, uint8 pt, uint8 pi, uint32 ts, uint8 pri, uint16 dsn, uint32 dpid,
		   RefCount<Buffer> buf)
      : RtpiStateBase(new RtpiHeader(pid,subid,RtpiHeader::PKT_DELTA_STATE,sn,0,1),
		      new RtpiDeltaStateHeader(pt,pi,ts,pri,dsn,dpid), buf) {}
    RtpiDeltaState(uint32 pid, uint32 subid, uint16 sn, uint8 pt, uint8 pi, uint32 ts, uint8 pri, uint16 dsn, uint32 dpid,
		   RefCount<Buffer> buf, char* payload, int payloadSize)
      : RtpiStateBase(new RtpiHeader(pid,subid,RtpiHeader::PKT_DELTA_STATE,sn,0,1),
		      new RtpiDeltaStateHeader(pt,pi,ts,pri,dsn,dpid), buf, payload, payloadSize) {}
    RtpiDeltaState(uint32 subid, uint8 pt, uint8 pi, uint32 ts, uint8 pri, uint16 dsn, uint32 dpid,
		   char* payload, int payloadSize)
      : RtpiStateBase(new RtpiHeader(0,subid,RtpiHeader::PKT_DELTA_STATE,0,0,1),
		      new RtpiDeltaStateHeader(pt,pi,ts,pri,dsn,dpid), NULL, payload, payloadSize) {}
    virtual ~RtpiDeltaState() {}

    RtpiAdu* Copy();

    uint8 GetType() {return RtpiHeader::PKT_DELTA_STATE;}

    RtpiDeltaStateHeader* GetAduHeader() {return (RtpiDeltaStateHeader*)aduHeader;}

  private:
    friend class UnreliableTransport;
    friend class TransparentReliableTransport;
    friend class RtpiHandler;
    friend class RtpiParticipant;
  };

  class RtpiStateQuery : public RtpiAdu {
  public:
    RtpiStateQuery()
      : RtpiAdu() {}
    RtpiStateQuery(uint32 pid, uint32 subid, uint16 sn, uint8 pt, uint32 ts, uint8 pri)
      : RtpiAdu(new RtpiHeader(pid,subid,RtpiHeader::PKT_STATE_QUERY,sn,0,1), new RtpiStateQueryHeader(pt,0,ts,pri)) {}
    RtpiStateQuery(uint32 subid, uint8 pt, uint32 ts, uint8 pri)
      : RtpiAdu(new RtpiHeader(0,subid,RtpiHeader::PKT_STATE_QUERY,0,0,1), new RtpiStateQueryHeader(pt,0,ts,pri)) {}
    virtual ~RtpiStateQuery() {}

    RtpiAdu* Copy();

    uint8 GetType() {return RtpiHeader::PKT_STATE_QUERY;}

    RtpiStateQueryHeader* GetAduHeader() {return (RtpiStateQueryHeader*)aduHeader;}

    char* GetPayload() {
      throw RtpiException(RtpiException::ERR_RTPI_PACKET_FORMAT, "RtpiAdu::GetPayload - A State Query does not have a payload");
    }
    int GetPayloadSize() {
      throw RtpiException(RtpiException::ERR_RTPI_PACKET_FORMAT, "RtpiAdu::GetPayloadSize - A State Query does not have a payload");
    }
    void SetPayload(RefCount<Buffer> _buf, char* _payload, int _payloadSize) {
      throw RtpiException(RtpiException::ERR_RTPI_PACKET_FORMAT, "RtpiAdu::SetPayload - A State Query does not have a payload");
    }

  private:
    friend class UnreliableTransport;
    friend class TransparentReliableTransport;
    friend class RtpiHandler;
    friend class RtpiParticipant;
  };

  typedef std::vector<RefCount<RtpiAdu> > RtpiAduVector;
  typedef std::list<RefCount<RtpiAdu> > RtpiAduList;


  class RtpiCompoundPacket {
  public:
    RtpiCompoundPacket(RefCount<Buffer> _buf) : buf(_buf) {}
    ~RtpiCompoundPacket() {}

    void AddPacket(RefCount<RtpiPacket> packet);

    char* GetBufferStart() {
      if (buf == RefCount<Buffer>())
	throw RtpiException(RtpiException::ERR_RTPI_INTERNAL, "RtpiCompoundPacket::GetStart - no buffer.");
      return buf->getCont();
    }
    int GetBufferSize() {  // returns size of buffer
      if (buf == RefCount<Buffer>())
	throw RtpiException(RtpiException::ERR_RTPI_INTERNAL, "RtpiCompoundPacket::GetSize - no buffer.");
      return buf->getSize();
    }
    RefCount<Buffer> GetBuffer() {
      if (buf == RefCount<Buffer>())
	throw RtpiException(RtpiException::ERR_RTPI_INTERNAL, "RtpiCompoundPacket::GetBuffer - no buffer.");
      return buf;
    }
    void SetBuffer(RefCount<Buffer> _buf) {
      buf = _buf;
    }

    void Print();

    RtpiPacketVector packets;
    RefCount<Buffer> buf;
  };

  typedef std::vector<RefCount<RtpiCompoundPacket> > RtpiCompoundPacketVector;

#ifdef WIN32
  using std::list<RefCount<RtpiPacket> >;
  using std::list<std::list<RefCount<RtpiPacket> >*>;
#endif

  /*
  class RtpiAduFactory {
  public:
    RtpiAduFactory() {}
    ~RtpiAduFactory();

    // Flush
    // juergi: method added - please check if ok
    RtpiPacketVector* CreateRtpiPackets(RtpiAdu* adu);
    RtpiPacketVector* CreateRtpiPackets(RtpiAdu* adu, RefCount<Buffer> buf);
    RtpiCompoundPacketVector* CreateRtpiPackets(RefCount<RtpiAduVector> adus);

    // Parse
    RtpiAduVector* CreateRtpiAdusFromPackets(RefCount<Buffer> buf);
    RtpiAduVector* CreateRtpiAdusFromPackets(RtpiPacketVector* rtpiPackets);
    RtpiAdu* CreateRtpiAduFromAdu(RefCount<Buffer> b);

    void DeleteAduFragments(uint32 pid, uint64 subid, uint8 type, uint16 sn);
  private:
    RtpiAdu* CreateRtpiAdu(int type);
    bool AduComplete(list<RefCount<RtpiPacket> >* fragmentList);
    int  GetSizeOfAdu(list<RefCount<RtpiPacket> >* fragmentList);
    void Print(list<RefCount<RtpiPacket> >* fragmentList);

    // manage lists of ADU fragments / RTP/I packets
    list<RefCount<RtpiPacket> >* FindFragmentList(uint64 subid, uint8 type, uint32 pid, uint16 sn);
    list<RefCount<RtpiPacket> >* InsertFragmentList(uint64 subid, uint8 type, uint32 pid, uint16 sn);
    void DeleteFragmentList(uint64 subid, uint8 type, uint32 pid, uint16 sn);

    list<list<RefCount<RtpiPacket> >*> fragments;
  };
  */

}

#endif
