// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//
// Copyright (c) 2000 University of Mannheim. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for educational, research and non-profit purposes,
// without fee, and without a written agreement is hereby granted,
// provided that the above copyright notice, this paragraph and the
// following four paragraphs appear in all copies.
//
// Permission to incorporate this software into commercial products may
// be obtained from the following address:
// University of Mannheim
// Lehrstuhl fuer Praktische Informatik IV
// L15, 16
// 68161 Mannheim, Germany
//
// IN NO EVENT SHALL THE UNIVERSITY OF MANNHEIM BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
// ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF MANNHEIM HAS BEEN ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//
// THE UNIVERSITY OF MANNHEIM SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER
// IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF MANNHEIM HAS NO OBLIGATIONS
// TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The University of Mannheim will appreciate receiving any enhancements and
// modifications of this software made by others, and reserves the right to
// include them in future releases. The University also recommends that this
// software not be redistributed by third parties, but be obtained by all
// directly from the University's server, so that all users of the software
// may receive bug notifications and other relevant announcements.
//
// Authors:
//   Volker Hilt, Juergen Vogel, Martin Mauve
//   University of Mannheim
//   {hilt,vogel,mauve}@informatik.uni-mannheim.de
//
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#ifndef _RtpiSubComponentSpace_h_
#define _RtpiSubComponentSpace_h_

#ifdef WIN32
#  pragma warning (disable: 4786 4181)
#endif

#include <map>
#include <list>
#include <iterator>
#include <string>
#include <utility>
#include <iostream>

#include <util/types.h>
#include <util/strtools.h>
#include <util/tools.h>
#include <util/random.h>
#include <rtpi/rtpiPacket.h>
#include <rtpi/rtcpiPacket.h>
#include <rtpi/rtpiException.h>
#include <rtpi/rtpiInterfaces.h>


namespace RTPI {

  class RtpiADUType;
  class RtpiSubComponent;
  class RtpiSubComponentSpace;
  class RtpiHandler;
  class RtpiParticipant;

  class RtpiParticipantSeq {
  public:
    enum seq_type {
      SEQ_OK = 0,
      SEQ_DUPLICATE = 1,
      SEQ_GAP = 2,
      SEQ_JUMP = 3,
      SEQ_INITIALIZED = 4
    };

    static void err_out(uint8);

    RtpiParticipantSeq() : initialized(false), last_seq(0), base_seq(0),
      seq_cycles(0) { };

    void init_seq(uint16 _seq = 0) {
      // std::cout << "INIT seq. " << std::endl;
      base_seq = _seq - 1;
      last_seq = base_seq;
      seq_cycles = 0;
      initialized = true;
    }
    uint8 seq_ok(uint16);
    uint16 get_next_seq();
    void insert_gap(uint16);

  private:
    friend class RtpiADUType;
    friend class RtpiSubComponentSpace;
    friend std::ostream& operator<<(std::ostream&, std::map<u_int32,RtpiParticipantSeq>&);

    bool initialized;

    // pid is first attribute of map
    // sequence number generation
    u_int16	last_seq;  // last seq
    u_int32	base_seq; // first seq - 1
    u_int32	seq_cycles;
  };
  typedef std::map<u_int32,RtpiParticipantSeq> ParticipantSeqMap;
  typedef std::pair<const uint32,RtpiParticipantSeq> ParticipantSeqMapElement;


  class RtpiADUType {
  public:
    RtpiADUType() { };

  private:
    friend class RtpiSubComponent;
    friend class RtpiSubComponentSpace;
    friend std::ostream& operator<<(std::ostream&, std::map<u_int8,RtpiADUType>&);

    // type is first attribute of map
    ParticipantSeqMap participants;
  };
  typedef std::map<u_int8,RtpiADUType> ADUTypeSeqMap;


  class RtpiSubComponent {
  public:
    enum change_code {
      SUBCOMP_OK = 0x0,
      SUBCOMP_NO_CHANGES = 0x01,
      SUBCOMP_CHANGED_NAME = 0x02,
      SUBCOMP_CHANGED_ACTIVE = 0x04,
      SUBCOMP_PAYLOAD_CHANGED = 0x08,
      SUBCOMP_SET_LOCAL = 0x10,
      SUBCOMP_INITIALIZED = 0x20
    };

    RtpiSubComponent(uint32 _subid) :
      initialized(false), local_active(false), global_active(false),
      reported_active(false), local(false), was_local(false), name(),
      name_length(0), last_report(), reported_by(0), timeout_base(unixtime()),
      last_pt(-1), flag(false), alt_subid(_subid), subid(_subid) {
      name[0] = '\0';
      timerclear(&last_report);
    };

    RtpiSubComponent(uint32 _subid, uint32 _alt_subid) :
      initialized(false), local_active(false), global_active(false),
      reported_active(false), local(false), was_local(false), name(),
      name_length(0), last_report(), reported_by(0), timeout_base(unixtime()),
      last_pt(-1), flag(false), alt_subid(_alt_subid), subid(_subid) {
      name[0] = '\0';
      timerclear(&last_report);
    };

    // methods for application
    // add sub-component
    uint8 add_component(bool _active, char* _name = NULL, int _name_length = 0);
    // releases sub-component
    void release_component();
    // get / set activation state
    void set_active(bool _active) {
      initialized = true;
      local = true;
      local_active = _active;
    };
    bool is_active() { return (local_active || global_active); }
    // get name (to set name please call add component)
    char* get_name() { return name; }
    int get_name_length() { return name_length; }

    // methods for protocol stack
    // packet received for a sub-component
    uint8 data_received(int _last_pt = -1);
    // packet sent for a sub-component
    uint8 data_sent(int _last_pt = -1);
    // report received for a sub-component
    uint8 report_received(uint32 _pid, bool _active, char* _name, int _name_length);
    // report sent for a sub-component
    uint8 report_sent(uint32 _pid);

    // check timeouts
    bool must_report(uint32 _pid, int interval);
    bool is_timed_out(int interval);

    int  get_last_pt() { return last_pt; }

    // access user defined flag
    bool get_flag() { return flag; }
    void set_flag(bool f) { flag = f; }

    uint32 get_subid() {return subid;}

  private:
    friend class RtpiSubComponentSpace;
    friend class RtpiHandler;
    friend class RtpiParticipant;
    friend std::ostream& operator<<(std::ostream&, std::map<uint32,RtpiSubComponent>&);

    bool get_local_active() { return local_active; }
    bool get_global_active() { return global_active; }

    void set_local(bool _local) { local = _local; }
    uint8 refresh_component();

    bool      initialized;
    bool      local_active;
    bool      global_active;
    bool      reported_active;
    bool      local;
    bool      was_local;
    char      name[256];
    int       name_length;
    timeval   last_report;
    uint32    reported_by;
    timeval   timeout_base;
    int       last_pt;
    uint32    alt_subid;
    uint32    subid;

    // may be used by the application
    bool      flag;

    // subids are first attribute of maps
    ADUTypeSeqMap ADUtypes;
  };
  typedef std::map<uint32,RtpiSubComponent> SubComponentMap;
  typedef std::map<uint32,SubComponentMap::iterator> SubComponentIteratorMap;
  typedef std::pair<const uint32,RtpiSubComponent> SubComponentMapElement;


  class RtpiSubComponentSpace {
  public:
    // get components and participant_seqs
    SubComponentMap::iterator find_component(uint32 _subid);
    SubComponentMap::iterator altsubid_find_component(uint32 _alt_subid);
    SubComponentMap::iterator insert_component(uint32 _subid) {
      return insert_component(_subid, _subid);
    }
    SubComponentMap::iterator insert_component(uint32 _subid, uint32 _alt_subid);

    ParticipantSeqMapElement* find_participant_seq(uint32 _subid, u_int8 _type, u_int32 _pid);
    ParticipantSeqMapElement* insert_participant_seq(uint32 _subid, u_int8 _type, u_int32 _pid);

    // manage participants / adu types / sub components
    void delete_participant_seq(u_int32);
    void delete_component(uint32);
    SubComponentMap::iterator delete_component(SubComponentMap::iterator);
    SubComponentMap::iterator get_components_begin() {
      return subComponents.begin(); }
    SubComponentMap::iterator get_components_end() {
      return subComponents.end(); }
    SubComponentIteratorMap::iterator get_alt_components_begin() {
      return altsubid_subComponents.begin(); }
    SubComponentIteratorMap::iterator get_alt_components_end() {
      return altsubid_subComponents.end(); }
    bool components_empty() {
      return subComponents.empty(); }

    // translate between alt_subid and subid
    std::pair<bool, uint32> altsubid_to_subid(uint32 _alt_subid) {
      SubComponentMap::iterator pos = altsubid_find_component(_alt_subid);
      if (pos != subComponents.end()) {
	return std::make_pair(true, pos->first);
      } else {
	return std::make_pair(false, 0);
      }
    }
    std::pair<bool, uint32> subid_to_altsubid(uint32 _subid) {
      SubComponentMap::iterator pos = find_component(_subid);
      if (pos != subComponents.end()) {
		  return std::make_pair(true, pos->second.alt_subid);
      } else {
	return std::make_pair(false, 0);
      }
    }



    // helpers to update all subcomponents via rtcpi packets
    // used by the protocol stack
    void clear_subcomps_local();
    void release_nonlocal_subcomps();

    void release_all_components();

    // helpers
    void clear_subcomp_flag();

  private:
    friend std::ostream& operator<<(std::ostream&, RtpiSubComponentSpace&);
    friend class RtpiSubComponent;

    SubComponentMap subComponents;
    SubComponentIteratorMap altsubid_subComponents;
  };
};

#endif






