// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//
// Copyright (c) 2000 University of Mannheim. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for educational, research and non-profit purposes,
// without fee, and without a written agreement is hereby granted,
// provided that the above copyright notice, this paragraph and the
// following four paragraphs appear in all copies.
//
// Permission to incorporate this software into commercial products may
// be obtained from the following address:
// University of Mannheim
// Lehrstuhl fuer Praktische Informatik IV
// L15, 16
// 68161 Mannheim, Germany
//
// IN NO EVENT SHALL THE UNIVERSITY OF MANNHEIM BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
// ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF MANNHEIM HAS BEEN ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//
// THE UNIVERSITY OF MANNHEIM SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER
// IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF MANNHEIM HAS NO OBLIGATIONS
// TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The University of Mannheim will appreciate receiving any enhancements and
// modifications of this software made by others, and reserves the right to
// include them in future releases. The University also recommends that this
// software not be redistributed by third parties, but be obtained by all
// directly from the University's server, so that all users of the software
// may receive bug notifications and other relevant announcements.
//
// Authors:
//   Volker Hilt, Juergen Vogel, Martin Mauve
//   University of Mannheim
//   {hilt,vogel,mauve}@informatik.uni-mannheim.de
//
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#include <math.h>
#ifdef WIN32
#include <util/win32.h>
#endif

#include "rtpiTransport.h"


//
// Functions from class UnreliableTransport
//
RTPI::UnreliableTransport::~UnreliableTransport() {
  for (std::list<AduFragments*>::iterator i = aduFragmentList.begin(); i != aduFragmentList.end(); ++i)
    delete *i;
}


void RTPI::UnreliableTransport::CheckFragmentLifetime() {
  struct timeval time;
  gettimeofday(&time, NULL);
  for (std::list<AduFragments*>::iterator i = aduFragmentList.begin(); i != aduFragmentList.end(); ++i) 
    if ((*i)->time.tv_sec + lifetime < time.tv_sec) {
      // adu has not been completed in lifetime
      delete *i;
      aduFragmentList.erase(i);
    }
}


// fragment a RTP/I ADU into one or more RTP/I packets; these packtes are flushed into the buffer given
// by the appliction; ! if this buffer is deleted, all RTP/I packets become obsolete and  must be destroyed
// if no header extension shall be attached, the packets are completely flushed;
//   however, the packet header can be changed and the packet flushed again (the payload stays in its assigned place
// if a header extension shall be attached, this has to be done manually by calling RtpiPacket::SetHeaderExtension for
//   each packet of the vector returend; afterwards all packets have to be flushed manually (call RtpiPacket::Flush)
//   the payload is already in it's correct position
RTPI::RtpiPacketVector* RTPI::UnreliableTransport::CreateRtpiPackets(RefCount<RtpiAdu> adu) {

  int aduSize = adu->GetAduSize();
  int reservedPacketSize = RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize();
  int maxRtpiPayloadSize = RtpiPacket::MAX_PKT_SIZE - reservedPacketSize;
  int rtpiPacketsNeeded  = (int)ceil((double)aduSize / maxRtpiPayloadSize);

  RtpiPacketVector* packets = new RtpiPacketVector();
  RefCount<RtpiPacket> packet;
  RtpiHeader* hdr;

  int totalAduBytesFlushed = 0;
  int bytesFlushed = 0;
  int fc = 0;

  while (totalAduBytesFlushed < aduSize) {
    // reserve a Buffer for each packet
    RefCount<Buffer> buf = new Buffer(RtpiPacket::MAX_PKT_SIZE);
    char* pos = buf->getCont();

    // create new RTP/I packet
    packet = new RtpiPacket(buf, pos);
    packets->push_back(packet);

    pos += (RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize());

    // write payload of RTP/I packet  = piece of RTP/I adu
    int bytesLeft = aduSize - totalAduBytesFlushed;
    int bytesToFlush = (bytesLeft <= maxRtpiPayloadSize) ? bytesLeft : maxRtpiPayloadSize;
    adu->Fragment(pos, totalAduBytesFlushed, bytesToFlush);

    // write RTP/I packet header
    packet->SetPayload(buf, pos, bytesToFlush);
    hdr = new RtpiHeader();
    hdr->e = (fc == rtpiPacketsNeeded - 1) ? 1 : 0;
    hdr->type = adu->GetType();
    hdr->rt = adu->packetHeader->rt;
    hdr->pid = adu->packetHeader->pid;
    hdr->subid = adu->packetHeader->subid;
    hdr->sn = adu->packetHeader->sn;
    hdr->fc = fc;
    packet->SetHeader(hdr);
    // if no header extension the packet is flushed completely; if extension, the packet has to be flushed again after
    // the extension has been attached
    bytesFlushed = packet->Flush();

    // resize buffer to number of bytes written; can be smaller than value above
    buf->setSize(bytesFlushed);

    // prepare for next round
    totalAduBytesFlushed += bytesToFlush;

    fc++;
  }

  return packets;
}


// fragment a RTP/I ADU into one or more RTP/I packets; these packtes are flushed into the buffer given
// by the appliction; ! if this buffer is deleted, all RTP/I packets become obsolete and  must be destroyed
// if no header extension shall be attached, the packets are completely flushed;
//   however, the packet header can be changed and the packet flushed again (the payload stays in its assigned place
// if a header extension shall be attached, this has to be done manually by calling RtpiPacket::SetHeaderExtension for
//   each packet of the vector returend; afterwards all packets have to be flushed manually (call RtpiPacket::Flush)
//   the payload is already in it's correct position
RTPI::RtpiPacketVector* RTPI::UnreliableTransport::CreateRtpiPackets(RefCount<RtpiAdu> adu, RefCount<Buffer> buf) {

  int aduSize = adu->GetAduSize();
  int reservedPacketSize = RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize() + RtpiPacket::GetPaddingBorder();
  // juergi: wozu ist hier GetPaddingBorder noetig? Stimmt das Ergebnis?
  int maxRtpiPayloadSize = RtpiPacket::MAX_PKT_SIZE - reservedPacketSize;
  int rtpiPacketsNeeded  = (int)ceil((double)aduSize / maxRtpiPayloadSize);

  // resize buffer
  buf->resize(aduSize + rtpiPacketsNeeded * reservedPacketSize);

  RtpiPacketVector* packets = new RtpiPacketVector();
  RefCount<RtpiPacket> packet;
  RtpiHeader* hdr;

  char* pos = buf->getCont();
  int totalBytesFlushed = 0;
  int totalAduBytesFlushed = 0;
  int bytesFlushed = 0;
  int fc = 0;

  while (totalAduBytesFlushed < aduSize) {
    // create new RTP/I packet
    packet = new RtpiPacket(buf, pos);
    packets->push_back(packet);

    pos += (RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize());

    // write payload of RTP/I packet  = piece of RTP/I adu
    int bytesLeft = aduSize - totalAduBytesFlushed;
    int bytesToFlush = (bytesLeft <= maxRtpiPayloadSize) ? bytesLeft : maxRtpiPayloadSize;
    adu->Fragment(pos, totalAduBytesFlushed, bytesToFlush);

    // write RTP/I packet header
    packet->SetPayload(buf, pos, bytesToFlush);
    hdr = new RtpiHeader();
    hdr->e = (fc == rtpiPacketsNeeded - 1) ? 1 : 0;
    hdr->type = adu->GetType();
    hdr->rt = adu->packetHeader->rt;
    hdr->pid = adu->packetHeader->pid;
    hdr->subid = adu->packetHeader->subid;
    hdr->sn = adu->packetHeader->sn;
    hdr->fc = fc;
    packet->SetHeader(hdr);
    // if no header extension the packet is flushed completely; if extension, the packet has to be flushed again after
    // the extension has been attached
    bytesFlushed = packet->Flush();
    totalBytesFlushed += bytesFlushed;

    // prepare for next round
    totalAduBytesFlushed += bytesToFlush;
    pos += (bytesFlushed - RtpiHeader::GetSize() - RtpiHeaderExtension::GetSize()); // payload plus padding ; rest has been taken care of above

    fc++;
  }

  // resize buffer to number of bytes written; can be smaller than value above
  buf->setSize(totalBytesFlushed);

  return packets;
}


// create RTP/I packets from a list of RTP/I ADUs: if a ADU is too large for one transport packet it is
// fragmented into several RTP/I packets; if not, only one RTP/I packet is created per ADU. If several
// RTP/I packets fit into one tranpsort packet, a compound packet is created containing a list of
// RTP/I packets which share the same buffer.
// ! if a buffer is deleted, all corresponding RTP/I packets become obsolete and  must be destroyed
// if no header extension shall be attached, the packets are completely flushed;
//   however, the packet header can be changed and the packet flushed again (the payload stays in its assigned place
// if a header extension shall be attached, this has to be done manually by calling RtpiPacket::SetHeaderExtension for
//   each packet of the vector returend; afterwards all packets have to be flushed manually (call RtpiPacket::Flush)
//   the payload is already in it's correct position
RTPI::RtpiCompoundPacketVector* RTPI::UnreliableTransport::CreateRtpiPackets(RefCount<RtpiAduVector> adus) {
  if ((!adus) || (adus->size() == 0))
    return NULL;

  int totalSize = 0;
  int aduSize = 0;
  int reservedPacketSize = RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize() + RtpiPacket::GetPaddingBorder();
  int maxRtpiPayloadSize = RtpiPacket::MAX_PKT_SIZE - reservedPacketSize;

  RefCount<Buffer> buf = new Buffer(RtpiPacket::MAX_PKT_SIZE);
  int bytesFlushedIntoBuffer = 0;
  RtpiCompoundPacketVector* compoundPackets = new RtpiCompoundPacketVector();
  RefCount<RtpiCompoundPacket> compoundPacket = new RtpiCompoundPacket(buf);
  compoundPackets->push_back(compoundPacket);

  char* pos = buf->getCont();
  RefCount<RtpiPacket> packet;
  RtpiHeader* hdr = NULL;

  for (RtpiAduVector::iterator i = adus->begin(); i != adus->end(); ++i) {
    int totalAduBytesFlushed = 0;
    int bytesFlushed = 0;
    aduSize = (*i)->GetAduSize();
    int rtpiPacketsNeeded = (int)ceil((double)aduSize / maxRtpiPayloadSize);
    int fc = 0;

    if (rtpiPacketsNeeded == 1) {
      //
      // 1. only one RTP/I packet is needed for the complete ADU
      //
      // create new RTP/I packet
      packet = new RtpiPacket(buf, pos);
      compoundPacket->AddPacket(packet);

      pos += (RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize());

      // write payload of RTP/I packet  = complete RTP/I ADU
      (*i)->Fragment(pos, 0, aduSize);

      // write RTP/I packet header
      packet->SetPayload(buf, pos, aduSize);
      hdr = new RtpiHeader();
      hdr->e = 1;
      hdr->type = (*i)->GetType();
      hdr->rt = (*i)->packetHeader->rt;
      hdr->pid = (*i)->packetHeader->pid;
      hdr->subid = (*i)->packetHeader->subid;
      hdr->sn = (*i)->packetHeader->sn;
      hdr->fc = 0;
      packet->SetHeader(hdr);
      // if no header extension the packet is flushed completely; if extension, the packet has to be flushed again after
      // the extension has been attached
      bytesFlushed = packet->Flush();
      bytesFlushedIntoBuffer += bytesFlushed;

      if ((i + 1) != adus->end()) { // there is at least one more ADU
	// prepare for next round
	if ((buf->getSize() - bytesFlushedIntoBuffer) < ((*(i+1))->GetAduSize() + reservedPacketSize)) {
	  // next ADU does not fit into the current buffer -> create new RTP/I Compound packet and new buffer
	  buf->setSize(bytesFlushedIntoBuffer);
	  buf = new Buffer(RtpiPacket::MAX_PKT_SIZE);
	  bytesFlushedIntoBuffer = 0;
	  pos = buf->getCont();
	  compoundPacket = new RtpiCompoundPacket(buf);
	  compoundPackets->push_back(compoundPacket);
	} else {
	  // next ADU does fit into the current buffer
	  pos += (bytesFlushed - RtpiHeader::GetSize() - RtpiHeaderExtension::GetSize()); // payload plus padding ; rest has been taken care of above
	}
      } else {
	// no more ADUs
	buf->setSize(bytesFlushedIntoBuffer);
      }
    } else {
      //
      // 2. the ADU has to be fragmented into several RTP/I packets
      //
      while (totalAduBytesFlushed < aduSize) {
	// create new RTP/I packet
	packet = new RtpiPacket(buf, pos);
	compoundPacket->AddPacket(packet);

	pos += (RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize());

	// write payload of RTP/I packet  = piece of RTP/I adu
	int bytesLeft = aduSize - totalAduBytesFlushed;
	int bytesToFlush = (bytesLeft <= maxRtpiPayloadSize) ? bytesLeft : maxRtpiPayloadSize;
	(*i)->Fragment(pos, totalAduBytesFlushed, bytesToFlush);

	// write RTP/I packet header
	packet->SetPayload(buf, pos,bytesToFlush);
	hdr = new RtpiHeader();
	hdr->e = (fc == rtpiPacketsNeeded - 1) ? 1 : 0;
	hdr->type = (*i)->GetType();
	hdr->rt = (*i)->packetHeader->rt;
	hdr->pid = (*i)->packetHeader->pid;
	hdr->subid = (*i)->packetHeader->subid;
	hdr->sn = (*i)->packetHeader->sn;
	hdr->fc = fc;
	packet->SetHeader(hdr);
	// if no header extension the packet is flushed completely; if extension, the packet has to be flushed again after
	// the extension has been attached
	bytesFlushed = packet->Flush();
	buf->setSize(bytesFlushed);  // resize buffer to number of bytes written; can be smaller than value above

	// prepare for next round
	totalAduBytesFlushed += bytesToFlush;

	fc++;

	// create new compound packet and new buffer
	buf = new Buffer(RtpiPacket::MAX_PKT_SIZE);
	pos = buf->getCont();
	compoundPacket = new RtpiCompoundPacket(buf);
	compoundPackets->push_back(compoundPacket);
      }

      if ((i + 1) == adus->end()) { // there are no more ADUs
	// delete last compound packet
	compoundPackets->pop_back();
      }
    }
  }

  return compoundPackets;
}


// create RTP/I ADUs from RTP/I packets stored in buf
RTPI::RtpiAduVector* RTPI::UnreliableTransport::CreateRtpiAdusFromPackets(RefCount<Buffer> buf) {
  RtpiPacketVector* rtpiPackets = RtpiPacketFactory::CreateRtpiPackets(buf);

  if (rtpiPackets == NULL)
    throw RtpiException(RtpiException::ERR_RTPI_PARSE, "UnreliableTransport::CreateRtpiAdusFromPackets - Buffer corrupt or empty");

  //std::cout << "UnreliableTransport: Created the following packets: " << std::endl;
  //for (RtpiPacketVector::iterator ppos = rtpiPackets->begin(); ppos < rtpiPackets->end(); ++ppos)
  //  (*ppos)->Print();

  RtpiAduVector* rtpiAdus = CreateRtpiAdusFromPackets(rtpiPackets);

  // clean up
  delete rtpiPackets; // only vector is destroyed - the packets live until the corresponding adu is complete

  return rtpiAdus;
}


// create RTP/I ADUs from RTP/I packets (=ADU fragments); returns all ADUs that have been completed
RTPI::RtpiAduVector* RTPI::UnreliableTransport::CreateRtpiAdusFromPackets(RtpiPacketVector* packets) {
  //std::cout << "UnreliableTransport::CreateRtpiAdusFromPackets" << std::endl;

  if (packets->size() == 0) return NULL; // nothing to do

  RtpiAduVector* aduList = new RtpiAduVector;
  RefCount<RtpiPacket> packet;
  RtpiAdu* adu;

  for (RtpiPacketVector::iterator i = packets->begin(); i != packets->end(); ++i) {
    packet = *i;

    if ((packet->hdr->e == 1) && (packet->hdr->fc == 0)) {
      // packet contains one complete ADU -> create ADU
      adu = RtpiAdu::CreateRtpiAdu(packet->hdr->type);

      // create new buffer which contains RTP/I packet header  + RTP/I Adu Header + Payload
      adu->buf = new Buffer(packet->hdr->GetSize() + packet->GetPayloadSize());
      adu->packetHeaderStart = adu->buf->getCont(); // ! extension

      // create Adu Header
      adu->ParseHeader(packet->GetPayloadStart(), packet->hdr->type); // parse RTP/I ADU header
      adu->aduHeader->Flush(adu->packetHeaderStart + packet->hdr->GetSize()); // ! extension

      // set beginning of ADU Header in Buffer
      adu->aduHeaderStart = adu->packetHeaderStart + packet->hdr->GetSize(); // ! extension

      // set beginning of payload in Buffer
      if (adu->GetType() != RtpiHeader::PKT_STATE_QUERY) {
	adu->payload = adu->aduHeaderStart + adu->aduHeader->GetSize();
	adu->payloadSize = packet->GetPayloadSize() - adu->aduHeader->GetSize();
      }

      // copy RTP/I ADU Header and payload (if exists)
      memcpy(adu->aduHeaderStart, packet->GetPayloadStart(), packet->GetPayloadSize());

      // create single RTP/I packet header at beginning of RTP/I ADU
      adu->packetHeader = new RtpiHeader();
      adu->packetHeader->v = packet->hdr->v;
      adu->packetHeader->e = 1;
      adu->packetHeader->x = packet->hdr->x; // ! extension
      adu->packetHeader->type = packet->hdr->type;
      adu->packetHeader->p = 0;
      adu->packetHeader->rt = packet->hdr->rt;
      adu->packetHeader->length = adu->aduHeader->GetSize() + adu->payloadSize;
      adu->packetHeader->pid = packet->hdr->pid;
      adu->packetHeader->subid = packet->hdr->subid;
      adu->packetHeader->sn = packet->hdr->sn;
      adu->packetHeader->fc = 0;
      adu->packetHeader->Flush(adu->packetHeaderStart);

     // create RTP/I packet header extension ?

      aduList->push_back(adu);

      continue;
    }

    // packet contains one fragment of the ADU
    AduFragments* aduF = FindAduFragments(packet->hdr->subid, packet->hdr->type, packet->hdr->pid, packet->hdr->sn);
    if (aduF == NULL) {
      // first fragment of ADU -> generate new list
      aduF = InsertAduFragments(packet->hdr->subid, packet->hdr->type, packet->hdr->pid, packet->hdr->sn);
      aduF->fragments.push_front(packet);
    } else {
      // some fragments already there -> insert new fragment
			std::list<RefCount<RtpiPacket> >::iterator pos = aduF->fragments.begin();
      while (**pos < *packet)
	if (++pos == aduF->fragments.end()) break;

      aduF->fragments.insert(pos,packet);
      gettimeofday(&(aduF->time), NULL);  // update arrival time of last fragment

      // check if ADU complete
      if (AduComplete(aduF)) {
	// create ADU

				std::list<RefCount<RtpiPacket> >::iterator i = aduF->fragments.begin();
	RefCount<RtpiPacket> first = *i;
	adu = RtpiAdu::CreateRtpiAdu(first->hdr->type);
	int aduSize = GetSizeOfAdu(aduF);

	// create new buffer which contains RTP/I packet header  + RTP/I Adu Header + Payload
	adu->buf = new Buffer(first->hdr->GetSize() + aduSize); // ! extension
	adu->packetHeaderStart = adu->buf->getCont(); // ! extension

	// create Adu Header
	adu->ParseHeader(first->GetPayloadStart(), first->hdr->type); // parse RTP/I ADU header
	adu->aduHeader->Flush(adu->packetHeaderStart + first->hdr->GetSize()); // ! extension

	// set beginning of ADU Header in Buffer
	adu->aduHeaderStart = adu->packetHeaderStart + first->hdr->GetSize(); // ! extension

	// set beginning of payload in Buffer
	adu->payload = adu->aduHeaderStart + adu->aduHeader->GetSize();
	adu->payloadSize = aduSize - adu->aduHeader->GetSize();

	// create single RTP/I packet header at beginning of RTP/I ADU
	adu->packetHeader = new RtpiHeader();
	adu->packetHeader->v = first->hdr->v;
	adu->packetHeader->e = 1;
	adu->packetHeader->x = first->hdr->x; // ! extension
	adu->packetHeader->type = first->hdr->type;
	adu->packetHeader->p = 0;
	adu->packetHeader->rt = first->hdr->rt;
	adu->packetHeader->length = aduSize;
	adu->packetHeader->pid = first->hdr->pid;
	adu->packetHeader->subid = first->hdr->subid;
	adu->packetHeader->sn = first->hdr->sn;
	adu->packetHeader->fc = 0;
	adu->packetHeader->Flush(adu->packetHeaderStart);

	// create RTP/I packet header extension ?

	// copy RTP/I ADU Header and payload
	char* p = adu->aduHeaderStart;
	int bytesToCopy = first->GetPayloadSize();
	memcpy(p, first->GetPayloadStart(), bytesToCopy);
	p += bytesToCopy;

	++i;
	for ( ; i != aduF->fragments.end(); ++i) {
	  bytesToCopy = (*i)->GetPayloadSize();
	  memcpy(p, (*i)->GetPayloadStart(), bytesToCopy);
	  p += bytesToCopy;
	}

	aduList->push_back(adu);

	// clean up
	DeleteAduFragments(packet->hdr->subid, packet->hdr->type, packet->hdr->pid, packet->hdr->sn);

      }
    }
  }

  if (aduList->size() == 0) {
    // no comlete ADU's
    delete aduList;
    return NULL;
  } else
    return aduList;
}


bool RTPI::UnreliableTransport::AduComplete(RTPI::UnreliableTransport::AduFragments* adu) {
  int i = 0;
	std::list<RefCount<RtpiPacket> >::iterator p;

  for (p = adu->fragments.begin(); p != adu->fragments.end(); ++p) {
    if ((*p)->hdr->fc != i) return false; // gap discovered
    ++i;
  }

  p = adu->fragments.end();
  --p;
  if ((*p)->hdr->e != 1) return false; // tail not there

  return true;
}


int RTPI::UnreliableTransport::GetSizeOfAdu(RTPI::UnreliableTransport::AduFragments* adu) {
  int size = 0;

  for (std::list<RefCount<RtpiPacket> >::iterator i = adu->fragments.begin(); i != adu->fragments.end(); ++i)
    size += (*i)->GetPayloadSize();

  return size;
}


void RTPI::UnreliableTransport::Print(RTPI::UnreliableTransport::AduFragments* adu) {
  for (std::list<RefCount<RtpiPacket> >::iterator i = adu->fragments.begin(); i != adu->fragments.end(); ++i)
    (*i)->Print();
}


RTPI::UnreliableTransport::AduFragments* RTPI::UnreliableTransport::FindAduFragments(uint32 subid, uint8 type, uint32 pid, uint16 sn) {
  static bool firstCheck = true;
  static struct timeval lastCheckTime;
  static struct timeval currentTime;
  
  if (firstCheck) {
    gettimeofday(&lastCheckTime, NULL);
    firstCheck = false;
  }

  gettimeofday(&currentTime, NULL);

  if (lastCheckTime.tv_sec + lifetime < currentTime.tv_sec) {
    // traverse whole list searching the adu and doing timer check along
    AduFragments* adu = NULL;
    RefCount<RtpiPacket> first;
    for (std::list<AduFragments*>::iterator i = aduFragmentList.begin(); i != aduFragmentList.end(); ++i) {
      first = (*i)->fragments.front();
      if ((first->hdr->subid == subid) && (first->hdr->type == type) && (first->hdr->pid == pid) && (first->hdr->sn == sn)) {
	adu = *i;
      } else { // do timer check
	if ((*i)->time.tv_sec + lifetime < currentTime.tv_sec) {
	  // adu has not been completed in lifetime
	  delete *i;
	  aduFragmentList.erase(i);
	}
      }
    }

    gettimeofday(&lastCheckTime, NULL);

    return adu;
  } else {
    // searching adu only
    RefCount<RtpiPacket> first;
    for (std::list<AduFragments*>::iterator i = aduFragmentList.begin(); i != aduFragmentList.end(); ++i) {
      first = (*i)->fragments.front();
      if ((first->hdr->subid == subid) && (first->hdr->type == type) && (first->hdr->pid == pid) && (first->hdr->sn == sn))
	return *i;
    }

    return NULL; // not found
  }
}


RTPI::UnreliableTransport::AduFragments* RTPI::UnreliableTransport::InsertAduFragments(uint32 subid, uint8 type, uint32 pid, uint16 sn) {
  AduFragments* adu = new AduFragments;
  gettimeofday(&(adu->time), NULL);
  aduFragmentList.push_back(adu);
  return adu;
}



void RTPI::UnreliableTransport::DeleteAduFragments(uint32 subid, uint8 type, uint32 pid, uint16 sn) {
  RefCount<RtpiPacket> first;
  for (std::list<AduFragments*>::iterator i = aduFragmentList.begin(); i != aduFragmentList.end(); ++i) {
    first = (*i)->fragments.front();
    if ((first->hdr->subid == subid) && (first->hdr->type == type) && (first->hdr->pid == pid) && (first->hdr->sn == sn)) {
      delete *i;
      aduFragmentList.erase(i);
      return;
    }
  }
}



//
// Functions from class TransparentReliableTransport
//
RTPI::TransparentReliableTransport::~TransparentReliableTransport() {
  for (std::list<std::list<RefCount<RtpiPacket> >*>::iterator i = fragments.begin(); i != fragments.end(); ++i) {
    delete *i;
  }
}


// fragment a RTP/I ADU into one or more RTP/I packets; these packtes are flushed into the buffer given
// by the appliction; ! if this buffer is deleted, all RTP/I packets become obsolete and  must be destroyed
// if no header extension shall be attached, the packets are completely flushed;
//   however, the packet header can be changed and the packet flushed again (the payload stays in its assigned place
// if a header extension shall be attached, this has to be done manually by calling RtpiPacket::SetHeaderExtension for
//   each packet of the vector returend; afterwards all packets have to be flushed manually (call RtpiPacket::Flush)
//   the payload is already in it's correct position
RTPI::RtpiPacketVector* RTPI::TransparentReliableTransport::CreateRtpiPackets(RefCount<RtpiAdu> adu) {

  int aduSize = adu->GetAduSize();
  int reservedPacketSize = RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize();
  int maxRtpiPayloadSize = RtpiPacket::MAX_PKT_SIZE - reservedPacketSize;
  int rtpiPacketsNeeded  = (int)ceil((double)aduSize / maxRtpiPayloadSize);

  RtpiPacketVector* packets = new RtpiPacketVector();
  RefCount<RtpiPacket> packet;
  RtpiHeader* hdr;

  int totalAduBytesFlushed = 0;
  int bytesFlushed = 0;
  int fc = 0;

  while (totalAduBytesFlushed < aduSize) {
    // reserve a Buffer for each packet
    RefCount<Buffer> buf = new Buffer(RtpiPacket::MAX_PKT_SIZE);
    char* pos = buf->getCont();

    // create new RTP/I packet
    packet = new RtpiPacket(buf, pos);
    packets->push_back(packet);

    pos += (RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize());

    // write payload of RTP/I packet  = piece of RTP/I adu
    int bytesLeft = aduSize - totalAduBytesFlushed;
    int bytesToFlush = (bytesLeft <= maxRtpiPayloadSize) ? bytesLeft : maxRtpiPayloadSize;
    adu->Fragment(pos, totalAduBytesFlushed, bytesToFlush);

    // write RTP/I packet header
    packet->SetPayload(buf, pos, bytesToFlush);
    hdr = new RtpiHeader();
    hdr->e = (fc == rtpiPacketsNeeded - 1) ? 1 : 0;
    hdr->type = adu->GetType();
    hdr->rt = adu->packetHeader->rt;
    hdr->pid = adu->packetHeader->pid;
    hdr->subid = adu->packetHeader->subid;
    hdr->sn = adu->packetHeader->sn;
    hdr->fc = fc;
    packet->SetHeader(hdr);
    // if no header extension the packet is flushed completely; if extension, the packet has to be flushed again after
    // the extension has been attached
    bytesFlushed = packet->Flush();

    // resize buffer to number of bytes written; can be smaller than value above
    buf->setSize(bytesFlushed);

    // prepare for next round
    totalAduBytesFlushed += bytesToFlush;

    fc++;
  }

  return packets;
}


// fragment a RTP/I ADU into one or more RTP/I packets; these packtes are flushed into the buffer given
// by the appliction; ! if this buffer is deleted, all RTP/I packets become obsolete and  must be destroyed
// if no header extension shall be attached, the packets are completely flushed;
//   however, the packet header can be changed and the packet flushed again (the payload stays in its assigned place
// if a header extension shall be attached, this has to be done manually by calling RtpiPacket::SetHeaderExtension for
//   each packet of the vector returend; afterwards all packets have to be flushed manually (call RtpiPacket::Flush)
//   the payload is already in it's correct position
RTPI::RtpiPacketVector* RTPI::TransparentReliableTransport::CreateRtpiPackets(RefCount<RtpiAdu> adu, RefCount<Buffer> buf) {

  int aduSize = adu->GetAduSize();
  int reservedPacketSize = RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize() + RtpiPacket::GetPaddingBorder();
  // juergi: wozu ist hier GetPaddingBorder noetig? Stimmt das Ergebnis?
  int maxRtpiPayloadSize = RtpiPacket::MAX_PKT_SIZE - reservedPacketSize;
  int rtpiPacketsNeeded  = (int)ceil((double)aduSize / maxRtpiPayloadSize);

  // resize buffer
  buf->resize(aduSize + rtpiPacketsNeeded * reservedPacketSize);

  RtpiPacketVector* packets = new RtpiPacketVector();
  RefCount<RtpiPacket> packet;
  RtpiHeader* hdr;

  char* pos = buf->getCont();
  int totalBytesFlushed = 0;
  int totalAduBytesFlushed = 0;
  int bytesFlushed = 0;
  int fc = 0;

  while (totalAduBytesFlushed < aduSize) {
    // create new RTP/I packet
    packet = new RtpiPacket(buf, pos);
    packets->push_back(packet);

    pos += (RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize());

    // write payload of RTP/I packet  = piece of RTP/I adu
    int bytesLeft = aduSize - totalAduBytesFlushed;
    int bytesToFlush = (bytesLeft <= maxRtpiPayloadSize) ? bytesLeft : maxRtpiPayloadSize;
    adu->Fragment(pos, totalAduBytesFlushed, bytesToFlush);

    // write RTP/I packet header
    packet->SetPayload(buf, pos, bytesToFlush);
    hdr = new RtpiHeader();
    hdr->e = (fc == rtpiPacketsNeeded - 1) ? 1 : 0;
    hdr->type = adu->GetType();
    hdr->rt = adu->packetHeader->rt;
    hdr->pid = adu->packetHeader->pid;
    hdr->subid = adu->packetHeader->subid;
    hdr->sn = adu->packetHeader->sn;
    hdr->fc = fc;
    packet->SetHeader(hdr);
    // if no header extension the packet is flushed completely; if extension, the packet has to be flushed again after
    // the extension has been attached
    bytesFlushed = packet->Flush();
    totalBytesFlushed += bytesFlushed;

    // prepare for next round
    totalAduBytesFlushed += bytesToFlush;
    pos += (bytesFlushed - RtpiHeader::GetSize() - RtpiHeaderExtension::GetSize()); // payload plus padding ; rest has been taken care of above

    fc++;
  }

  // resize buffer to number of bytes written; can be smaller than value above
  buf->setSize(totalBytesFlushed);

  return packets;
}


// create RTP/I packets from a list of RTP/I ADUs: if a ADU is too large for one transport packet it is
// fragmented into several RTP/I packets; if not, only one RTP/I packet is created per ADU. If several
// RTP/I packets fit into one tranpsort packet, a compound packet is created containing a list of
// RTP/I packets which share the same buffer.
// ! if a buffer is deleted, all corresponding RTP/I packets become obsolete and  must be destroyed
// if no header extension shall be attached, the packets are completely flushed;
//   however, the packet header can be changed and the packet flushed again (the payload stays in its assigned place
// if a header extension shall be attached, this has to be done manually by calling RtpiPacket::SetHeaderExtension for
//   each packet of the vector returend; afterwards all packets have to be flushed manually (call RtpiPacket::Flush)
//   the payload is already in it's correct position
RTPI::RtpiCompoundPacketVector* RTPI::TransparentReliableTransport::CreateRtpiPackets(RefCount<RtpiAduVector> adus) {
  if ((!adus) || (adus->size() == 0))
    return NULL;

  int totalSize = 0;
  int aduSize = 0;
  int reservedPacketSize = RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize() + RtpiPacket::GetPaddingBorder();
  int maxRtpiPayloadSize = RtpiPacket::MAX_PKT_SIZE - reservedPacketSize;

  RefCount<Buffer> buf = new Buffer(RtpiPacket::MAX_PKT_SIZE);
  int bytesFlushedIntoBuffer = 0;
  RtpiCompoundPacketVector* compoundPackets = new RtpiCompoundPacketVector();
  RefCount<RtpiCompoundPacket> compoundPacket = new RtpiCompoundPacket(buf);
  compoundPackets->push_back(compoundPacket);

  char* pos = buf->getCont();
  RefCount<RtpiPacket> packet;
  RtpiHeader* hdr = NULL;

  for (RtpiAduVector::iterator i = adus->begin(); i != adus->end(); ++i) {
    int totalAduBytesFlushed = 0;
    int bytesFlushed = 0;
    aduSize = (*i)->GetAduSize();
    int rtpiPacketsNeeded = (int)ceil((double)aduSize / maxRtpiPayloadSize);
    int fc = 0;

    if (rtpiPacketsNeeded == 1) {
      //
      // 1. only one RTP/I packet is needed for the complete ADU
      //
      // create new RTP/I packet
      packet = new RtpiPacket(buf, pos);
      compoundPacket->AddPacket(packet);

      pos += (RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize());

      // write payload of RTP/I packet  = complete RTP/I ADU
      (*i)->Fragment(pos, 0, aduSize);

      // write RTP/I packet header
      packet->SetPayload(buf, pos, aduSize);
      hdr = new RtpiHeader();
      hdr->e = 1;
      hdr->type = (*i)->GetType();
      hdr->rt = (*i)->packetHeader->rt;
      hdr->pid = (*i)->packetHeader->pid;
      hdr->subid = (*i)->packetHeader->subid;
      hdr->sn = (*i)->packetHeader->sn;
      hdr->fc = 0;
      packet->SetHeader(hdr);
      // if no header extension the packet is flushed completely; if extension, the packet has to be flushed again after
      // the extension has been attached
      bytesFlushed = packet->Flush();
      bytesFlushedIntoBuffer += bytesFlushed;

      if ((i + 1) != adus->end()) { // there is at least one more ADU
	// prepare for next round
	if ((buf->getSize() - bytesFlushedIntoBuffer) < ((*(i+1))->GetAduSize() + reservedPacketSize)) {
	  // next ADU does not fit into the current buffer -> create new RTP/I Compound packet and new buffer
	  buf->setSize(bytesFlushedIntoBuffer);
	  buf = new Buffer(RtpiPacket::MAX_PKT_SIZE);
	  bytesFlushedIntoBuffer = 0;
	  pos = buf->getCont();
	  compoundPacket = new RtpiCompoundPacket(buf);
	  compoundPackets->push_back(compoundPacket);
	} else {
	  // next ADU does fit into the current buffer
	  pos += (bytesFlushed - RtpiHeader::GetSize() - RtpiHeaderExtension::GetSize()); // payload plus padding ; rest has been taken care of above
	}
      } else {
	// no more ADUs
	buf->setSize(bytesFlushedIntoBuffer);
      }
    } else {
      //
      // 2. the ADU has to be fragmented into several RTP/I packets
      //
      while (totalAduBytesFlushed < aduSize) {
	// create new RTP/I packet
	packet = new RtpiPacket(buf, pos);
	compoundPacket->AddPacket(packet);

	pos += (RtpiHeader::GetSize() + RtpiHeaderExtension::GetSize());

	// write payload of RTP/I packet  = piece of RTP/I adu
	int bytesLeft = aduSize - totalAduBytesFlushed;
	int bytesToFlush = (bytesLeft <= maxRtpiPayloadSize) ? bytesLeft : maxRtpiPayloadSize;
	(*i)->Fragment(pos, totalAduBytesFlushed, bytesToFlush);

	// write RTP/I packet header
	packet->SetPayload(buf, pos,bytesToFlush);
	hdr = new RtpiHeader();
	hdr->e = (fc == rtpiPacketsNeeded - 1) ? 1 : 0;
	hdr->type = (*i)->GetType();
	hdr->rt = (*i)->packetHeader->rt;
	hdr->pid = (*i)->packetHeader->pid;
	hdr->subid = (*i)->packetHeader->subid;
	hdr->sn = (*i)->packetHeader->sn;
	hdr->fc = fc;
	packet->SetHeader(hdr);
	// if no header extension the packet is flushed completely; if extension, the packet has to be flushed again after
	// the extension has been attached
	bytesFlushed = packet->Flush();
	buf->setSize(bytesFlushed);  // resize buffer to number of bytes written; can be smaller than value above

	// prepare for next round
	totalAduBytesFlushed += bytesToFlush;

	fc++;

	// create new compound packet and new buffer
	buf = new Buffer(RtpiPacket::MAX_PKT_SIZE);
	pos = buf->getCont();
	compoundPacket = new RtpiCompoundPacket(buf);
	compoundPackets->push_back(compoundPacket);
      }

      if ((i + 1) == adus->end()) { // there are no more ADUs
	// delete last compound packet
	compoundPackets->pop_back();
      }
    }
  }

  return compoundPackets;
}


// create RTP/I ADUs from RTP/I packets stored in buf
RTPI::RtpiAduVector* RTPI::TransparentReliableTransport::CreateRtpiAdusFromPackets(RefCount<Buffer> buf) {
  RtpiPacketVector* rtpiPackets = RtpiPacketFactory::CreateRtpiPackets(buf);

  if (rtpiPackets == NULL)
    throw RtpiException(RtpiException::ERR_RTPI_PARSE, "TransparentReliableTransport::CreateRtpiAdusFromPackets - Buffer corrupt or empty");

  //std::cout << "TransparentReliableTransport: Created the following packets: " << std::endl;
  //for (RtpiPacketVector::iterator ppos = rtpiPackets->begin(); ppos < rtpiPackets->end(); ++ppos)
  //  (*ppos)->Print();

  RtpiAduVector* rtpiAdus = CreateRtpiAdusFromPackets(rtpiPackets);

  // clean up
  delete rtpiPackets; // only vector is destroyed - the packets live until the corresponding adu is complete

  return rtpiAdus;
}


// create RTP/I ADUs from RTP/I packets (=ADU fragments); returns all ADUs that have been completed
RTPI::RtpiAduVector* RTPI::TransparentReliableTransport::CreateRtpiAdusFromPackets(RtpiPacketVector* packets) {
  //std::cout << "TransparentReliableTransport::CreateRtpiAdusFromPackets" << std::endl;

  if (packets->size() == 0) return NULL; // nothing to do

  RtpiAduVector* aduList = new RtpiAduVector;
  RefCount<RtpiPacket> packet;
  RtpiAdu* adu;

  for (RtpiPacketVector::iterator i = packets->begin(); i != packets->end(); ++i) {
    packet = *i;

    if ((packet->hdr->e == 1) && (packet->hdr->fc == 0)) {
      // packet contains one complete ADU -> create ADU
      adu = RtpiAdu::CreateRtpiAdu(packet->hdr->type);

      // create new buffer which contains RTP/I packet header  + RTP/I Adu Header + Payload
      adu->buf = new Buffer(packet->hdr->GetSize() + packet->GetPayloadSize());
      adu->packetHeaderStart = adu->buf->getCont(); // ! extension

      // create Adu Header
      adu->ParseHeader(packet->GetPayloadStart(), packet->hdr->type); // parse RTP/I ADU header
      adu->aduHeader->Flush(adu->packetHeaderStart + packet->hdr->GetSize()); // ! extension

      // set beginning of ADU Header in Buffer
      adu->aduHeaderStart = adu->packetHeaderStart + packet->hdr->GetSize(); // ! extension

      // set beginning of payload in Buffer
      if (adu->GetType() != RtpiHeader::PKT_STATE_QUERY) {
	adu->payload = adu->aduHeaderStart + adu->aduHeader->GetSize();
	adu->payloadSize = packet->GetPayloadSize() - adu->aduHeader->GetSize();
      }

      // copy RTP/I ADU Header and payload (if exists)
      memcpy(adu->aduHeaderStart, packet->GetPayloadStart(), packet->GetPayloadSize());

      // create single RTP/I packet header at beginning of RTP/I ADU
      adu->packetHeader = new RtpiHeader();
      adu->packetHeader->v = packet->hdr->v;
      adu->packetHeader->e = 1;
      adu->packetHeader->x = packet->hdr->x; // ! extension
      adu->packetHeader->type = packet->hdr->type;
      adu->packetHeader->p = 0;
      adu->packetHeader->rt = packet->hdr->rt;
      adu->packetHeader->length = adu->aduHeader->GetSize() + adu->payloadSize;
      adu->packetHeader->pid = packet->hdr->pid;
      adu->packetHeader->subid = packet->hdr->subid;
      adu->packetHeader->sn = packet->hdr->sn;
      adu->packetHeader->fc = 0;
      adu->packetHeader->Flush(adu->packetHeaderStart);

     // create RTP/I packet header extension ?

      aduList->push_back(adu);

      continue;
    }

    // packet contains one fragment of the ADU
		std::list<RefCount<RtpiPacket> >* fragmentList = FindFragmentList(packet->hdr->subid, packet->hdr->type, packet->hdr->pid, packet->hdr->sn);
    if (fragmentList == NULL) {
      // first fragment of ADU -> generate new list
      fragmentList = InsertFragmentList(packet->hdr->subid, packet->hdr->type, packet->hdr->pid, packet->hdr->sn);
      fragmentList->push_front(packet);
    } else {
      // some fragments already there -> insert new fragment
      std::list<RefCount<RtpiPacket> >::iterator pos = fragmentList->begin();
      while (**pos < *packet)
	if (++pos == fragmentList->end()) break;

      fragmentList->insert(pos,packet);

      // check if ADU complete
      if (AduComplete(fragmentList)) {
	// create ADU

				std::list<RefCount<RtpiPacket> >::iterator i = fragmentList->begin();
	RefCount<RtpiPacket> first = *i;
	adu = RtpiAdu::CreateRtpiAdu(first->hdr->type);
	int aduSize = GetSizeOfAdu(fragmentList);

	// create new buffer which contains RTP/I packet header  + RTP/I Adu Header + Payload
	adu->buf = new Buffer(first->hdr->GetSize() + aduSize); // ! extension
	adu->packetHeaderStart = adu->buf->getCont(); // ! extension

	// create Adu Header
	adu->ParseHeader(first->GetPayloadStart(), first->hdr->type); // parse RTP/I ADU header
	adu->aduHeader->Flush(adu->packetHeaderStart + first->hdr->GetSize()); // ! extension

	// set beginning of ADU Header in Buffer
	adu->aduHeaderStart = adu->packetHeaderStart + first->hdr->GetSize(); // ! extension

	// set beginning of payload in Buffer
	adu->payload = adu->aduHeaderStart + adu->aduHeader->GetSize();
	adu->payloadSize = aduSize - adu->aduHeader->GetSize();

	// create single RTP/I packet header at beginning of RTP/I ADU
	adu->packetHeader = new RtpiHeader();
	adu->packetHeader->v = first->hdr->v;
	adu->packetHeader->e = 1;
	adu->packetHeader->x = first->hdr->x; // ! extension
	adu->packetHeader->type = first->hdr->type;
	adu->packetHeader->p = 0;
	adu->packetHeader->rt = first->hdr->rt;
	adu->packetHeader->length = aduSize;
	adu->packetHeader->pid = first->hdr->pid;
	adu->packetHeader->subid = first->hdr->subid;
	adu->packetHeader->sn = first->hdr->sn;
	adu->packetHeader->fc = 0;
	adu->packetHeader->Flush(adu->packetHeaderStart);

	// create RTP/I packet header extension ?

	// copy RTP/I ADU Header and payload
	char* p = adu->aduHeaderStart;
	int bytesToCopy = first->GetPayloadSize();
	memcpy(p, first->GetPayloadStart(), bytesToCopy);
	p += bytesToCopy;

	++i;
	for ( ; i != fragmentList->end(); ++i) {
	  bytesToCopy = (*i)->GetPayloadSize();
	  memcpy(p, (*i)->GetPayloadStart(), bytesToCopy);
	  p += bytesToCopy;
	}

	aduList->push_back(adu);

	// clean up
	DeleteFragmentList(packet->hdr->subid, packet->hdr->type, packet->hdr->pid, packet->hdr->sn);

      }
    }
  }

  if (aduList->size() == 0) {
    // no comlete ADU's
    delete aduList;
    return NULL;
  } else
    return aduList;
}


bool RTPI::TransparentReliableTransport::AduComplete(std::list<RefCount<RtpiPacket> >* fragmentList) {
  int i = 0;
	std::list<RefCount<RtpiPacket> >::iterator p;

  for (p = fragmentList->begin(); p != fragmentList->end(); ++p) {
    if ((*p)->hdr->fc != i) return false; // gap discovered
    ++i;
  }

  p = fragmentList->end();
  --p;
  if ((*p)->hdr->e != 1) return false; // tail not there

  return true;
}


int RTPI::TransparentReliableTransport::GetSizeOfAdu(std::list<RefCount<RtpiPacket> >* fragmentList) {
  int size = 0;

  for (std::list<RefCount<RtpiPacket> >::iterator i = fragmentList->begin(); i != fragmentList->end(); ++i)
    size += (*i)->GetPayloadSize();

  return size;
}


void RTPI::TransparentReliableTransport::Print(std::list<RefCount<RtpiPacket> >* fragmentList) {
  for (std::list<RefCount<RtpiPacket> >::iterator i = fragmentList->begin(); i != fragmentList->end(); ++i)
    (*i)->Print();
}


std::list<RefCount<RTPI::RtpiPacket> >* RTPI::TransparentReliableTransport::FindFragmentList(uint32 subid, uint8 type, uint32 pid, uint16 sn) {
  RefCount<RtpiPacket> first;
  for (std::list<std::list<RefCount<RtpiPacket> >*>::iterator i = fragments.begin(); i != fragments.end(); ++i) {
    first = (*i)->front();
    if ((first->hdr->subid == subid) && (first->hdr->type == type) && (first->hdr->pid == pid) && (first->hdr->sn == sn))
      return *i;
  }

  return NULL; // not found
}


std::list<RefCount<RTPI::RtpiPacket> >* RTPI::TransparentReliableTransport::InsertFragmentList(uint32 subid, uint8 type, uint32 pid, uint16 sn) {
	std::list<RefCount<RtpiPacket> >* fragmentList = new std::list<RefCount<RtpiPacket> >;
  fragments.push_back(fragmentList);
  return fragmentList;
}


void RTPI::TransparentReliableTransport::DeleteFragmentList(uint32 subid, uint8 type, uint32 pid, uint16 sn) {
  RefCount<RtpiPacket> first;
  for (std::list<std::list<RefCount<RtpiPacket> >*>::iterator i = fragments.begin(); i != fragments.end(); ++i) {
    first = (*i)->front();
    if ((first->hdr->subid == subid) && (first->hdr->type == type) && (first->hdr->pid == pid) && (first->hdr->sn == sn)) {
      delete *i;
      fragments.erase(i);
      return;
    }
  }
}


