// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//
// Copyright (c) 2000 University of Mannheim. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for educational, research and non-profit purposes,
// without fee, and without a written agreement is hereby granted,
// provided that the above copyright notice, this paragraph and the
// following four paragraphs appear in all copies.
//
// Permission to incorporate this software into commercial products may
// be obtained from the following address:
// University of Mannheim
// Lehrstuhl fuer Praktische Informatik IV
// L15, 16
// 68161 Mannheim, Germany
//
// IN NO EVENT SHALL THE UNIVERSITY OF MANNHEIM BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
// ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF MANNHEIM HAS BEEN ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//
// THE UNIVERSITY OF MANNHEIM SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER
// IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF MANNHEIM HAS NO OBLIGATIONS
// TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The University of Mannheim will appreciate receiving any enhancements and
// modifications of this software made by others, and reserves the right to
// include them in future releases. The University also recommends that this
// software not be redistributed by third parties, but be obtained by all
// directly from the University's server, so that all users of the software
// may receive bug notifications and other relevant announcements.
//
// Authors:
//   Volker Hilt, Juergen Vogel, Martin Mauve
//   University of Mannheim
//   {hilt,vogel,mauve}@informatik.uni-mannheim.de
//
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#ifndef __RTPI_TRANSPORT_H__
#define __RTPI_TRANSPORT_H__

#ifdef WIN32
#  pragma warning (disable: 4786 4181)
#endif

#ifndef WIN32
#include <sys/time.h>
#endif

#include <list>

#ifdef WIN32
#include <util/refcount.cpp>
#else
#include <util/refcount.cc>
#endif

#include <util/Buffer.h>
#include <util/types.h>
#include <rtpi/rtpiPacket.h>


namespace RTPI {

  class Transport {
  public:
    Transport() {};
    virtual ~Transport() {};

    // handle outgoing packets
    virtual RtpiPacketVector* CreateRtpiPackets(RefCount<RtpiAdu> adu) = 0;
    virtual RtpiPacketVector* CreateRtpiPackets(RefCount<RtpiAdu> adu, RefCount<Buffer> buf) = 0;
    virtual RtpiCompoundPacketVector* CreateRtpiPackets(RefCount<RtpiAduVector> adus) = 0;

    // handle incoming packets
    virtual RtpiAduVector* CreateRtpiAdusFromPackets(RefCount<Buffer> buf) = 0;
    virtual RtpiAduVector* CreateRtpiAdusFromPackets(RtpiPacketVector* rtpiPackets) = 0;
  };

  
  class UnreliableTransport : public Transport { 
  public:
    UnreliableTransport() : lifetime(300) {} 
    ~UnreliableTransport();
    
    void SetFragmentLifetime(uint32 _lifetime) {lifetime = _lifetime;}  // define time in secs that  fragments of incomplete adus are stored
    uint32 GetFragmentLifeTime() {return lifetime;}
    void CheckFragmentLifetime();                                       // delete all fragments of adus that will not be completed anymore
    
    // handle outgoing packets
    RtpiPacketVector* CreateRtpiPackets(RefCount<RtpiAdu> adu);
    RtpiPacketVector* CreateRtpiPackets(RefCount<RtpiAdu> adu, RefCount<Buffer> buf);
    RtpiCompoundPacketVector* CreateRtpiPackets(RefCount<RtpiAduVector> adus);
    
    // handle incoming packets
    RtpiAduVector* CreateRtpiAdusFromPackets(RefCount<Buffer> buf);
    RtpiAduVector* CreateRtpiAdusFromPackets(RtpiPacketVector* rtpiPackets);
    
  private:
    uint32 lifetime;
    
    struct AduFragments {
      struct timeval time;  // arrival time of last fragment
      std::list<RefCount<RtpiPacket> > fragments;
    };
    
    std::list<AduFragments*> aduFragmentList;
    
    bool AduComplete(AduFragments* adu);
    int  GetSizeOfAdu(AduFragments* adu);
    
    void Print(AduFragments* adu);
    
    // manage lists of ADU fragments / RTP/I packets
    AduFragments* FindAduFragments(uint32 subid, uint8 type, uint32 pid, uint16 sn);
    AduFragments* InsertAduFragments(uint32 subid, uint8 type, uint32 pid, uint16 sn);
    void DeleteAduFragments(uint32 subid, uint8 type, uint32 pid, uint16 sn);
  };
  
#ifdef WIN32
  using std::list<RefCount<RtpiPacket> >;
  using std::list<UnreliableTransport::AduFragments*>;
#endif
  

  class TransparentReliableTransport : public Transport {
  public:
    TransparentReliableTransport() {}
    ~TransparentReliableTransport();

    // handle outgoing packets
    RtpiPacketVector* CreateRtpiPackets(RefCount<RtpiAdu> adu);
    RtpiPacketVector* CreateRtpiPackets(RefCount<RtpiAdu> adu, RefCount<Buffer> buf);
    RtpiCompoundPacketVector* CreateRtpiPackets(RefCount<RtpiAduVector> adus);

    // handle incoming packets
    RtpiAduVector* CreateRtpiAdusFromPackets(RefCount<Buffer> buf);
    RtpiAduVector* CreateRtpiAdusFromPackets(RtpiPacketVector* rtpiPackets);

  private:
    bool AduComplete(std::list<RefCount<RtpiPacket> >* fragmentList);
    int  GetSizeOfAdu(std::list<RefCount<RtpiPacket> >* fragmentList);

    void Print(std::list<RefCount<RtpiPacket> >* fragmentList);

    // manage lists of ADU fragments / RTP/I packets
    std::list<RefCount<RtpiPacket> >* FindFragmentList(uint32 subid, uint8 type, uint32 pid, uint16 sn);
    std::list<RefCount<RtpiPacket> >* InsertFragmentList(uint32 subid, uint8 type, uint32 pid, uint16 sn);
    void DeleteFragmentList(uint32 subid, uint8 type, uint32 pid, uint16 sn);

    std::list<std::list<RefCount<RtpiPacket> >*> fragments;
  };
}
  



#endif
