// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//
// Copyright (c) 2000 University of Mannheim. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for educational, research and non-profit purposes,
// without fee, and without a written agreement is hereby granted,
// provided that the above copyright notice, this paragraph and the
// following four paragraphs appear in all copies.
//
// Permission to incorporate this software into commercial products may
// be obtained from the following address:
// University of Mannheim
// Lehrstuhl fuer Praktische Informatik IV
// L15, 16
// 68161 Mannheim, Germany
//
// IN NO EVENT SHALL THE UNIVERSITY OF MANNHEIM BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
// ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF MANNHEIM HAS BEEN ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//
// THE UNIVERSITY OF MANNHEIM SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER
// IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF MANNHEIM HAS NO OBLIGATIONS
// TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The University of Mannheim will appreciate receiving any enhancements and
// modifications of this software made by others, and reserves the right to
// include them in future releases. The University also recommends that this
// software not be redistributed by third parties, but be obtained by all
// directly from the University's server, so that all users of the software
// may receive bug notifications and other relevant announcements.
//
// Authors:
//   Volker Hilt, Juergen Vogel, Martin Mauve
//   University of Mannheim
//   {hilt,vogel,mauve}@informatik.uni-mannheim.de
//
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#ifndef __RTPI_VIRTUAL__
#define __RTPI_VIRTUAL__

#ifdef WIN32
#  pragma warning (disable: 4786 4181)
#endif

#include <map>
#include <list>

#include <tcl.h>

#include <util/types.h>
#include <rtpi/rtpi.h>
//#include <rtpi/TclTimerManager.h>


namespace RTPI {
  class VirtualRtpi;

  class _Subcomponent {
  public:
    _Subcomponent(bool _active, std::string _name)
      : active(_active), name(_name) {}

    bool active;
    std::string name;
  };


  class RtcpiTimer {
  public:
    enum types {
      ADD_PARTICIPANT  = 0,
      REM_PARTICIPANT  = 1,
      ADD_SUBCOMPONENT = 2,
      REM_SUBCOMPONENT = 3,
      ACT_SUBCOMPONENT = 4,
      DEA_SUBCOMPONENT = 5,
      REPORT = 6
    };

    RtcpiTimer(int _type, uint32 _id, uint32 _runningTime, uint32 _sessionID, uint32 _pid)
      : type(_type), psid(_id), runningTime(_runningTime), sessionID(_sessionID), pid(_pid) {
      static uint32 idCounter = 0;
      id = idCounter++;
    }

    long token;
    int type;
    uint32 id;
    uint32 runningTime;
    uint32 psid;
    uint32 sessionID;
    uint32 pid;
  };

  
  class VirtualRtpi {
  public:
    VirtualRtpi();
    virtual ~VirtualRtpi();

    // group membership
    uint32 join_group(int sessionID, uint32 latency, std::string username, RtcpiReportInterface* rtcpiReport, 
		      RtpiFeedbackInterface* rtpiFeedback, Transport* rtpiTransport);
    void leave_group();
    bool group_joined();

    // manage participants
    //RtpiParticipant* get_own_participant();
    //RtpiParticipant* get_participant(uint32 pid);
    bool participant_exists(uint32 pid);
    uint32 get_number_of_participants();

    // manage subcomponents
    void add_subcomponent(uint32 subid, std::string name = "", bool active = true);
    void delete_subcomponent(uint32 subid);

    RtpiSubComponent* get_subcomponent(uint32 subid);
    bool subcomponent_exists(uint32 subid);
    uint32 get_number_of_subcomponents();
    
    void activate_subcomponent(uint32 subid);
    void deactivate_subcomponent(uint32 subid);
    bool subcomponent_is_active(uint32 subid);

    void set_subcomponent_name(uint32 subid, std::string name);
    std::string get_subcomponent_name(uint32 subid);
    bool subcomponent_has_name(uint32 subid);

    // helper functions
    void send_rtcpi_note(std::string note = "");
    uint32 get_current_time();

    // outgoing data
    RefCount<BufferVector> create_rtpi_packets(RefCount<RtpiAdu> adu);
    RefCount<BufferVector> create_rtpi_packets(RefCount<RtpiAduVector> adus);

    // incoming data
    RefCount<RtpiAduVector> create_rtpi_adus(RefCount<Buffer> buf);
    void handle_received_rtpi(RefCount<RtpiAduVector> adus, uint32 senderIpAddress);
    void handle_received_rtcpi(RefCount<Buffer> buf, uint32 senderIpAddress);
    
  private:
    static void ExecuteTimer(ClientData data);  

    RtcpiReportInterface* rtcpiReport;
    RtpiFeedbackInterface* rtpiFeedback;

    RtcpiPacketFactory* rtcpiPacketFactory;
    RtpiPacketFactory* rtpiPacketFactory;
    Transport* transport;

    int sessionID;
    bool joined;
    uint32 pid;
    uint32 transmissionTime;

    static std::map<int, bool> first; // sessionID, first report timer running
    static std::map<int, std::map<uint32, _Subcomponent*> > subcomponentList; // sessionID, subid, Subcomponent
    static std::map<int, std::map<uint32, VirtualRtpi*> > participantList; // sessionID, pid , VirtualRtpi
    static std::list<RtcpiTimer*> timerList;

    static const uint32 minInt;
    static const uint32 maxInt;
  };
}


#endif
