// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//
// Copyright (c) 2000 University of Mannheim. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for educational, research and non-profit purposes,
// without fee, and without a written agreement is hereby granted,
// provided that the above copyright notice, this paragraph and the
// following four paragraphs appear in all copies.
//
// Permission to incorporate this software into commercial products may
// be obtained from the following address:
// University of Mannheim
// Lehrstuhl fuer Praktische Informatik IV
// L15, 16
// 68161 Mannheim, Germany
//
// IN NO EVENT SHALL THE UNIVERSITY OF MANNHEIM BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
// ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF MANNHEIM HAS BEEN ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//
// THE UNIVERSITY OF MANNHEIM SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER
// IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF MANNHEIM HAS NO OBLIGATIONS
// TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The University of Mannheim will appreciate receiving any enhancements and
// modifications of this software made by others, and reserves the right to
// include them in future releases. The University also recommends that this
// software not be redistributed by third parties, but be obtained by all
// directly from the University's server, so that all users of the software
// may receive bug notifications and other relevant announcements.
//
// Authors:
//   Volker Hilt, Juergen Vogel, Martin Mauve
//   University of Mannheim
//   {hilt,vogel,mauve}@informatik.uni-mannheim.de
//
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#include <rtpi/rtpihandler.h>

#ifdef WIN32
#include <process.h>
#endif

#define xDEBUG0
#define xDEBUG1
#define xDEBUG2
#define xDEBUG3
#define xDEBUG4
#define xDEBUG5
#define xDEBUG8
#define xDEBUG_PARTICIPANT_LIST

#define DEBUG_ERRORS

// if WAIT_FOR_RTCPI is set, we wait for a valid rtcpi packet
// before we consider a site valid.
#define WAIT_FOR_RTCPI 1

// if WAIT_FOR_PROBABTION is set we use a probation interval of
// MIN_SEQUENTIAL rtpi packets in order before we consider a site valid.
#define WAIT_FOR_PROBATION 0


RTPI::RtpiHandler::RtpiHandler() : participants(), master_pos(0), reporter(0),
    psfi(0), average_rtcpi_packet_size(RTCPI_INITIAL_PACKET_SIZE),
    bye_counter(0),
    inactive_timer(Timer::newTimer("rtcpi_inactive_timer", this)),
    seq_space(new RtpiSubComponentSpace()),
    del_seq_space(true) {
  inactive_timer->msched(RTPI_INACTIVE_CHECK_INTERVAL);
 };


RTPI::RtpiHandler::RtpiHandler(RtcpiReportInterface* ri,
                               RtpiFeedbackInterface* _psfi,
			       bool in_only,
			       bool always_new_subids,
                               RtpiSubComponentSpace* pss) :
      participants(),
      master_pos(0), reporter(ri), psfi(_psfi),
      average_rtcpi_packet_size(RTCPI_INITIAL_PACKET_SIZE),
      bye_counter(0),
      inactive_timer(Timer::newTimer("rtcpi_inactive_timer", this)),
      seq_space((pss == NULL) ? new RtpiSubComponentSpace() : pss),
      del_seq_space(pss == NULL),
      report_incoming_only(in_only),
      always_use_new_subids(always_new_subids) {
  inactive_timer->msched(RTPI_INACTIVE_CHECK_INTERVAL);
};


RTPI::RtpiHandler::~RtpiHandler() {
  RtpiParticipantList::iterator pos;
  for (pos = participants.begin(); pos != participants.end(); ++pos) {
    delete (*pos);
  }
  participants.clear();

  if (inactive_timer)
    delete(inactive_timer);

  if (del_seq_space && (seq_space != NULL))
    delete seq_space;
};


void RTPI::RtpiHandler::createMasterParticipant(u_int32 local_addr, string user_name, bool deny_subcomp_reporting) {
  #ifdef DEBUG4
  std::cout << "RtpiHandler::createMasterParticipant()" << std::endl;
  #endif

  if (master_pos == 0) {
    RtpiParticipant* master = new RtpiParticipant(reporter, this, reporter->get_new_pid(), RTPI_MASTER);

    // initialize the master participant
    master->set_state(RTPI_VALID);
    master->set_addr(local_addr);
    master->set_cname(get_new_cname(user_name, local_addr));
    master->set_deny_subcomp_reporting(deny_subcomp_reporting);

    participants.push_back(master);       // append it to the end of the list!
    master_pos = participants.end();    // get its position
    --master_pos;
  };

  #ifdef DEBUG4
  std::cout << "RtpiHandler::createMasterParticipant() done." << std::endl;
  #endif
};



// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// * We use the algorithm described in RFC 1889 chapter 8.2 here to do
// * collision resolution and loop detection
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
RTPI::RtpiParticipantList::iterator RTPI::RtpiHandler::checkParticipant(rtpi_packet_direction direction,
							 bool& notify_part_created,
							 u_int32 pid,
							 u_int32 addr,
							 string cname) {

  #ifdef DEBUG4
  std::cout << "RtpiHandler::checkParticipant("<< pid <<") " << std::endl;
  #endif

  #ifdef DEBUG_PARTICIPANT_LIST
  {
  std::cout << "Participant list ......................................... " << std::endl;
  RtpiParticipantList::iterator pos;
  for (pos = participants.begin(); pos != participants.end(); ++pos) {
    std::cout << (*(*pos)) << std::endl;
    std::cout << ".......................................................... " << std::endl;
  }
  //timeval tmp = unixtime();
  //std::cout << "Timestamp NOW: " << tmp.tv_sec << "." << tmp.tv_usec << std::endl;
  //std::cout << "SubComponent list ........................................ " << std::endl;
  //std::cout << *seq_space << std::endl;
  std::cout << "List end ................................................. " << std::endl;
  }
  #endif

  // clean up participant list
  pack_members();

  // search for an existing participant using this pid
  RtpiParticipantList::iterator pos;
  for (pos = participants.begin(); pos != participants.end(); ++pos) {

    if (direction == RTPI_PACKET_OUTGOING) {
      // we're sending - use our alt_pid
      if ((*pos)->get_type() == RTPI_LOCAL) {
	if ((*pos)->get_alt_pid() == pid) {
          #ifdef DEBUG3
	  std::cout << "RtpiHandler::checkParticipant("<<pid<<") "
	       << "OUTGOING LOCAL FOUND" << std::endl;
	  #endif
	  break;
	};
      } else {
	if ((*pos)->get_type() == RTPI_MASTER) {
	  // master participant never uses alt_pid when sending
	  if ((*pos)->get_pid() == pid) {
            #ifdef DEBUG3
	    std::cout << "RtpiHandler::checkParticipant("<<pid<<") "
		 << "OUTGOING MASTER FOUND" << std::endl;
	    #endif
	    break;
	  }
	}
      }
      // remote outgoing can't happen since packets from a remote participant
      // can never be outgoing.
    } else {
      // we're receiving
      if ((*pos)->get_pid() == pid) {
        #ifdef DEBUG3
	std::cout << "RtpiHandler::checkParticipant("<<pid<<") "
	     << "INCOMING FOUND" << std::endl;
        #endif
	break;
      };
    };
  };

  // we have to add this participant
  if (pos == participants.end()) {

    // check weather we already have a site with this cname...
    pos = participants.begin();
    while ((pos != participants.end()) && ((*pos)->get_cname() != cname))
      ++pos;
    if (pos != participants.end()) {
      // cname found but with different pid! -> display a warning
      #ifdef DEBUG2
      std::cout << "RtpiHandler::checkParticipant(" << pid << ")"
	   << "Cname :" << cname << " already exists for pid: "
	   << (*pos)->get_pid() << std::endl
	   << "Warning: Creating participant with same cname!" << std::endl;
      #endif
    };

    // create a brand new participant!
    RtpiParticipant* participant;

    if (direction==RTPI_PACKET_OUTGOING) {
      // outgoing sites get a new pid for sending
      participant = new RtpiParticipant(reporter, this, reporter->get_new_pid(), RTPI_LOCAL);
      // set the cname for this site (if we already have one)
      if (!cname.empty()) {
	participant->set_cname(cname);
      }
      // set the addr for this site
      participant->set_addr(addr);
      // set the session note...
      participant->set_note(reporter->get_session_name());
      // we will still need the original pid to identify
      // the packets that we read from the file...
      participant->set_alt_pid(pid);
    } else {
      // incoming sites get the regular pid in pid and no alt_pid
      participant = new RtpiParticipant(reporter, this, pid, RTPI_REMOTE);
       // set the cname for this site (if we already have one)
      if (!cname.empty()) {
	participant->set_cname(cname);
      }
      // set the addr for this site
      participant->set_addr(addr);
      // set an alternate pid (for recording)
      // participant->set_alt_pid(reporter->get_new_pid());
      // not needed since we do not modify the pid during recording
    };

    // add the new guy to the list and get its position
    #ifdef DEBUG2
    std::cout << "RtpiHandler::checkParticipant("<<pid<<")"
	 << " adding participant... " << std::endl;
    #endif

    participants.push_back(participant);  // append it to the end of the list!
    pos = participants.end();             // get its position
    --pos;

    // inform the application about our new participant
    if (psfi != NULL)
      notify_part_created = true;

    #ifdef DEBUG2
    std::cout << "RtpiHandler::checkParticipant("<<pid<<") "
	 << "new participant added:" << std::endl
	 << "\tcname   : " << cname << std::endl
	 << "\tpid     : " << participant->get_pid() << std::endl;
    if (direction == RTPI_PACKET_OUTGOING) {
      std::cout << "\talt_pid : " << participant->get_alt_pid() << std::endl;
    }
    #endif

    return pos;

  } else {
    // pid found in list as participant pos

    // see if participant is marked as to be removed => ignore packets
    if (((*pos)->get_state() == RTPI_REMOVE) || ((*pos)->get_state() == RTPI_BYED)) {
      throw RtpiException(RtpiException::ERR_PARTICIPANT_TO_BE_REMOVED, "Participant is to be removed. Ignoring packet.");
    }

    // see if packet is from local participant
    if (reporter->addr_is_local(addr)) {
      if (direction == RTPI_PACKET_OUTGOING) {
	// outgoing packet should be ok; just return iterator
	return pos;
      } else {
	// if this is an incoming packet, always ignore packets from this
	// host (LOCAL or MASTER) if we generated them!
	if ((*pos)->get_type() != RTPI_REMOTE) {
          #ifdef DEBUG4
	  std::cout << "RtpiHandler::checkParticipant("<<pid<<") "
	       << "own packet ignored!" << std::endl;
          #endif
	  throw RtpiException(RtpiException::ERR_OWN_PARTICIPANT_IGNORED, "Ignoring own packets.");
	} else {
          #ifdef DEBUG3
	  std::cout << "RtpiHandler::checkParticipant("<<pid<<") "
	       << "packet from this host but not from us." << std::endl;
          #endif
	  return pos;
	};
      };
    }

    // address is not local
    if ((*pos)->get_addr() == addr) {
      // everything ok, participant already seen and addresses match
      #ifdef DEBUG3
      std::cout << "RtpiHandler::checkParticipant("<<pid<<") "
	   << "packet is ok." << std::endl;
      #endif
      return pos;

    } else {
      // we have a participant with same pid but different address
      // => loop!
      switch ((*pos)->get_type()) {
      case RTPI_MASTER:
      case RTPI_LOCAL:
	// it is one of our own pids!
        #ifdef DEBUG_ERRORS
	std::cout << "RtpiHandler::checkParticipant("<<pid<<") ";
        #endif
	if (cname.empty() || (*pos)->get_cname().empty()) {
          #ifdef DEBUG_ERRORS
	  std::cout << "Traffic loop or pid collision detected for local participant!" << std::endl;
          #endif
	  throw RtpiException(RtpiException::ERR_PARTICIPANT_REJECTED, "traffic loop or pid collision detected.");
	} else {
	  if (cname != (*pos)->get_cname()) {
            #ifdef DEBUG_ERRORS
	    std::cout << "Pid collision detected for local participant!" << std::endl;
            #endif
	    throw RtpiException(RtpiException::ERR_PARTICIPANT_REJECTED, "pid collision detected.");
	  } else {
            #ifdef DEBUG_ERRORS
	    std::cout << "Traffic loop detected for local participant!" << std::endl;
            #endif
	    throw RtpiException(RtpiException::ERR_PARTICIPANT_REJECTED, "traffic loop detected.");
	  }
	}
	break;

      default:
	// must be a third party, not really our business :-)
        #ifdef DEBUG_ERRORS
	std::cout << "RtpiHandler::checkParticipant("<<pid<<") "
	     << "third party loop detected (" << (*pos)->get_name() << ")!" << std::endl;
        #endif
	throw RtpiException(RtpiException::ERR_PARTICIPANT_REJECTED, "third party traffic loop detected.");
	break;
      }
    }
  }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
RTPI::IRtpiParticipantList RTPI::RtpiHandler::handle_rtcpi(RefCount<RTPI::RtcpiCompoundPacket> rtcpi, u_int32 addr, RTPI::rtpi_packet_direction direction, bool posthandle) {
  #ifdef DEBUG2
  std::cout << "RtpiHandler::handle_rtcpi() " << std::endl;
  #endif

  // include received rtcp packet size in average
  add_to_avg_rtcpi_size(rtcpi->GetBufferSize());

  IRtpiParticipantList parts;  // return list of pointers to participants in this packet
  RtpiParticipantList::iterator ppos;  // current participant

  bool notify_db_modification = false;
  bool notify_part_added = false;

  // proceed all packets in a compound packet individually!
  RtcpiPacketVector::iterator hpos = rtcpi->packets.begin();
  while (hpos != rtcpi->packets.end()) {

    // count byes for bye backoff
    if ((*hpos)->GetType() == RtcpiPacket::PKT_BYE) {
      bye_counter++;
    }

    bool del_hdr;
    try {

      del_hdr = false;

      switch((*hpos)->GetType()) {
      case RtcpiPacket::PKT_SDES:
	// handling of sdes packets is somewhat different
	// so treat them seperately
	ppos = handle_rtcpi_sdes(hpos, addr, direction, notify_part_added);

	// do timer calculation and set pids
	if (direction==RTPI_PACKET_INCOMING) {
	  (*ppos)->check_incoming_rtcpi(hpos);
	}
	if (direction == RTPI_PACKET_OUTGOING) {
	  (*ppos)->check_outgoing_rtcpi(hpos);
	}
	break;
      case RtcpiPacket::PKT_SUBREP:
	// handling of subrep packets is somewhat different
	// so treat them seperately
	ppos = handle_rtcpi_subrep(hpos, addr, direction, notify_db_modification, notify_part_added);
	break;
      default:
	// standard packet treatment

	// find the participant corresponding to this pid
	// and check for collisions and loops
	// exeption is thrown if packet is from us (via loopback)
	ppos = checkParticipant(direction, notify_part_added, (*hpos)->pid, addr);

	// do timer calculation and set pids
	if (direction==RTPI_PACKET_INCOMING) {
	  (*ppos)->check_incoming_rtcpi(hpos);
	} else {
	  // direction == RTPI_PACKET_OUTGOING
	  (*ppos)->check_outgoing_rtcpi(hpos);
	};

	// if desired do posthandling of packets here
	// (otherwise the application must call posthandle_rtcpi itself)
	if (posthandle) {
	  switch((*hpos)->GetType()) {
	  case RtcpiPacket::PKT_BYE:
	    // for bye packets we must remove our participants
	    posthandle_rtcpi_bye(hpos, direction, ppos);
	    break;
	  default:
	    break;
	  }
	}
	break;
      }

      // add valid participant
      parts.push_back(ppos);

      // report changes to application
      if (notify_part_added) {
	psfi->participant_added((*ppos)->get_pid());
      }

    } catch (RtpiException e) {
      #ifdef DEBUG0
      if (e.errCode != RtpiException::ERR_OWN_PARTICIPANT_IGNORED) {
	std::cout << "RtpiHandler::handle_rtcpi() packet rejected. Reason: ";
	std::cout << e.errMsg << " (" << e.errCode << ")" <<std::endl;
      }
      #endif
      // delete this header from the list
      del_hdr = true;
    }

    // next header
    if (del_hdr)
      hpos = rtcpi->DeletePacket(hpos);
    else
      ++hpos;
  }

  // clean up subcomponent list
  if (direction==RTPI_PACKET_OUTGOING) {
    // all components that don't have the local flag set are subject to timeout
    seq_space->release_nonlocal_subcomps();
  }
  // report changes to application
  if (notify_db_modification) {
    psfi->subcomponent_db_modified();
  }

  // see if we have found at least one valid participant
  if (parts.empty()) {
    throw RtpiException(RtpiException::ERR_PARTICIPANT_REJECTED, "No valid participant in RTCPI compound packet.");
  }
  // see if both lists are consistent (should always be ok ;) )
  if (parts.size() != rtcpi->packets.size()) {
    #ifdef DEBUG4
    std::cout << "RtpiHandler::handle_rtcpi() "
	 << "Inconsistent participant and packet list. Internal error." << std::endl;
    #endif
    throw RtpiException(RtpiException::ERR_PARTICIPANT_REJECTED, "Inconsistent participant list for RTCPI compound packet.");
  }

  return parts;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiHandler::posthandle_rtcpi(RefCount<RtcpiCompoundPacket> rtcpi, IRtpiParticipantList& parts, rtpi_packet_direction direction) {
  #ifdef DEBUG2
  std::cout << "RtpiHanlder::handle_rtcpi() " << std::endl;
  #endif

  // process all packets in a compound packet individually!
  IRtpiParticipantList::iterator ppos = parts.begin();
  RtcpiPacketVector::iterator hpos = rtcpi->packets.begin();
  while  ((hpos != rtcpi->packets.end()) && (ppos != parts.end())) {

    try {
      switch((*hpos)->GetType()) {
      case RtcpiPacket::PKT_BYE:
	// for bye packets we must remove our participants
	posthandle_rtcpi_bye(hpos, direction, (*ppos));
	break;
      default:
	break;
      }

    } catch (RtpiException e) {
      #ifdef DEBUG0
      if (e.errCode != RtpiException::ERR_OWN_PARTICIPANT_IGNORED) {
	std::cout << "RtpiHandler::handle_rtcpi() packet rejected. Reason: "
	     << e.errMsg << " (" << e.errCode << ")" << std::endl;
      }
      #endif
    }

    // next packet
    ++hpos;
    ++ppos;
  }
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiHandler::posthandle_rtcpi_bye(RtcpiPacketVector::iterator hdrpos, rtpi_packet_direction direction, RtpiParticipantList::iterator partpos) {

  #ifdef DEBUG2
  std::cout << "RtpiHandler::posthandle_rtcpi_bye() " << std::endl;
  #endif

  #ifdef DEBUG1
  // try to cast to the correct RtcpiByePacket type
  RtcpiByePacket* b_hdr = (RtcpiByePacket*)(&(**hdrpos));

  std::cout << "RtpiHandler::posthandle_rtcpi_bye() : Participant "
       << (*partpos)->pid << " is leaving. ";
  if (b_hdr->rlength != 0) {
    std::cout << "Reason: " << b_hdr->reason;
  }
  std::cout << std::endl;
  #endif

  // for an outgoing packet we must send a bye to all other participants
  // and we remove this participant immediately
  if (direction == RTPI_PACKET_OUTGOING) {
    #ifdef DEBUG3
    std::cout << "RtpiHandler::posthandle_rtcpi_bye(" << (*partpos)->pid << ") "
	 << "BYE packet => send BYE and delete from list..." << std::endl;
    #endif
    // the particpant is set to remove after the bye packet is sent
    (*partpos)->send_bye_and_remove("end of data");
  } else {
    // for incoming packets we keep the participant for a while but
    // set it as byed
    #ifdef DEBUG3
    std::cout << "RtpiHandler::posthandle_rtcpi_bye()" << (*partpos)->pid << ") "
	 << "received BYE packet => set it as byed..." << std::endl;
    #endif
    (*partpos)->set_state(RTPI_BYED);
  }

  // finally we must start the reverse timer reconsideration
  reverseTimerReconsiderations();
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
RTPI::RtpiParticipantList::iterator RTPI::RtpiHandler::handle_rtcpi_subrep(RtcpiPacketVector::iterator hpos, u_int32 addr, rtpi_packet_direction direction, bool& notify_db_change, bool& notify_part_added) {
  #ifdef DEBUG2
  std::cout << "RtpiHandler::handle_rtcpi_subrep(SUBREP RECEIVED)" << std::endl;
  #endif

  // try to cast to the correct RtcpiHdr_sr header type
  RtcpiSubcomponentReportPacket* subrep_hdr = (RtcpiSubcomponentReportPacket*)(&(**hpos));

  RtpiParticipantList::iterator ppos;  // current participant

  if (direction == RTPI_PACKET_INCOMING) {
    #ifdef DEBUG4
    std::cout << "RtpiHandler::handle_rtcpi_subrep(): Incoming packet (replace subids : " << (always_use_new_subids ? "on" : "off") << ")" << std::endl;
    #endif

    // validate participant
    ppos = checkParticipant(direction, notify_part_added, (*hpos)->pid, addr);

    // update sub component tables
    std::vector<RtcpiSubcomponentReportItem*>::iterator ipos = (subrep_hdr->GetItems())->begin();
    for (ipos; ipos != (subrep_hdr->GetItems())->end(); ++ipos) {
      #ifdef DEBUG4
      std::cout << "RtpiHandler::handle_rtcpi_subrep(): Processing item with id: " << (*ipos)->subid << std::endl;
      #endif

      // update component state
      SubComponentMap::iterator spos = seq_space->find_component((*ipos)->subid);
      if (spos == seq_space->get_components_end()) {
	// create a new subcomponent
	spos = seq_space->insert_component((*ipos)->subid,
					   (always_use_new_subids ? reporter->get_new_subid() : (*ipos)->subid));
	if (spos == seq_space->get_components_end()) {
	  // something went wrong
	  throw RtpiException(RtpiException::ERR_SUBCOMPONENT_INVALID, "Could not create subcomponent. Ignoring packet.");
	}
      }
      uint8 sub_change = spos->second.report_received((*ppos)->get_pid(), (*ipos)->active, (*ipos)->name, (*ipos)->length);
      // report the application about the changes
      if ((sub_change != RtpiSubComponent::SUBCOMP_OK) &&
	  (sub_change != RtpiSubComponent::SUBCOMP_NO_CHANGES)) {
        #ifdef DEBUG4
	std::cout << "RtpiHandler::handle_rtcpi_subrep(): sub component new or changed (" << (int)sub_change << ")." << std::endl;
        #endif
	if (psfi != NULL) {
	  notify_db_change = true;
	  if ((sub_change & RtpiSubComponent::SUBCOMP_INITIALIZED) != 0) {
	    psfi->subcomponent_added(&(*spos), RTPI::RtpiFeedbackInterface::PKT_RTCPI);
	  } else {
	    if (sub_change != RtpiSubComponent::SUBCOMP_OK) {
	      psfi->subcomponent_changed(&(*spos), sub_change);
	    }
	  }
	}
      }
    }

     // update participant state
    (*ppos)->check_incoming_rtcpi(hpos);

  } else {
    // direction == RTPI_PACKET_OUTGOING
    #ifdef DEBUG4
    std::cout << "RtpiHandler::handle_rtcpi_subrep(): Outgoing packet. Updating component db." << std::endl;
    #endif

    // we don't validate this participant since it is a dummy entry
    ppos = master_pos;

    // clear local flag of ALL subcomponents
    // if all packets are processed release_nonlocal_subcomps() must be called to
    // finally release the components that are not refreshed.
    seq_space->clear_subcomps_local();
    // reinstall local sub components
    std::vector<RtcpiSubcomponentReportItem*>::iterator ipos = (subrep_hdr->GetItems())->begin();
    for (ipos; ipos != (subrep_hdr->GetItems())->end(); ++ipos) {
      #ifdef DEBUG4
      std::cout << "RtpiHandler::handle_rtcpi_subrep(): Processing item with id: " << (*ipos)->subid << std::endl;
      #endif

      // find component in db (packet contains the alt subid)
      // Note: this is only useful for recorded packets. A regular participant
      // should never use outgoing rtcpi packets.
      SubComponentMap::iterator spos = seq_space->altsubid_find_component((*ipos)->subid);
      if (spos == seq_space->get_components_end()) {
	// create a new subcomponent
	spos = seq_space->insert_component(reporter->get_new_subid(), (*ipos)->subid);
	if (spos == seq_space->get_components_end()) {
	  // something went wrong
	  throw RtpiException(RtpiException::ERR_SUBCOMPONENT_INVALID, "Could not create subcomponent. Ignoring packet.");
	}
      }
      // Replace the subid with our new one
      (*ipos)->subid = spos->first;

      uint8 sub_change = spos->second.add_component((*ipos)->active, (*ipos)->name, (*ipos)->length);
      // report the application about the changes
      if ((sub_change != RtpiSubComponent::SUBCOMP_OK) &&
          (sub_change != RtpiSubComponent::SUBCOMP_NO_CHANGES)) {
        #ifdef DEBUG4
        std::cout << "RtpiHandler::handle_rtcpi_subrep(): sub component new or changed (" << (int)sub_change << ")." << std::endl;
        #endif
	if (!report_incoming_only && (psfi != NULL)) {
	  notify_db_change = true;
	  if ((sub_change & RtpiSubComponent::SUBCOMP_INITIALIZED) != 0) {
            psfi->subcomponent_added(&(*spos), RTPI::RtpiFeedbackInterface::PKT_RTCPI);
          } else {
	    if (sub_change != RtpiSubComponent::SUBCOMP_OK) {
	      psfi->subcomponent_changed(&(*spos), sub_change);
	    }
	  }
        }
      }
    }

     // update participant state
    (*ppos)->check_outgoing_rtcpi(hpos);

  }

  return ppos;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
RTPI::RtpiParticipantList::iterator RTPI::RtpiHandler::handle_rtcpi_sdes(RtcpiPacketVector::iterator hdrpos, u_int32 addr, rtpi_packet_direction direction, bool& notify_part_added) {

  #ifdef DEBUG2
  std::cout << "RtpiHandler::handle_rtcpi_sdes(SDES RECEIVED ("
       << (*hdrpos)->pid << "))" << std::endl;
  #endif

  // try to cast to the correct RtcpiHdr_Bye header type
  RtcpiSourceDescriptionPacket* sdes_hdr = (RtcpiSourceDescriptionPacket*)(&(**hdrpos));

  if (sdes_hdr == NULL)
    throw RtpiException(RtpiException::ERR_RTCPI_PACKET_FORMAT, "RTCP/I SDES expected.");

  string cname = "";
  string name = "";
  string email = "";
  string phone = "";
  string tool = "";
  string note = "";
  string loc = "";
  string priv = "";

  // seek the cname items
  std::vector<RtcpiSourceDescriptionItem*>::iterator ipos = sdes_hdr->items.begin();
  while  (ipos != sdes_hdr->items.end()) {

    // set the content according to the type of the sdes item
    switch ((*ipos)->GetType()) {
    case RtcpiSourceDescriptionItem::CNAME:
      if (direction == RTPI_PACKET_OUTGOING) {
	cname = create_new_cname(string((*ipos)->GetData(), (*ipos)->GetDataLength()));
      } else {
	cname = string((*ipos)->GetData(), (*ipos)->GetDataLength());
      }
      break;
    case RtcpiSourceDescriptionItem::NAME:
      name = string((*ipos)->GetData(), (*ipos)->GetDataLength());
      break;
    case RtcpiSourceDescriptionItem::EMAIL:
      email = string((*ipos)->GetData(), (*ipos)->GetDataLength());
      break;
    case RtcpiSourceDescriptionItem::PHONE:
      phone = string((*ipos)->GetData(), (*ipos)->GetDataLength());
      break;
    case RtcpiSourceDescriptionItem::LOC:
      loc = string((*ipos)->GetData(), (*ipos)->GetDataLength());
      break;
    case RtcpiSourceDescriptionItem::TOOL:
      tool = string((*ipos)->GetData(), (*ipos)->GetDataLength());
      break;
    case RtcpiSourceDescriptionItem::NOTE:
      note = string((*ipos)->GetData(), (*ipos)->GetDataLength());
      break;
    case RtcpiSourceDescriptionItem::PRIV:
      priv = string((*ipos)->GetData(), (*ipos)->GetDataLength());
      break;
    default:
      // unknown SDES item, ignored
      break;
    }
    ++ipos;
  }

  // find the participant corresponding to this pid
  RtpiParticipantList::iterator ppos = checkParticipant(direction, notify_part_added, sdes_hdr->pid, addr, cname);

  // set the participant infos
  bool changed = false;
  if (!cname.empty() && (cname != (*ppos)->get_plain_cname())) {
    (*ppos)->set_cname(cname);
    changed = true;
  }
  if (!name.empty() && (name != (*ppos)->get_plain_name())) {
    (*ppos)->set_name(name);
    changed = true;
  }
  if (!email.empty() && (email != (*ppos)->get_plain_email())) {
    (*ppos)->set_email(email);
    changed = true;
  }
  if (!phone.empty() && (phone != (*ppos)->get_plain_phone())) {
    (*ppos)->set_phone(phone);
    changed = true;
  }
  if (!tool.empty() && (tool != (*ppos)->get_plain_tool())) {
    (*ppos)->set_tool(tool);
    changed = true;
  }
  if (!note.empty() && (note != (*ppos)->get_plain_note())) {
    (*ppos)->set_note(note);
    changed = true;
  }
  if (!loc.empty() && (loc != (*ppos)->get_plain_loc())) {
    (*ppos)->set_loc(loc);
    changed = true;
  }
  if (!priv.empty() && (priv != (*ppos)->get_plain_priv())) {
    (*ppos)->set_priv(priv);
    changed = true;
  }

  // tell the application
  if ((direction == RTPI_PACKET_INCOMING) && changed && (psfi != NULL)) {
    psfi->participant_changed((*ppos)->get_pid());
  }

  #ifdef DEBUG2
  std::cout << "RtpiHandler::handle_rtcpi_sdes() done." << std::endl;
  #endif

  return ppos;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
RTPI::IRtpiParticipantList RTPI::RtpiHandler::handle_rtpi(RefCount<RtpiAduVector> rtpi,  u_int32 addr, rtpi_packet_direction direction, bool new_packets_out) {
  #ifdef DEBUG2
  std::cout << "RtpiHandler::handle_rtpi() " << std::endl;
  #endif

  IRtpiParticipantList parts;  // return list of pointers to participants
  RtpiParticipantList::iterator ppos;  // current participant

  bool notify_db_modification = false;
  bool notify_part_added = false;

  // now proceed all adus individually!
  RtpiAduVector::iterator hpos = rtpi->begin();
  while (hpos != rtpi->end()) {
    #ifdef DEBUG2
    std::cout << "RtpiHandler::handle_rtpi("
	 << ((direction==RTPI_PACKET_INCOMING)?"INCOMING":"OUTGOING") << ") "
	 << "PT: " << (int)(*hpos)->aduHeader->pt
	 << " PID: " << ((*hpos)->packetHeader->pid) << std::endl;
    #endif

    bool del_hdr;
    try {
      del_hdr = false;

      if ((direction == RTPI_PACKET_OUTGOING) && new_packets_out) {
	ppos = master_pos;
      } else {
	ppos = checkParticipant(direction, notify_part_added, (*hpos)->packetHeader->pid, addr);
      }

      // check and register seq. number and timestamps etc.
      if (direction == RTPI_PACKET_INCOMING) {
	// update component state
	SubComponentMap::iterator spos = seq_space->find_component((*hpos)->packetHeader->subid);
	if (spos == seq_space->get_components_end()) {
	  // create a new subcomponent
	  spos = seq_space->insert_component((*hpos)->packetHeader->subid,
					     (always_use_new_subids ? reporter->get_new_subid() : (*hpos)->packetHeader->subid));
	  if (spos == seq_space->get_components_end()) {
	    // something went wrong
	    throw RtpiException(RtpiException::ERR_SUBCOMPONENT_INVALID, "Could not create subcomponent. Ignoring packet.");
	  }
	}
	uint8 seq_ok = spos->second.data_received((*hpos)->aduHeader->pt);
	if ((seq_ok != RtpiSubComponent::SUBCOMP_OK) &&
	    (seq_ok != RtpiSubComponent::SUBCOMP_NO_CHANGES)) {
          #ifdef DEBUG4
	  std::cout << "RtpiHandler::handle_rtpi(): "
	       << "sub component new or changed (" << (int)seq_ok << ")." << std::endl;
          #endif
	  if (psfi != NULL) {
	    notify_db_modification = true;
	    if ((seq_ok & RtpiSubComponent::SUBCOMP_INITIALIZED) != 0) {
	      psfi->subcomponent_added(&(*spos), (*hpos)->packetHeader->GetType());
	    } else {
	      if (seq_ok != RtpiSubComponent::SUBCOMP_OK) {
		psfi->subcomponent_changed(&(*spos), seq_ok);
	      }
	    }
	  }
	}

	// update participant state
	(*ppos)->check_incoming_rtpi(*hpos);

      } else {
	// direction == RTPI_PACKET_OUTGOING

	// state queries are not treated as data!
	if ((*hpos)->packetHeader->GetType() != RtpiHeader::PKT_STATE_QUERY) {
	  // find component in db
	  SubComponentMap::iterator spos;
	  // are we allowed to substitute subids?
	  if (!new_packets_out) {
	    // packet contains the alt subid
	    spos = seq_space->altsubid_find_component((*hpos)->packetHeader->subid);
	    if (spos == seq_space->get_components_end()) {
	      // create a new subcomponent
	      spos = seq_space->insert_component(reporter->get_new_subid(), (*hpos)->packetHeader->subid);
	      if (spos == seq_space->get_components_end()) {
		// something went wrong
		throw RtpiException(RtpiException::ERR_SUBCOMPONENT_INVALID, "Could not create subcomponent. Ignoring packet.");
	      }
	    }
	    // Replace all subids containd in this rtpi adu
	    replace_rtpi_subids(*hpos, spos->first);
	  } else {
	    // packet already contains correct subid
	    spos = seq_space->find_component((*hpos)->packetHeader->subid);
	    if (spos == seq_space->get_components_end()) {
	      // create a new subcomponent
	      spos = seq_space->insert_component((*hpos)->packetHeader->subid,
						 (always_use_new_subids ? reporter->get_new_subid() : (*hpos)->packetHeader->subid));
	      if (spos == seq_space->get_components_end()) {
		// something went wrong
		throw RtpiException(RtpiException::ERR_SUBCOMPONENT_INVALID, "Could not create subcomponent. Ignoring packet.");
	      }
	    }
	  }

	  // update component state
	  uint8 seq_ok = spos->second.data_sent((*hpos)->aduHeader->pt);
	  if ((seq_ok != RtpiSubComponent::SUBCOMP_OK) &&
	      (seq_ok != RtpiSubComponent::SUBCOMP_NO_CHANGES)) {
            #ifdef DEBUG2
	    std::cout << "RtpiHandler::handle_rtpi(): "
		 << "Subcomponent db modified by sending data!" << std::endl;
            #endif
	    if (!report_incoming_only && (psfi != NULL)) {
	      notify_db_modification = true;
	      if ((seq_ok & RtpiSubComponent::SUBCOMP_INITIALIZED) != 0) {
		psfi->subcomponent_added(&(*spos), (*hpos)->packetHeader->GetType());
	      } else {
		if (seq_ok != RtpiSubComponent::SUBCOMP_OK) {
		  psfi->subcomponent_changed(&(*spos), seq_ok);
		}
	      }
	    }
	  }
	}

	// update participant state
	(*ppos)->check_outgoing_rtpi(*hpos);
      }

      // add valid participant to return value
      parts.push_back(ppos);

      // report changes to application
      if (notify_part_added) {
	psfi->participant_added((*ppos)->get_pid());
      }

    } catch (RtpiException e) {
      #ifdef DEBUG0
      if (e.errCode != RtpiException::ERR_OWN_PARTICIPANT_IGNORED) {
	std::cout << "RtpiHandler::handle_rtpi() packet rejected. Reason: ";
	std::cout << e.errMsg << " (" << e.errCode << ")" <<std::endl;
      }
      #endif
      // delete this header from the list
      del_hdr = true;
    }

    // next header
    if (del_hdr)
      hpos = rtpi->erase(hpos);
    else
      ++hpos;
  }

  // report changes to application
  if (notify_db_modification) {
    psfi->subcomponent_db_modified();
  }

  // see if we have found at least one valid participant
  if (parts.empty()) {
    throw RtpiException(RtpiException::ERR_PARTICIPANT_REJECTED, "No valid participant in RTCPI compound packet.");
  }
  // see if both lists are consistent (should always be ok ;) )
  if (parts.size() != rtpi->size()) {
    #ifdef DEBUG4
    std::cout << "RtpiHandler::handle_rtpi() "
	 << "Inconsistent participant and packet list. Internal error." << std::endl;
    #endif
    throw RtpiException(RtpiException::ERR_PARTICIPANT_REJECTED, "Inconsistent participant list for RTPI ADUs.");
  }

  #ifdef DEBUG2
  std::cout << "RtpiHandler::handle_rtpi() done." << std::endl;
  #endif

  return parts;
};


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiHandler::timerHandler(string id) {

  #ifdef DEBUG2
  std::cout << "RtpiHandler::timerHandler("<< id << ")" << std::endl;
  #endif

  if (id == "rtcpi_inactive_timer") {

    RtpiParticipantList::iterator pos = participants.begin();
    while (pos != participants.end()) {
      (*pos)->update_state();
      ++pos;
    }

    // reschedule the timer
    inactive_timer->msched(RTPI_INACTIVE_CHECK_INTERVAL);
  }

  #ifdef DEBUG2
  std::cout << "RtpiHandler::timerHandler("<< id << ") done." << std::endl;
  #endif
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiHandler::reverseTimerReconsiderations() {

  #ifdef DEBUG2
  std::cout << "RtpiHandler::reverseTimerReconsiderations()" << std::endl;
  #endif

  RtpiParticipantList::iterator pos = participants.begin();
  while (pos != participants.end()) {
    switch ((*pos)->get_type()) {
      // only RTPI_MASTER and RTPI_LOCALS issue reports
      case RTPI_MASTER:
      case RTPI_LOCAL:
	// and only if they are valid
	  (*pos)->reverseTimerReconsideration();
	break;
    }
    ++pos;
  }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiHandler::removeAllParticipants(string reason) {

  #ifdef DEBUG2
  std::cout << "RtpiHandler::removeAllParticipants()" << std::endl;
  #endif

  RtpiParticipantList::iterator pos = participants.begin();
  while (pos != participants.end()) {
    #ifdef DEBUG2
    std::cout << "RtpiHandler::removeAllParticipants() pid:"
	 << (*pos)->get_pid() << std::endl;
    #endif
    switch ((*pos)->get_type()) {
      case RTPI_MASTER:
      case RTPI_LOCAL:
	// say bye to the others
        #ifdef DEBUG2
        std::cout << "RtpiHandler::removeAllParticipants() sending bye." << std::endl;
        #endif
	// participant is set to remove if bye is sent
        (*pos)->send_bye_and_remove(reason);
	++pos;
	break;
      default:
	// remove partcipant from sequence number space
	seq_space->delete_participant_seq((*pos)->get_pid());
	// remove participant object
	delete (*pos);
	pos = participants.erase(pos);
	break;
    };
  };

  // we let the master participant live a little longer
  // but we do not consider it as master any more
  master_pos = 0;

  // clean up member list
  pack_members();

  #ifdef DEBUG2
  std::cout << "RtpiHandler::removeAllParticipants() done." << std::endl;
  #endif
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiHandler::removeLocalParticipants(string reason) {

  #ifdef DEBUG1
  std::cout << "RtpiHandler::removeLocalParticipants()" << std::endl;
  #endif

  RtpiParticipantList::iterator pos = participants.begin();
  while (pos != participants.end()) {
     switch ((*pos)->get_type()) {
      case RTPI_LOCAL:
        #ifdef DEBUG2
	std::cout << "RtpiHandler::removeLocalParticipants(): pid:"
	     << (*pos)->get_pid() << std::endl;
        #endif
	// say bye to the others
	// participant is set to remove if bye is sent
        (*pos)->send_bye_and_remove(reason);
	break;
      default:
        break;
     };
     ++pos;
  }

  // clean up member list
  pack_members();

  // try to fix timers
  reverseTimerReconsiderations();
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
RTPI::RtpiParticipantList::iterator RTPI::RtpiHandler::getParticipantsBegin() {
  return getNextParticipant(participants.begin());
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
RTPI::RtpiParticipantList::iterator RTPI::RtpiHandler::getParticipantsEnd() {
  return participants.end();
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
RTPI::RtpiParticipantList::iterator RTPI::RtpiHandler::getNextParticipant(const RTPI::RtpiParticipantList::iterator& spos) {
  RtpiParticipantList::iterator pos = spos;
  while ((pos != participants.end()) && ((*pos)->get_state() != RTPI_VALID)) {
    ++pos;
  }
  return pos;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
RTPI::RtpiParticipantList::iterator RTPI::RtpiHandler::findParticipant(uint32 pid) {
  RtpiParticipantList::iterator pos = participants.begin();
  while ((pos != participants.end()) &&
	 (((*pos)->get_state() != RTPI_VALID) || ((*pos)->get_pid() != pid))) {
    ++pos;
  }
  return pos;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiHandler::sendRtcpiNote() {
  #ifdef DEBUG2
  std::cout << "RtpiHandler::sendRtcpiNote()" << std::endl;
  #endif

  RtpiParticipantList::iterator pos;
  for (pos = participants.begin(); pos != participants.end(); ++pos) {
    switch ((*pos)->get_type()) {
    case RTPI_MASTER:
    case RTPI_LOCAL:
      (*pos)->send_report(RtcpiSourceDescriptionItem::NOTE);
      break;
    };
  };

  #ifdef DEBUG2
  std::cout << "RtpiHandler::sendRtcpiNote() done." << std::endl;
  #endif
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiHandler::pack_members() {
  #ifdef DEBUG4
  std::cout << "RtpiHandler::pack_members(): "
       << participants.size() << std::endl;
  #endif

  RtpiParticipantList::iterator pos = participants.begin();
  while (pos != participants.end()) {
    if ((*pos)->get_state() == RTPI_REMOVE) {
      #ifdef DEBUG4
      std::cout << "RtpHandler::pack_members(): Removing participant: "
	   << (*pos)->pid << std::endl;
      #endif
      if (pos == master_pos)
	master_pos = 0;
      // remove partcipant from sequence number space
      seq_space->delete_participant_seq((*pos)->get_pid());
      // remove participant object
      delete (*pos);
      pos = participants.erase(pos);
    } else {
      ++pos;
    }
  };

  #ifdef DEBUG4
  std::cout << "RtpiHandler::pack_members(): "
       << participants.size() << " done. " << std::endl;
  #endif
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
u_int32 RTPI::RtpiHandler::get_members() {
  #ifdef DEBUG4
  std::cout << "RtpiHandler::get_members(): "
       << participants.size();
  #endif

  int count = 0;
  RtpiParticipantList::iterator pos = participants.begin();
  while (pos != participants.end()) {
      // count only valid sites
    if ((*pos)->get_state() == RTPI_VALID) {
      ++count;
    }
    ++pos;
  };

  #ifdef DEBUG4
  std::cout << " valid: " << count << std::endl;
  #endif

  return count;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiHandler::reset_transmission(bool include_master) {
  #ifdef DEBUG1
  std::cout << "RtpiHandler::reset_transmission()" << std::endl;
  #endif

  RtpiParticipantList::iterator pos;
  for (pos = participants.begin(); pos != participants.end(); ++pos) {
    if (((*pos)->get_type() == RTPI_LOCAL) ||
	(include_master && ((*pos)->get_type() == RTPI_MASTER))) {
      (*pos)->reset_transmission();
    };
  };
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiHandler::reset_reception() {
  #ifdef DEBUG1
  std::cout << "RtpiHandler::reset_reception()" << std::endl;
  #endif

  RtpiParticipantList::iterator pos;
  for (pos = participants.begin(); pos != participants.end(); ++pos) {
    if ((*pos)->get_type() == RTPI_REMOTE) {
      (*pos)->reset_transmission();
    };
  };
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
string RTPI::RtpiHandler::create_new_cname(string old_cname) {
  in_addr a;
  a.s_addr = htonl(reporter->get_local_address());
  string cname = "";
  int p = old_cname.find("-recorded");
  if (p != string::npos) {
    cname = old_cname.substr(0, p) + "-recorded-" + itos(getpid())
      + "@" + string(inet_ntoa(a));
  } else {
    p = old_cname.find("@");
    if (p != string::npos) {
      cname = old_cname.substr(0, p) + "-recorded-" + itos(getpid())
	+ "@";
    } else {
      cname = "recorded-" + itos(getpid()) + "@";
    }
    cname += string(inet_ntoa(a));
  };
  return cname;
};

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
string RTPI::RtpiHandler::get_new_cname(string user, u_int32 i) {
  struct in_addr a;
  if (i!=0) {
    a.s_addr=htonl(i);
  } else {
    a.s_addr=htonl(get_local_addr());
  };
  return user + "@" + string(inet_ntoa(a));
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
RTPI::RtcpiCompoundPacket* RTPI::RtpiHandler::create_participant_db_entry(RtpiParticipantList::iterator pos) {
  #ifdef DEBUG2
  std::cout << "RtpiHandler::create_participant_db_entry() "
       << "build sdes infos for: " << (*pos)->get_pid() << std::endl;
  #endif

  // create a new compound packet
  RtcpiCompoundPacket* rtcpi = new RtcpiCompoundPacket(RefCount<Buffer>(new Buffer(4096)), RtcpiCompoundPacket::UNLIMITED_PKT_SIZE);

  // create a new sdes packet
  RtcpiSourceDescriptionPacket* sdes_packet = new RtcpiSourceDescriptionPacket((*pos)->pid);

  // build list of sdes items
  int sdes_item_list[5];
  sdes_item_list[0] = RtcpiSourceDescriptionItem::CNAME;
  sdes_item_list[1] = RtcpiSourceDescriptionItem::NAME;
  sdes_item_list[2] = RtcpiSourceDescriptionItem::EMAIL;
  sdes_item_list[3] = RtcpiSourceDescriptionItem::PHONE;
  sdes_item_list[4] = RtcpiSourceDescriptionItem::TOOL;

  // build items
  for (int i = 0; i < 5; i++) {
    RtcpiSourceDescriptionItem* tmp_item = NULL;
    try {
      tmp_item = (*pos)->build_sdes_item(sdes_item_list[i]);
    } catch (RtpiException e) {
      #ifdef DEBUG_1
      std::cout << "RtpiHandler()::create_participant_db_entry(): could not build item. Reason: "
	   << e.errMsg << " (" << e.errCode << ")" << std::endl;
      #endif
      if (tmp_item != NULL) {
	delete tmp_item;
	tmp_item = NULL;
      }
    }
    // add item to packet
    if (tmp_item != NULL)  {
      try {
	sdes_packet->AddItem(tmp_item, rtcpi->GetMaxSize());
      } catch (RtpiException e) {
	// current sdes packet is full. create a new one.
	// store old packet in compound packet
	if (sdes_packet->ContainsItems()) {
	  try {
	    rtcpi->AddPacket(sdes_packet);
	  } catch (RtpiException e) {
	    // current compound packet is full. something is wrong here.
            #ifdef DEBUG_ERRORS
	    std::cout << "RtpiHandler::create_participant_db_entry(): Could not add packet to compound packet. Discarding sdes packet. Reason: "
		 << e.errMsg << " (" << e.errCode << ")" << std::endl;
            #endif
	    delete sdes_packet;
	  }
	  // create new sdes packet
	  sdes_packet = new RtcpiSourceDescriptionPacket((*pos)->pid);
	}
	// try again with the new sdes packet
	try {
	  sdes_packet->AddItem(tmp_item, rtcpi->GetMaxSize());
	} catch (RtpiException e) {
	  // something is really wrong here.
          #ifdef DEBUG_ERRORS
	  std::cout << "RtpiHandler::create_participant_db_entry(): Could not add item to sdes packet. Discarding item. Reason: "
	       << e.errMsg << " (" << e.errCode << ")" << std::endl;
          #endif
	  delete tmp_item;
	}
      }
    } else {
      // no item created
      if (i == 0) {
	// we must have a cname!
	break;
      }
    }
  }

  // see if we have components to report
  if (!sdes_packet->ContainsItems()) {
    delete sdes_packet;
  } else {
    // add our last sdes packet to compound packet
    try {
      rtcpi->AddPacket(sdes_packet);
    } catch (RtpiException e) {
      // something is really wrong here.
      #ifdef DEBUG_ERRORS
      std::cout << "RtpiHandler()::create_participant_db_entry(): Could not add packet to compound packet. Discarding sdes packet. Reason: "
	   << e.errMsg << " (" << e.errCode << ")" << std::endl;
      #endif
      delete sdes_packet;
    }
  }

  // flush our compound packet
  if (!rtcpi->ContainsPackets()) {
    delete rtcpi;
    #ifdef DEBUG2
    std::cout << "RtpiHandler()::create_participant_db_entry(): No sdes items to report." << std::endl;
    #endif
    return NULL;
  } else {
    try {
      rtcpi->Flush();
    } catch (RtpiException e) {
      #ifdef DEBUG2
      std::cout << "RtpiHanlder()::create_participant_db_entry(): Could not flush packets. Reason: " << e.errMsg << " (" << e.errCode << ")" << std::endl;
      #endif
      delete rtcpi;
      return NULL;
    }

    #ifdef DEBUG2
    std::cout << "RtpiHanlder()::create_participant_db_entry(): Report created.";
    rtcpi->Print();
    #endif

    return rtcpi;
  }
}



// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
RTPI::RtcpiCompoundPacket* RTPI::RtpiHandler::create_subcomponent_db_entry() {
  #ifdef DEBUG2
  std::cout << "RtpiHandler::create_subcomp_db_entry() " << std::endl;
  #endif

  // create a brand new RTCPI compound packet
  RtcpiCompoundPacket* rtcpi = new RtcpiCompoundPacket(RefCount<Buffer>(new Buffer(4096)), RtcpiCompoundPacket::UNLIMITED_PKT_SIZE);
  RtcpiSubcomponentReportPacket* subrep_packet = new RtcpiSubcomponentReportPacket((*master_pos)->pid, RtcpiSubcomponentReportPacket::EMPTY_ALLOWED);

  // create an item for each (active and passive!) sub-component
  SubComponentMap::iterator spos = seq_space->get_components_begin();
  while (spos != seq_space->get_components_end()) {
    RtcpiSubcomponentReportItem* tmp_item = NULL;
    try {
      tmp_item = new RtcpiSubcomponentReportItem(spos->first,
						 spos->second.is_active(),
						 spos->second.get_name(),
						 spos->second.get_name_length());

    } catch (RtpiException e) {
      #ifdef DEBUG_ERRORS
      std::cout << "RtpiHandler()::create_subcomponent_db_entry(): could not build item. Reason: "
	   << e.errMsg << " (" << e.errCode << ")" << std::endl;
      #endif
      if (tmp_item != NULL) {
	delete tmp_item;
	tmp_item = NULL;
      }
    }

    // add item to packet
    if (tmp_item != NULL)  {
      try {
	subrep_packet->AddItem(tmp_item, rtcpi->GetMaxSize());
      } catch (RtpiException e) {
	// current subrep packet is full. create a new one.
	// store old packet in compound packet
	if (subrep_packet->ContainsItems()) {
	  try {
	    rtcpi->AddPacket(subrep_packet);
	  } catch (RtpiException e) {
	    // current compound packet is full?! something is wrong here!
            #ifdef DEBUG_ERRORS
	    std::cout << "RtpiHandler()::create_subcomp_db_entry(): Could not add packet to compound packet. Discarding subrep packet. Reason: "
		 << e.errMsg << " (" << e.errCode << ")" << std::endl;
            #endif
	    delete subrep_packet;
	  }
	  // create new sdes packet
	  subrep_packet = new RtcpiSubcomponentReportPacket((*master_pos)->pid);
	}
	// try again with the new sdes packet
	try {
	  subrep_packet->AddItem(tmp_item, rtcpi->GetMaxSize());
	} catch (RtpiException e) {
	  // something is really wrong here.
          #ifdef DEBUG_ERRORS
	  std::cout << "RtpiHandler()::create_subcomp_db_entry(): Could not add item to subrep packet. Discarding item. Reason: "
	       << e.errMsg << " (" << e.errCode << ")" << std::endl;
          #endif
	  delete tmp_item;
	}
      }
    }

    ++spos;
  }

  // see if we have components to report (one single empty packet is allowed!)
  if (rtcpi->ContainsPackets() && !subrep_packet->ContainsItems()) {
    delete subrep_packet;
  } else {
    // add our last subrep packet to compound packet
    try {
      rtcpi->AddPacket(subrep_packet);
    } catch (RtpiException e) {
      // current compound packet is full?! something is wrong here!
      #ifdef DEBUG_ERRORS
      std::cout << "RtpiHandler()::create_subcomp_db_entry(): Could not add packet to compound packet. Discarding subrep packet. Reason: "
	   << e.errMsg << " (" << e.errCode << ")" << std::endl;
      #endif
      delete subrep_packet;
    }
  }

  if (!rtcpi->ContainsPackets()) {
    delete rtcpi;
    return NULL;
  } else {
    try {
      rtcpi->Flush();
    } catch (RtpiException e) {
      #ifdef DEBUG2
      std::cout << "RtpiHandler()::create_subcomponent_db_entry(): Could not flush packets. Reason: " << e.errMsg << " (" << e.errCode << ")" << std::endl;
      #endif
      delete rtcpi;
      return NULL;
    }

    #ifdef DEBUG2
    std::cout << "RtpiHandler()::create_subcomponent_db_entry(): Report created.";
    rtcpi->Print();
    #endif

    return rtcpi;
  }
}



// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
uint32 RTPI::RtpiHandler::get_current_timestamp() {
  return rtpitime();
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// *
// *
// *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiHandler::replace_rtpi_subids(RefCount<RtpiAdu> rtpi, uint32 new_subid) {
  // this function is only called for OUTGOING packets if new_packets_out = false

  // replace the subid in the packet header
  rtpi->packetHeader->subid = new_subid;

  // replace all subids contained in the rtpi adu
  switch (rtpi->packetHeader->GetType()) {
  case RtpiHeader::PKT_STATE:
  case RtpiHeader::PKT_DELTA_STATE: {
    // replace object subid in state base header
    RtpiStateBaseHeader* s_hdr = (RtpiStateBaseHeader*)rtpi->aduHeader;
    if (s_hdr->ObjectIDIsSet()) {
      std::pair<bool,uint32> res = seq_space->altsubid_to_subid(s_hdr->GetObjectID());
      if (!res.first) {
	// referred object does not exist.
	// create a new subcomponent
	SubComponentMap::iterator spos = seq_space->insert_component(reporter->get_new_subid(), s_hdr->GetObjectID());
	if (spos == seq_space->get_components_end()) {
	  // something went wrong
	  #ifdef DEBUG1
	  std::cout << "RtpiHandler::replace_rtpi_subids(): WARNING: Could not create referred subcomponent. "
	       << " Leaving id unchanged! " << std::endl;
	  #endif
	  return;
	}
	res.second = spos->first;
      }
      s_hdr->SetObjectID(res.second);
    }
    break;
  }
  case RtpiHeader::PKT_EVENT: {
    // replace object subid in state base header
    RtpiEventHeader* e_hdr = (RtpiEventHeader*)rtpi->aduHeader;
    if (e_hdr->ObjectIDIsSet()) {
      std::pair<bool,uint32> res = seq_space->altsubid_to_subid(e_hdr->GetObjectID());
      if (!res.first) {
	// referred object does not exist.
	// create a new subcomponent
	SubComponentMap::iterator spos = seq_space->insert_component(reporter->get_new_subid(), e_hdr->GetObjectID());
	if (spos == seq_space->get_components_end()) {
	  // something went wrong
	  #ifdef DEBUG1
	  std::cout << "RtpiHandler::replace_rtpi_subids(): WARNING: Could not create referred subcomponent. "
	       << " Leaving id unchanged! " << std::endl;
	  #endif
	  return;
	}
	res.second = spos->first;
      }
      e_hdr->SetObjectID(res.second);
    }
    break;
  }
  }
}
