// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//
// Copyright (c) 2000 University of Mannheim. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for educational, research and non-profit purposes,
// without fee, and without a written agreement is hereby granted,
// provided that the above copyright notice, this paragraph and the
// following four paragraphs appear in all copies.
//
// Permission to incorporate this software into commercial products may
// be obtained from the following address:
// University of Mannheim
// Lehrstuhl fuer Praktische Informatik IV
// L15, 16
// 68161 Mannheim, Germany
//
// IN NO EVENT SHALL THE UNIVERSITY OF MANNHEIM BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
// ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF MANNHEIM HAS BEEN ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//
// THE UNIVERSITY OF MANNHEIM SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER
// IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF MANNHEIM HAS NO OBLIGATIONS
// TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The University of Mannheim will appreciate receiving any enhancements and
// modifications of this software made by others, and reserves the right to
// include them in future releases. The University also recommends that this
// software not be redistributed by third parties, but be obtained by all
// directly from the University's server, so that all users of the software
// may receive bug notifications and other relevant announcements.
//
// Authors:
//   Volker Hilt, Juergen Vogel, Martin Mauve
//   University of Mannheim
//   {hilt,vogel,mauve}@informatik.uni-mannheim.de
//
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#ifndef _RtpiHandler_h_
#define _RtpiHandler_h_

#ifdef WIN32
#  pragma warning (disable: 4786)
#endif


#ifdef WIN32
#include <winsock.h>
#else
#include <netinet/in.h>
#endif

#include <map>
#include <list>
#include <iterator>
#include <string>

#include <util/types.h>
#include <util/strtools.h>
#include <util/tools.h>
#include <util/timerinterface.h>
#include <rtpi/rtpiparticipant.h>
#include <rtpi/rtpiPacket.h>
#include <rtpi/rtcpiPacket.h>
#include <rtpi/rtpiException.h>
#include <rtpi/rtpiSeqNumber.h>
#include <rtpi/rtpiInterfaces.h>

/*
 * Gain (smoothing constant) for the low-pass filter that estimates
 * the average rtcpi packet size.
 */
#define RTCPI_SIZE_GAIN (1./16.)
/*
 * Initial rtcpi packet size used as starting point for the
 * average rtcpi packet size estimate. Should correspond roughly
 * to the size of the first rtcpi packet that will be sent.
 *
 * 128 roughly assumes a compound packet consisting of
 * one SUBREP (1 sub component)
 * and one SDES with two reasonably sized items.
 */
#define RTCPI_INITIAL_PACKET_SIZE 128
/*
 * How often will we check weather a site changed its state (time in ms)
 */
#define RTPI_INACTIVE_CHECK_INTERVAL 2000

namespace RTPI {
  enum rtpi_packet_direction { RTPI_PACKET_INCOMING, RTPI_PACKET_OUTGOING };

  typedef std::list<RtpiParticipant*> RtpiParticipantList;
  typedef std::list<RtpiParticipantList::iterator> IRtpiParticipantList;

  class RtpiHandler : public TimerInterface {
    friend class RtpiParticipant;

  private:
    RtpiParticipantList participants;
    RtpiParticipantList::iterator master_pos;

    RtcpiReportInterface* reporter;
    // the pointer to the RtpiFeedbackInterface to inform
    // applications about changes in subcomponents
    RtpiFeedbackInterface* psfi;

    int		          average_rtcpi_packet_size;

    unsigned int          bye_counter;

    // the timer checking participant states
    P2Timer	          inactive_timer;

    // SubComponentSpace is not deleted if given as param to RtpiHander()
    RtpiSubComponentSpace* seq_space;
    bool                   del_seq_space;

    bool                   report_incoming_only;
    bool                   always_use_new_subids;

  public:
    RtpiHandler();
    RtpiHandler(RtcpiReportInterface*,
                RtpiFeedbackInterface* = NULL,
		bool = true,               // true: report chages about incoming subcomps only
		bool = false,              // true: always use new subids for outgoing packets
                RtpiSubComponentSpace* = NULL);

    virtual ~RtpiHandler();

    // process rtpi and rtcpi adus

    // Processes a compound packet and updates participant infos
    // Returns a list of participants for which
    // a valid packet is contained in the compound packet.
    // posthandle - false: No participants are deleted from the participant
    //                     list (even if a bye packet is processed).
    //                     This enables an application to process rtcpi packets
    //                     having all required participants available.
    //                     posthandle_rtcpi MUST be called after processing the
    //                     packet.
    //            - true: posthandle_rtcpi is done automatically.
    IRtpiParticipantList handle_rtcpi(RefCount<RtcpiCompoundPacket> rtcpi, u_int32 addr,
				  rtpi_packet_direction direction = RTPI_PACKET_INCOMING,
				  bool posthandle = true);
    // clean up participant list after the packet is processed completely.
    void posthandle_rtcpi(RefCount<RtcpiCompoundPacket> rtcpi, IRtpiParticipantList& parts,
			  rtpi_packet_direction direction = RTPI_PACKET_INCOMING);
    // Processes a rtpi packet and updates participant infos
    // Returns the participant who has sent the packet
    // If direction == RTPI_PACKET_OUTGOING && new_packets_out == true
    // the following header fields are filled in (using master participant):
    //             - sequence number
    //             - pid
    // If direction == RTPI_PACKET_OUTGOING && new_packets_out == false
    // the following header fields are changed
    //             - sequence number
    //             - pid is replaced by a new pid
    //             - subids are replaced by new subids
    IRtpiParticipantList handle_rtpi(RefCount<RtpiAduVector> rtpi,
				 u_int32 addr,
				 rtpi_packet_direction direction,
				 bool new_packets_out = true);


    // manage participants

    // The master participant represents this application.
    // Local participants are only present if this application handles
    // third party data. They are generated automatically on the basis
    // of the handled data packets.
    // if deny_subcomp_reporting = true the master will not send any sub_components reports
    void createMasterParticipant(u_int32 local_addr, string user_name, bool deny_subcomp_reporting = false);
    // Remove all participants (local and master; remote participants stay)
    void removeAllParticipants(string reason = "");
    // Remove third party participants (master and remote participants stay).
    void removeLocalParticipants(string reason = "");
    RtpiParticipantList::iterator getMasterParticipant() {
      if (master_pos == 0)
	throw RtpiException(RtpiException::ERR_PARTICIPANT_INVALID, "no master participant.");
      return master_pos;
    }
    // get the first valid participant
    RtpiParticipantList::iterator getParticipantsBegin();
    // get the end iterator of the participant list
    RtpiParticipantList::iterator getParticipantsEnd();
    // get next valid participant
    RtpiParticipantList::iterator getNextParticipant(const RtpiParticipantList::iterator&);
    // find an existing participant
    RtpiParticipantList::iterator findParticipant(uint32 pid);



    // manage transmission (only needed for inter-stream synchronization)

    // reset transmission data for local participants
    // (reset timestamps for stream synchronization)
    void reset_transmission(bool include_master = true);
    // reset reception data for remote participants
    // (reset timestamps for stream synchronization)
    void reset_reception();


    // request sending of a rtcpi note item

    // Send a note via rtcpi for (all local and master) participant
    // not must be set directly for a participant
    void sendRtcpiNote();


    // get helper functions

    // get subcomponents and sequence number space
    RtpiSubComponentSpace* get_seq_space() { return seq_space; };
    // get master participant
    RtpiParticipantList::iterator& get_master_pos() { return master_pos; };
    // creates a timestamp of current rtpi time
    uint32 get_current_timestamp();

    // the following function will typically not be needed by a typical application
    // however they might be useful to some

    // create rtcpi sdes packet that contains all relevant sdes infos
    // usually only required if participant activities are recorded
    RtcpiCompoundPacket* create_participant_db_entry(RtpiParticipantList::iterator);
    // create rtcpi subrep packet that contains infos about all current subcomponents
    // usually only required if subcomponents are recorded
    RtcpiCompoundPacket* create_subcomponent_db_entry();

  private:
    void add_to_avg_rtcpi_size(int s) {
      average_rtcpi_packet_size = int((s * RTCPI_SIZE_GAIN) + (average_rtcpi_packet_size * (1 - RTCPI_SIZE_GAIN)));
    }
    int get_avg_rtcpi_size() {
      return average_rtcpi_packet_size;
    }
    // remove marked participants
    void pack_members();
    // return valid members
    u_int32 get_members();

    // execute reverse timer reconsideration for master and local participants
    void reverseTimerReconsiderations();

    RtpiParticipantList::iterator checkParticipant(rtpi_packet_direction direction,
					       bool& notify_part_created,
					       u_int32 pid, u_int32 addr,
					       string cname = "");
    RtpiParticipantList::iterator handle_rtcpi_sdes(RtcpiPacketVector::iterator hdrpos,
					       u_int32 addr,
					       rtpi_packet_direction direction,
					       bool& notify_part_created);
    RtpiParticipantList::iterator handle_rtcpi_subrep(RtcpiPacketVector::iterator hpos,
					       u_int32 addr,
					       rtpi_packet_direction direction,
					       bool& notify_db_changed,
					       bool& notify_part_created);
    void posthandle_rtcpi_bye(RtcpiPacketVector::iterator hpos,
			      rtpi_packet_direction direction,
			      RtpiParticipantList::iterator ppos);

    // implements TimerInterface
    void timerHandler(string);

    RtpiFeedbackInterface* get_feedback_interface() { return psfi; }

    // some nice tools
    string create_new_cname(string);
    string get_new_cname(string user = "", u_int32 i = 0);
    void replace_rtpi_subids(RefCount<RtpiAdu>, uint32);
  };

};

#endif






