// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//
// Copyright (c) 2000 University of Mannheim. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for educational, research and non-profit purposes,
// without fee, and without a written agreement is hereby granted,
// provided that the above copyright notice, this paragraph and the
// following four paragraphs appear in all copies.
//
// Permission to incorporate this software into commercial products may
// be obtained from the following address:
// University of Mannheim
// Lehrstuhl fuer Praktische Informatik IV
// L15, 16
// 68161 Mannheim, Germany
//
// IN NO EVENT SHALL THE UNIVERSITY OF MANNHEIM BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
// ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF MANNHEIM HAS BEEN ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//
// THE UNIVERSITY OF MANNHEIM SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER
// IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF MANNHEIM HAS NO OBLIGATIONS
// TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The University of Mannheim will appreciate receiving any enhancements and
// modifications of this software made by others, and reserves the right to
// include them in future releases. The University also recommends that this
// software not be redistributed by third parties, but be obtained by all
// directly from the University's server, so that all users of the software
// may receive bug notifications and other relevant announcements.
//
// Authors:
//   Volker Hilt, Juergen Vogel, Martin Mauve
//   University of Mannheim
//   {hilt,vogel,mauve}@informatik.uni-mannheim.de
//
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#include <rtpi/rtpiparticipant.h>
#include <rtpi/rtpihandler.h>

#define xDEBUG0
#define xDEBUG1
#define xDEBUG2
#define xDEBUG3
#define xDEBUG4
#define xDEBUG5
#define xDEBUG8

#define DEBUG_ERRORS

#define UNSYNCED_CLOCK_WARNING_TIME 3000

// if WAIT_FOR_RTCPI is set, we wait for a valid rtcpi packet
// before we consider a site valid.
#define WAIT_FOR_RTCPI false

// how many inactivity check intervals will we wait until we remove
// a byed participant
#define RTPI_BYE_DELAY 6

// how many members must be in a group to use bye backoff (otherwise bye
// is sent immediately)
#define RTPI_BYE_BACKOFF_MEMBERS 50

// LOOPBACK_OFF must be set if multicast loopback is not enabled
//#define LOOPBACK_OFF 1

//friend
std::ostream& RTPI::operator<<(std::ostream& os, RTPI::RtpiParticipant &p)
  // {{{ body
{
  os << "Participant (pid:" << p.pid << " / alt_pid: " << p.alt_pid << ") { " << std::endl;
  os << "\tCName: \t" << p.get_cname() << std::endl;
  os << "\tName: \t" << p.get_name() << std::endl;
  os << "\tEMail: \t" << p.get_email() << std::endl;
  os << "\tPhone: \t" << p.get_phone() << std::endl;
  os << "\tLoc: \t" << p.get_loc() << std::endl;
  os << "\tTool: \t" << p.get_tool() << std::endl;
  os << "\tNote: \t" << p.get_note() << std::endl;
  os << "\tPriv: \t" << p.priv << std::endl;
  os << "\tParticipant type: ";
  switch (p.type) {
  case RTPI_MASTER:
    std::cout << "RTPI_MASTER ";
    break;
  case RTPI_LOCAL:
    std::cout << "RTPI_LOCAL ";
    break;
  case RTPI_REMOTE:
    std::cout << "RTPI_REMOTE ";
    break;
  }
  os << " state: ";
  switch (p.state) {
  case RTPI_INVALID:
    std::cout << "RTPI_INVALID";
    break;
  case RTPI_VALID:
    std::cout << "RTPI_VALID";
    break;
  case RTPI_BYED:
    std::cout << "RTPI_BYED";
    break;
  case RTPI_REMOVE:
    std::cout << "RTPI_REMOVE ";
    break;
  }
  os << " } " << std::endl;
  return os;
};
// }}}


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
RTPI::RtpiParticipant::RtpiParticipant(RtcpiReportInterface* p,
				       RtpiHandler* rh,
				       u_int32 pi,
				       int t)
{
  #ifdef DEBUG2
  std::cout << "RtpiParticipant()::RtpiParticipant("<< pi << ", " << t << ")" << std::endl;
  #endif

  prri = p;
  prh = rh;
  pid = pi;
  set_type(t);
  rtcpi_timer = Timer::newTimer("rtcpi_timer", this);

  set_state(RTPI_INVALID);

  which_sdes_item=0;

  cname= "";
  name = "";
  email = "";
  phone = "";
  loc = "";
  tool = "";
  note = "";
  priv = "";
  alt_pid=0;
  addr=0;

  first_data = true;

  // data needed to compute the rtcpi interval
  pmembers = 1;
  initial_report_to_send = true;

  timeval now = unixtime();
  last_data_sent = now;
  last_data_received = now;
  last_report_sent = now;
  last_report_read = now;
  last_report_received = now;
  next_report_sent = now;

  deny_subcomp_reports = false;

  flag=false;

  // should we consider a site valid only if we received a valid rtcpi packet
  wait_for_rtcpi = WAIT_FOR_RTCPI;

  payload_type=0;

  reported_bye = 0;
  bye_backoff_bypass = false;
  member_bypass = 0;

  // init the rtcpi report timer
  if ((type == RTPI_MASTER) || (type == RTPI_LOCAL))
    init_report_timer();
};

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
RTPI::RtpiParticipant::~RtpiParticipant() {
  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::~RtpiParticipant()" << std::endl;
  #endif

  if (rtcpi_timer)
    delete(rtcpi_timer);

  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::~RtpiParticipant() done." << std::endl;
  #endif

};


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiParticipant::reset_transmission() {
  first_data = true;
};

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiParticipant::init_report_timer(int sdes) {
  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::init_report_timer()" << std::endl;
  #endif

  // only the RTPI_MASTER and RTPI_LOCALs are allowed to send reports
  if ((type != RTPI_MASTER) && (type != RTPI_LOCAL)) {
    #ifdef DEBUG2
    std::cout << "RtpiParticipant("<<pid<<")::init_report_timer(): only master and local participants are allowed to send reports." << std::endl;
    #endif
    // cancel any running timer!
    rtcpi_timer->cancel();
    return;
  }

  // is this the first report
  if (initial_report_to_send) {
    #ifdef DEBUG2
    std::cout << "RtpiParticipant("<<pid<<")::init_report_timer(): scheduling timer." << std::endl;
    #endif
    // maybe a specific sdes item was requested
    ReportsToSend.push_back(RtcpiReportParameters(sdes));
    // schedule the next report ...
    schedule_rtcpi_timer(rtcpi_interval());
    initial_report_to_send = false;
  }
};


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiParticipant::send_report(int sdes, bool send_bye, string bye_msg) {
  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::send_report() " << name << std::endl;
  #endif

  // only the RTPI_MASTER and RTPI_LOCALs are allowed to send reports
  if ((type != RTPI_MASTER) && (type != RTPI_LOCAL))  {
    #ifdef DEBUG2
    std::cout << "RtpiParticipant("<<pid<<")::send_report(): only master and local participants are allowed to send reports." << std::endl;
    #endif
    ReportsToSend.clear();
    // cancel any running timer!
    rtcpi_timer->cancel();
    return;
  }

  // do we have a report interface to send a report to?
  if (prri==NULL) {
    #ifdef DEBUG2
    std::cout << "RtpiParticipant("<<pid<<")::send_report(): Internal error: PRRI not defined. No report sent - rescheduling timer. " << std::endl;
    #endif
    // keep timer active (maybe user gets valid later on ... )
    schedule_rtcpi_timer(rtcpi_interval());
    return;
  }

  // only valid participant are allowed to send reports
  // we never issue a report without a cname
  if (!(((state == RTPI_VALID) || ((state == RTPI_BYED) && bye_backoff_bypass)) &&
       !cname.empty())) {
    #ifdef DEBUG2
    std::cout << "RtpiParticipant("<<pid<<")::send_report(): participant not valid or has no cname right now. No report sent - rescheduling timer. " << std::endl;
    #endif
    ReportsToSend.clear();
    // keep timer active (maybe user gets valid later on ... )
    schedule_rtcpi_timer(rtcpi_interval());
    return;
  }

  // cancel any running timer!
  rtcpi_timer->cancel();

  // check ReportsToSend list if we have to send reports other than
  // those in the parameters (this may result in multiple sdes items)
  bool bye_found = false;
	std::list<RtcpiReportParameters>::iterator rep = ReportsToSend.begin();
  while (rep != ReportsToSend.end()) {
    // do we have a bye in our list -> apply it
    if (rep->send_bye && !bye_found) {
      #ifdef DEBUG3
      std::cout << "RtpiParticipant(" << pid << ")::send_report() "
	   << "Bye found in ReportsToSend list. ";
      if (send_bye)
	std::cout << "Using it instead of current bye request.";
      std::cout << std::endl;
      #endif
      send_bye = true;
      bye_msg = rep->bye_msg;
      bye_found = true;
    }

    // remove invalid entries from list
    if ((rep->sdes == 0) || (rep->sdes == RtcpiSourceDescriptionItem::CNAME)) {
      rep = ReportsToSend.erase(rep);
    } else {
      ++rep;
    }
  }
  if (sdes != 0) {
    ReportsToSend.push_back(RtcpiReportParameters(sdes));
  }

  // rtcpi compound packet starts with an SDES packet
  #ifdef DEBUG3
  std::cout << "RtpiParticipant(" << pid << ")::send_report() "
       << "building SDES ..." << std::endl;
  #endif

  // create a brand new vector of RTCPI compound packets
  RtcpiCompoundPacketVector rtcpi;
  rtcpi.push_back(new RtcpiCompoundPacket());

  build_sdes(rtcpi, ReportsToSend);
  ReportsToSend.clear();

  #ifdef DEBUG_ERRORS
  if (rtcpi.empty()) {
    std::cout << "RtpiParticipant(" << pid << ")::send_report(): could not build sdes packet." << std::endl;
  }
  #endif

  if (!rtcpi.empty()) {
    // do we have to say goodbye
    if (send_bye) {
      #ifdef DEBUG3
      std::cout << "RtpiParticipant(" << pid << ")::send_report() "
	   << "building BYE ..." << std::endl;
      #endif

      build_bye(rtcpi, bye_msg);

      // turn off bye backoff bypasses (in case that we continue)
      stop_bye_backoff();

      // mark participant to be byed
      set_state(RTPI_BYED);
    } else {
      // otherwise we send a SUBREP
      #ifdef DEBUG3
      std::cout << "RtpiParticipant(" << pid << ")::send_report() "
	   << "building sub component report ..." << std::endl;
      #endif

      build_subrep(rtcpi);
    };

    #ifdef DEBUG3
    std::cout << "RtpiParticipant(" << pid << ")::send_report() : flushing " << rtcpi.size() << " compound packets" << std::endl;
    #endif

    // try to send our packets
    for (int i = 0; i < rtcpi.size(); i++) {
      try {
        #ifdef DEBUG3
	std::cout << "RtpiParticipant(" << pid << ")::send_report() "
	     << "Sending report:" << std::endl;
	rtcpi[i]->Print();
	#endif

	// flush the compound packets
	rtcpi[i]->SetBuffer(new Buffer(rtcpi[i]->GetSize()));
	rtcpi[i]->Flush();

        #ifdef DEBUG3
	std::cout << "Sending " << rtcpi[i]->GetBufferSize() << " bytes ..." << std::endl;
        #endif

	// due to local loopback we dont have to
	// include sent rtcp packet size in average_rtcpi_size

	// remember the time when we sent this report
	last_report_sent = unixtime();
	// since we have sent a report we set this to false
	initial_report_to_send = false;

	// call our RtcpiReportInterface to send a report
	prri->send_rtcpi(rtcpi[i]->GetBuffer());

        #ifdef LOOPBACK_OFF
	// include received rtcp packet size in average
	prh->add_to_avg_rtcpi_size(rtcpi[i]->GetBufferSize());
        #endif

      } catch (RtpiException e) {
        #ifdef DEBUG1
	std::cout << "RtpiParticipant::send_report(): "
	     << " RTCPI flush error. Reason: "
	     << e.errMsg << " (" << e.errCode << ")" <<std::endl;
        #endif
      }
    }
  };

  // schedule the next report
  schedule_rtcpi_timer(rtcpi_interval());

  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::send_report() done." << std::endl;
  #endif
};


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//* sending bye using bye backoff
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiParticipant::send_bye_and_remove(string msg) {
  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::send_bye_and_remove("<<msg<<")" << std::endl;
  #endif

  // only the RTPI_MASTER and RTPI_LOCALs are allowed to send reports
  if ((type != RTPI_MASTER) && (type != RTPI_LOCAL))  {
    #ifdef DEBUG2
    std::cout << "RtpiParticipant("<<pid<<")::send_bye_and_remove(): only master and local participants are allowed to send byes." << std::endl;
    #endif
    return;
  }
  // don't send duplicate byes
  if (state == RTPI_BYED)  {
    #ifdef DEBUG2
    std::cout << "RtpiParticipant("<<pid<<")::send_bye_and_remove(): we are already in byed state." << std::endl;
    #endif
    return;
  }

  // use backoff only if more than a certain number of members are in our group
  if (get_number_of_members() > RTPI_BYE_BACKOFF_MEMBERS) {
    #ifdef DEBUG1
    std::cout << "RtpiParticipant("<<pid<<")::send_bye_and_remove(): Using bye backoff." << std::endl;
    #endif
    // Store bye parameters in list
    ReportsToSend.push_back(RtcpiReportParameters(true, msg));

    // reset parameters
    last_report_sent = unixtime();
    initial_report_to_send = true;
    pmembers = 1;
    // now we use our bypasses
    start_bye_backoff();

    // we're byed
    set_state(RTPI_BYED);

    // schedule bye report
    schedule_rtcpi_timer(rtcpi_interval());
  } else {

    // we're allowed to send rigth now :)
    send_report(0, true, msg);
  }
};


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiParticipant::send_bye_no_statechange(string msg) {
  #ifdef DEBUG1
  std::cout << "RtpiParticipant("<<pid<<")::send_bye_no_backoff("<<msg<<")" << std::endl;
  #endif

  send_report(0, true, msg);
};

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiParticipant::build_sdes(RtcpiCompoundPacketVector& rtcpi, std::list<RtcpiReportParameters>& sdes_items, bool automatic_sdes) {

  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::build_sdes()" << std::endl;
  #endif

  if (rtcpi.empty()) {
    rtcpi.push_back(new RtcpiCompoundPacket());
  }

  RtcpiSourceDescriptionPacket* sdes_packet = new RtcpiSourceDescriptionPacket(pid);

  // build list of sdes items
  // always send RTCPI_SDES_CNAME
  sdes_items.push_front(RtcpiReportParameters(RtcpiSourceDescriptionItem::CNAME));
  // the next sdes item is calculated by next_sdes_item()
  if (automatic_sdes) {
    int auto_item = next_sdes_item();
    // maybe we have it already ...
    bool found = false;
		std::list<RtcpiReportParameters>::iterator rep = sdes_items.begin();
    while ((rep != sdes_items.end()) && !found) {
      found = (rep->sdes == auto_item);
      ++rep;
    }
    if (!found) {
      sdes_items.push_back(auto_item);
    }
  }

  // build items
	std::list<RtcpiReportParameters>::iterator rep;
  for (rep = sdes_items.begin(); rep != sdes_items.end(); ++rep) {
    // build one item
    RtcpiSourceDescriptionItem* tmp_item = NULL;
    try {
      tmp_item = build_sdes_item(rep->sdes);
    } catch (RtpiException e) {
      #ifdef DEBUG_ERRORS
      std::cout << "RtpiParticipant(" << pid << ")::build_sdes(): could not build item. Reason: "
	   << e.errMsg << " (" << e.errCode << ")" << std::endl;
      #endif
      if (tmp_item != NULL) {
	delete tmp_item;
	tmp_item = NULL;
      }
    }
    // check item
    if ((tmp_item != NULL) && (tmp_item->GetSize() > rtcpi.back()->GetMaxSize())) {
      #ifdef DEBUG_ERRORS
      std::cout << "RtpiParticipant(" << pid << ")::build_sdes(): Item is too large. Discarding item." << std::endl;
      #endif
      delete tmp_item;
      tmp_item = NULL;
    }
    // add item to packet
    if (tmp_item != NULL)  {
      try {
	sdes_packet->AddItem(tmp_item, rtcpi.back()->GetMaxSize());
      } catch (RtpiException e) {
	// current sdes packet is full. create a new one.
	// store old packet in compound packet
	if (sdes_packet->ContainsItems()) {
	  try {
	    rtcpi.back()->AddPacket(sdes_packet);
	  } catch (RtpiException e) {
	    // current compound packet is full. create a new one.
	    if (rtcpi.back()->ContainsPackets()) {
	      rtcpi.push_back(new RtcpiCompoundPacket());
	    }
	    // try again with the new compound packet
	    try {
	      rtcpi.back()->AddPacket(sdes_packet);
	    } catch (RtpiException e) {
	      // something is really wrong here.
              #ifdef DEBUG_ERRORS
	      std::cout << "RtpiParticipant(" << pid << ")::build_sdes(): Could not add packet to compound packet. Discarding sdes packet. Reason: "
		   << e.errMsg << " (" << e.errCode << ")" << std::endl;
              #endif
	      delete sdes_packet;
	    }
	  }
	  // create new sdes packet
	  sdes_packet = new RtcpiSourceDescriptionPacket(pid);
	  // SDES packet must start with a CNAME item
	  if (tmp_item->GetType() != RtcpiSourceDescriptionItem::CNAME) {
	    RtcpiSourceDescriptionItem* tmp_item_2 = NULL;
	    try {
	      tmp_item_2 = build_sdes_item(RtcpiSourceDescriptionItem::CNAME);
	      sdes_packet->AddItem(tmp_item_2, rtcpi.back()->GetMaxSize());
	    } catch (RtpiException e) {
	      // uh. This should never happen.
              #ifdef DEBUG_ERRORS
	      std::cout << "RtpiParticipant(" << pid << ")::build_sdes(): could not build CNAME. Reason: "
		   << e.errMsg << " (" << e.errCode << ")" << std::endl;
              #endif
	      if (tmp_item_2 != NULL) {
		delete tmp_item_2;
		tmp_item_2 = NULL;
	      }
	    }
	  }
	}
	// try again with the new sdes packet
	try {
	  sdes_packet->AddItem(tmp_item, rtcpi.back()->GetMaxSize());
	} catch (RtpiException e) {
	  // something is really wrong here.
          #ifdef DEBUG_ERRORS
	  std::cout << "RtpiParticipant(" << pid << ")::build_sdes(): Could not add item to sdes packet. Discarding item. Reason: "
	       << e.errMsg << " (" << e.errCode << ")" << std::endl;
          #endif
	  delete tmp_item;
	}
      }
    }
  }

  // see if we have components to report
  if (!sdes_packet->ContainsItems()) {
    delete sdes_packet;
    if (!rtcpi.back()->ContainsPackets()) {
      // remove empty packet from vector
      rtcpi.pop_back();
    }
  } else {
    // add our last sdes packet to last compound packet
    try {
      rtcpi.back()->AddPacket(sdes_packet);
    } catch (RtpiException e) {
      // current compound packet is full. create a new one.
      if (rtcpi.back()->ContainsPackets()) {
	rtcpi.push_back(new RtcpiCompoundPacket());
      }
      // try again with the new compound packet
      try {
	rtcpi.back()->AddPacket(sdes_packet);
      } catch (RtpiException e) {
	// something is really wrong here.
        #ifdef DEBUG_ERRORS
	std::cout << "RtpiParticipant(" << pid << ")::build_sdes(): Could not add packet to compound packet. Discarding sdes packet. Reason: "
	     << e.errMsg << " (" << e.errCode << ")" << std::endl;
         #endif
	delete sdes_packet;
      }
    }
  }

  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::build_sdes() done." << std::endl;
  #endif
};


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiParticipant::build_bye(RtcpiCompoundPacketVector& rtcpi, string bye_msg) {

  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::build_bye("<<bye_msg<<")" << std::endl;
  #endif

  if (rtcpi.empty()) {
    #ifdef DEBUG3
    std::cout << "RtpParticipant("<<pid<<")::build_bye(): Bye can't be first in compount packet. Internal error. " << std::endl;
    #endif
    // no packet to add the bye to
    return;
  }

  RtcpiByePacket* bye_packet = NULL;
  try {
    bye_packet = new RtcpiByePacket(pid,
				    (bye_msg.empty() ? NULL : bye_msg.c_str()),
				    ((bye_msg.size() > 255) ? 255 : bye_msg.size()));
  } catch (RtpiException e) {
    #ifdef DEBUG_ERRORS
    std::cout << "RtpiParticipant(" << pid << ")::build_bye(): could not build bye packet. Reason: "
	 << e.errMsg << " (" << e.errCode << ")" << std::endl;
    #endif
    if (bye_packet != NULL) delete bye_packet;
  }

  // finally add our packet
  if (bye_packet != NULL) {
    try {
      rtcpi.back()->AddPacket(bye_packet);
    } catch (RtpiException e) {
      // current compound packet is full. create a new one.
      if (rtcpi.back()->ContainsPackets()) {
	rtcpi.push_back(new RtcpiCompoundPacket());
      }
      // the new compound packet must start with an CNAME SDES packet
      build_sdes(rtcpi, false);
      // try again with the new compound packet
      try {
	rtcpi.back()->AddPacket(bye_packet);
      } catch (RtpiException e) {
	// something is really wrong here.
        #ifdef DEBUG_ERRORS
	std::cout << "RtpiParticipant(" << pid << ")::build_bye(): Could not add packet to compound packet. Discarding bye packet. Reason: "
	     << e.errMsg << " (" << e.errCode << ")" << std::endl;
         #endif
	delete bye_packet;
      }
    }
  }
};


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiParticipant::build_subrep(RtcpiCompoundPacketVector& rtcpi) {

  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::build_subrep()" << std::endl;
  #endif

  #ifdef DEBUG4
  std::cout << "Current SubComponent List: " << std::endl;
  std::cout << *(prh->seq_space) << std::endl;
  #endif

  if (rtcpi.empty()) {
    #ifdef DEBUG3
    std::cout << "RtpParticipant("<<pid<<")::build_subrep(): SUBREP can't be first in compount packet. Internal error. " << std::endl;
    #endif
    // no packet to add the bye to
    return;
  }

  // create packet
  RtcpiSubcomponentReportPacket* subrep_packet = new RtcpiSubcomponentReportPacket(pid);

  // create items
  bool db_modified = false;
  int t_d = double2int(rtcpi_deterministic_interval(get_number_of_members()));
  SubComponentMap::iterator spos = prh->seq_space->get_components_begin();
  while (spos != prh->seq_space->get_components_end()) {
    // remove this item if not valid any more
    if (spos->second.is_timed_out(t_d)) {
      // tell the application about it
      if (prh->get_feedback_interface() != NULL) {
	prh->get_feedback_interface()->subcomponent_removed(spos->first);
	db_modified = true;
      }
      spos = prh->seq_space->delete_component(spos);
    } else {
      // do we have to report this item (and are we allowed to report)
      if (!deny_subcomp_reports && spos->second.must_report(pid, t_d)) {
	RtcpiSubcomponentReportItem* tmp_item = NULL;
	// do we have to report a name for this item
	try {
	  if ((prh->get_feedback_interface() != NULL) &&
	      (!prh->get_feedback_interface()->report_subcomponent_name(spos->first))) {
	    tmp_item = new RtcpiSubcomponentReportItem(spos->first,
						       spos->second.get_local_active());
	  } else {
	    tmp_item = new RtcpiSubcomponentReportItem(spos->first,
						       spos->second.get_local_active(),
						       spos->second.name,
						       spos->second.name_length);
	  }

	  // check item
	  if ((tmp_item != NULL) && (tmp_item->GetSize() > rtcpi.back()->GetMaxSize())) {
          #ifdef DEBUG_ERRORS
	    std::cout << "RtpiParticipant(" << pid << ")::build_subrep(): Item is too large. Discarding item." << std::endl;
          #endif
	    delete tmp_item;
	    tmp_item = NULL;
	  }
	  // add item to packet
	  if (tmp_item != NULL)  {
	    try {
	      subrep_packet->AddItem(tmp_item, rtcpi.back()->GetMaxSize());
	    } catch (RtpiException e) {
	      // current subrep packet is full. create a new one.
	      // store old packet in compound packet
	      if (subrep_packet->ContainsItems()) {
		try {
		  rtcpi.back()->AddPacket(subrep_packet);
		} catch (RtpiException e) {
		  // current compound packet is full. create a new one.
		  if (rtcpi.back()->ContainsPackets()) {
		    rtcpi.push_back(new RtcpiCompoundPacket());
		    // compound packet must start with sdes item
		    build_sdes(rtcpi, false);
		  }
		  // try again with the new compound packet
		  try {
		    rtcpi.back()->AddPacket(subrep_packet);
		  } catch (RtpiException e) {
		    // something is really wrong here.
                    #ifdef DEBUG_ERRORS
		    std::cout << "RtpiParticipant(" << pid << ")::build_subrep(): Could not add packet to compound packet. Discarding subrep packet. Reason: "
			 << e.errMsg << " (" << e.errCode << ")" << std::endl;
                    #endif
		    delete subrep_packet;
		  }
		}
		// create new sdes packet
		subrep_packet = new RtcpiSubcomponentReportPacket(pid);
	      }
	      // try again with the new sdes packet
	      try {
		subrep_packet->AddItem(tmp_item, rtcpi.back()->GetMaxSize());
	      } catch (RtpiException e) {
		// something is really wrong here.
                #ifdef DEBUG_ERRORS
		std::cout << "RtpiParticipant(" << pid << ")::build_subrep(): Could not add item to subrep packet. Discarding item. Reason: "
		     << e.errMsg << " (" << e.errCode << ")" << std::endl;
                #endif
		delete tmp_item;
	      }
	    }
	  }

	  // tell component db that we are reporting
	  spos->second.report_sent(pid);

	} catch (RtpiException e) {
          #ifdef DEBUG_ERRORS
	  std::cout << "RtpiParticipant(" << pid << ")::build_subrep(): could not build item. Reason: "
	       << e.errMsg << " (" << e.errCode << ")" << std::endl;
          #endif
	  if (tmp_item != NULL) delete tmp_item;
	}
      }
      ++spos;
    }
  }

  // tell the application about changes
  if (db_modified) {
    prh->get_feedback_interface()->subcomponent_db_modified();
  }

  // see if we have components to report
  if (!subrep_packet->ContainsItems()) {
    delete subrep_packet;
    if (!rtcpi.back()->ContainsPackets()) {
      // remove empty packet from vector
      rtcpi.pop_back();
    }
  } else {
    // add our last subrep packet to last compound packet
    try {
      rtcpi.back()->AddPacket(subrep_packet);
    } catch (RtpiException e) {
      // current compound packet is full. create a new one.
      if (rtcpi.back()->ContainsPackets()) {
	rtcpi.push_back(new RtcpiCompoundPacket());
	// compound packet must start with sdes item
	build_sdes(rtcpi, false);
      }
      // try again with the new compound packet
      try {
	rtcpi.back()->AddPacket(subrep_packet);
      } catch (RtpiException e) {
	// something is really wrong here.
        #ifdef DEBUG_ERRORS
	std::cout << "RtpiParticipant(" << pid << ")::build_subrep(): Could not add packet to compound packet. Discarding subrep packet. Reason: "
	     << e.errMsg << " (" << e.errCode << ")" << std::endl;
         #endif
	delete subrep_packet;
      }
    }
  }

};


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
int RTPI::RtpiParticipant::next_sdes_item() {

  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::next_sdes_item()" << std::endl;
  #endif

  // the schedule for what we send is sequenced by which_sdes_item:
  // we have 16 slots (slot 0 to slot 15)
  //   - send 'note'  every 0th & 8th slot
  //   - send 'email' every 2th & 10th slot
  //   - send 'phone' every 4nd & 12th slot
  //   - send 'tool'  every 6th & 14th slot
  //   - send 'name'  in all odd slots

  int this_sdes = (++which_sdes_item) & 0xF;
  switch (this_sdes) {
    case 0:
    case 8:
      return RtcpiSourceDescriptionItem::NOTE;
    case 2:
    case 10:
      return RtcpiSourceDescriptionItem::EMAIL;
    case 4:
    case 12:
      return RtcpiSourceDescriptionItem::PHONE;
    case 6:
    case 14:
      return RtcpiSourceDescriptionItem::TOOL;
    default:
      return RtcpiSourceDescriptionItem::NAME;
  };
};


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
RTPI::RtcpiSourceDescriptionItem* RTPI::RtpiParticipant::build_sdes_item(int sdes) {

  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::build_sdes_item("<< int(sdes) << ")" << std::endl;
  #endif

  // build item info string
  string str = "";

  switch (sdes) {
    case RtcpiSourceDescriptionItem::CNAME:
      str = get_cname();
      break;
    case RtcpiSourceDescriptionItem::NAME:
      str = get_name();
      break;
    case RtcpiSourceDescriptionItem::EMAIL:
      str = get_email();
      break;
    case RtcpiSourceDescriptionItem::PHONE:
      str = get_phone();
      break;
    case RtcpiSourceDescriptionItem::LOC:
      str = get_loc();
      break;
    case RtcpiSourceDescriptionItem::TOOL:
      str = get_tool();
      break;
    case RtcpiSourceDescriptionItem::NOTE:
      str = get_note();
      break;
    case RtcpiSourceDescriptionItem::PRIV:
      str = "";
      // ignore
      break;
  };

  #ifdef DEBUG2
  if (type!=RTPI_MASTER) {
    std::cout << "RtpiParticipant("<<pid<<")::build_sdes_item("
         << sdes << ") == \"" << str << "\"" << std::endl;
    std::cout << "cname = \"" << get_cname() << "\"" << std::endl;
    std::cout << "name  = \"" << get_name() << "\"" << std::endl;
    std::cout << "email = \"" << get_email() << "\"" << std::endl;
    std::cout << "phone = \"" << get_phone() << "\"" << std::endl;
    std::cout << "tool  = \"" << get_tool() << "\"" << std::endl;
  };
  #endif

  // build item header based on string
  RtcpiSourceDescriptionItem* item = NULL;
  if (!str.empty()) {
    item = new RtcpiSourceDescriptionItem(sdes, str.c_str(),
		      ((str.length() > 255) ? 255 : str.length()));
    #ifdef DEBUG2
  } else {
    std::cout << "RtpiParticipant("<<pid<<")::build_sdes_item(): Item empty. No sdes item built." << std::endl;
    #endif
  }

  return item;
};


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiParticipant::timerHandler(string id) {

  if (id == "rtcpi_timer") {

    #ifdef DEBUG2
    std::cout << "RtpiParticipant("<<pid<<")::timerHandler(" << id << ") "
	 << "checking report transmission time... ";
    #endif

    // re-check transmission timer (see draft 7.3.6):
    // only send a report if (last-report + newly_calculated_interval) <= now
    int interval = rtcpi_interval();
    int tc = (unixtime()).tv_sec;
    if ((last_report_sent.tv_sec + interval) > tc) {
      // we still have some time -> reschedule timer
      #ifdef DEBUG2
      std::cout << "rescheduling timer: " << last_report_sent.tv_sec + interval - tc << "sec." << std::endl;
      #endif
      schedule_rtcpi_timer(last_report_sent.tv_sec + interval - tc);
    return;
    }

    #ifdef DEBUG2
    std::cout << "transmitting report." << std::endl;
    #endif
    // send an rtcpi report (send_report will reschedule the timer...)
    send_report();

    return;
  }
}


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiParticipant::update_state() {
  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::update_state() type " << type << " state " << state << std::endl;
  #endif

  if (prh == NULL) {
    // we don't have a rtpi handler (should never happen)
    return;
  }

  #ifdef DEBUG1
  std::cout << "RtpiParticipant("<<pid<<")::update_state()"
       << "checking weather this site is still valid..." << std::endl;
  #endif

  // check weather this site is still valid
  if ((type != RTPI_MASTER) &&
      ((state == RTPI_VALID) || (state == RTPI_INVALID))) {
    // we mark a site inactive
    // if we did not hear from it for 5 report intervals
    timeval now = unixtime();
    timeval data_diff;
    timeval ctrl_diff;
    timeval diff;
    bool timer_set = false;
    timerclear(&data_diff);
    timerclear(&ctrl_diff);
    timerclear(&diff);
    if (type == RTPI_REMOTE) {
      if (timerisset(&last_data_received)) {
	data_diff=tvdiff(now, last_data_received);
	timer_set = true;
      };
      if (timerisset(&last_report_received)) {
	ctrl_diff=tvdiff(now, last_report_received);
	timer_set = true;
      };
    } else if (type == RTPI_LOCAL) {
      if (timerisset(&last_data_sent)) {
	data_diff=tvdiff(now, last_data_sent);
	timer_set = true;
      };
      if (timerisset(&last_report_read)) {
	ctrl_diff=tvdiff(now, last_report_read);
	timer_set = true;
      };
    }
    diff = ((timercmp(&data_diff,&ctrl_diff,<)) ? data_diff : ctrl_diff);
    int interval = double2int(rtcpi_deterministic_interval(get_number_of_members()));

    #ifdef DEBUG4
    std::cout << "Last data/report sent/received: " << diff.tv_sec << " Remove timeout: " << interval * RTPI_REMOVE_INTERVAL << std::endl;
    #endif

    if (!timer_set || (diff.tv_sec > interval * RTPI_REMOVE_INTERVAL))  {
      #ifdef DEBUG2
      std::cout << "RtpiParticipant("<<pid<<")::update_state() "
	   << "site will be removed from the participants list." << std::endl;
      #endif

      // cancel running report timer!
      rtcpi_timer->cancel();

      // if we remove a local participant be polite and send a bye
      if (type == RTPI_LOCAL) {
	// site is marked to be removed after sending the bye
	send_bye_and_remove("end of data");
      } else {
	set_state(RTPI_REMOVE);
      }

      // execute timer reconsideration
      prh->reverseTimerReconsiderations();

      return;
    }
  }

  // do we have a byed participant
  if ((state == RTPI_BYED) && (!bye_backoff_bypass))  {
    // we remove the site RTPI_BYE_DELAY inactivity check intervals after the bye
    if (reported_bye == RTPI_BYE_DELAY) {
      #ifdef DEBUG2
      std::cout << "RtpiParticipant("<<pid<<")::update_state() "
	   << "site said bye and will now be removed from the participants list." << std::endl;
      #endif

      // cancel any running timer!
      rtcpi_timer->cancel();

      set_state(RTPI_REMOVE);

      return;
    } else {
      reported_bye++;
    }
  }

}


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiParticipant::check_incoming_rtpi(RefCount<RtpiAdu> rtpi) {
  #ifdef DEBUG2
  std::cout << "RtpiParticipant():check_incoming_rtpi()" << std::endl;
  #endif
  #ifdef DEBUG4
  std::cout << "RtpiParticipant::check_incoming_rtpi(" << pid << ") rtpi:"
       << rtpi->aduHeader->ts << " type: " << (int)rtpi->packetHeader->type
       << " seq:" << (int)rtpi->packetHeader->sn
       << " RTPI time now: " << rtpitime() << " (Difference: "
       << (rtpi->aduHeader->ts - rtpitime()) << "/"
       << (rtpitime() - rtpi->aduHeader->ts) << ")." << std::endl;
  #endif
  #ifdef DEBUG_ERRORS
  uint32 diff = rtpi->aduHeader->ts - rtpitime();
  if (((rtpi->aduHeader->ts - rtpitime()) > UNSYNCED_CLOCK_WARNING_TIME) &&
      ((rtpitime() - rtpi->aduHeader->ts) > UNSYNCED_CLOCK_WARNING_TIME)) {
    std::cout << "WARNING: Difference between local clock and timestamp of ADU " << std::endl
	 << "from participant '" << get_name() << "' is very large. " << std::endl
	 << "Maybe clocks are not synchronized." << std::endl;
  }
  #endif

  // check/modify current state
  if (wait_for_rtcpi) {
    #ifdef DEBUG0
    std::cout << "RtpiParticipant():check_incoming_rtpi() still waiting for "
	 << "a valid rtcpi packet..." << std::endl;
    #endif
    set_state(RTPI_INVALID);

    throw RtpiException(RtpiException::ERR_PARTICIPANT_INVALID, "still waiting for a valid rtcpi packet.");
  };

  set_state(RTPI_VALID);
  last_data_received = unixtime();

  // find component/adu/participant
  ParticipantSeqMapElement* p_elem = prh->seq_space->
    find_participant_seq(rtpi->packetHeader->subid, rtpi->packetHeader->GetType(), pid);
  if (p_elem == NULL) {
    // participant not found -> insert new participant
    p_elem = prh->seq_space->insert_participant_seq(rtpi->packetHeader->subid,
				      rtpi->packetHeader->GetType(), pid);
    if (p_elem == NULL) {
      // subcomponent did not exist
      // this should never happen!
      throw RtpiException(RtpiException::ERR_SUBCOMPONENT_INVALID, "Subcomponent non-existent. Could not create participant. Ignoring packet.");
    }
  }
  // check sequence number
  uint8 seq_ret = p_elem->second.seq_ok(rtpi->packetHeader->sn);
  #ifdef DEBUG0
  if (seq_ret != RtpiParticipantSeq::SEQ_OK) {
    std::cout << "RtpiParticipant():check_incoming_rtpi(): Warning: Sequence number invalid: ";
    RTPI::RtpiParticipantSeq::err_out(seq_ret);
    std::cout << std::endl;
  };
  #endif

  // provide basis for timestamp calculation
  if (first_data) {
    #ifdef DEBUG1
    std::cout << "first data!" << std::endl;
    #endif
    first_data = false;
  };

};


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiParticipant::check_incoming_rtcpi(RtcpiPacketVector::iterator hpos) {
  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::check_incoming_rtcpi()" << std::endl;
  #endif

  // check if we can set this site to valid
  if (state == RTPI_INVALID) {
    #ifdef DEBUG4
    std::cout << "RtpiParticipant("<<pid<<")::check_incoming_rtcpi() state transition: RTPI_INVALID -> RTPI_VALID" << std::endl;
    #endif
    set_state(RTPI_VALID);
  };

  last_report_received = unixtime();

  wait_for_rtcpi = false;
};

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiParticipant::check_outgoing_rtpi(RefCount<RtpiAdu> rtpi) {
  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<< pid <<")::check_outgoing_rtpi()" << std::endl;
  #endif

  #ifdef DEBUG1
  std::cout << "RtpiParticipant::check_outgoing_rtpi() rtpi:" << rtpi->aduHeader->ts
       << " seq:" << rtpi->packetHeader->sn << std::endl;
  #endif

  last_data_sent = unixtime();

  // find component/adu/participant
  ParticipantSeqMapElement* p_elem = prh->seq_space->
    find_participant_seq(rtpi->packetHeader->subid, rtpi->packetHeader->GetType(), pid);
  if (p_elem == NULL) {
    // participant not found -> insert new participant
    p_elem = prh->seq_space->insert_participant_seq(rtpi->packetHeader->subid,
				      rtpi->packetHeader->GetType(), pid);
    if (p_elem == NULL) {
      // subcomponent did not exist
      // this should never happen!
      throw RtpiException(RtpiException::ERR_SUBCOMPONENT_INVALID, "Subcomponent non-existent. Could not create participant. Ignoring packet.");
    }
  }
  // generate a new sequence number for this subcomponent/adu_type/participant
  rtpi->packetHeader->sn = p_elem->second.get_next_seq();

  // provide basis for timestamp calculation
  if (first_data) {
    #ifdef DEBUG1
    std::cout << "first!" << std::endl;
    #endif
    first_data=false;
  };

  set_state(RTPI_VALID);

  // set the pid for playing !
  rtpi->packetHeader->pid = pid;

  // the rtpi packet is changed by this function.
  // rtpi->Flush() MUST be called to write these changes to memory buffer

};

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiParticipant::check_outgoing_rtcpi(RtcpiPacketVector::iterator hpos) {
  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::check_outgoing_rtcpi()" << std::endl;
  #endif

  // check if we can set this site to valid
  if (state == RTPI_INVALID) {
    #ifdef DEBUG4
    std::cout << "RtpiParticipant("<<pid<<")::check_outgoing_rtcpi() state transition: RTPI_INVALID -> RTPI_VALID" << std::endl;
    #endif
    set_state(RTPI_VALID);
  };

  // we do not really sent outgoing rtcpi packets
  // we just update participant infos and let the timer decide

  last_report_read = unixtime();

  // set the pid
  (*hpos)->pid = pid;

}



//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
double RTPI::RtpiParticipant::rtcpi_deterministic_interval(int members) {
  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::rtcpi_deterministic_interval()" << std::endl;
  #endif

  // this function is idempotent

  // retrieve init values
  double t_min = RTCPI_MIN_RPT_TIME;  // should better be retrieved via pt-info
  if (initial_report_to_send) {
    t_min /= 2;
  };
  double rtcpi_bw = prri->get_session_bandwith() * RTCPI_SESSION_BW_FRACTION;

  // calculate deterministic interval Td
  double t_d = ((double)get_average_rtcpi_size() * (double)members) / rtcpi_bw;
  if (t_d < t_min)
    t_d = t_min;

  #ifdef DEBUG3
  std::cout << " members: " << members << std::endl;
  std::cout << " average: " << get_average_rtcpi_size() << std::endl;
  std::cout << " rtcpi_bw: " << rtcpi_bw << std::endl;
  std::cout << " t_d: " << t_d << std::endl;
  #endif

  return t_d;
}


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
int RTPI::RtpiParticipant::rtcpi_interval() {
  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::rtcpi_interval()" << std::endl;
  #endif

  int members = get_number_of_members();

  double t_d = rtcpi_deterministic_interval(members);

  // calculate interval T
  double t = t_d * (drand48() + 0.5);
  t /= 1.21828;

  #if defined(DEBUG1) || defined(DEBUG3)
  std::cout << "RtpiParticipant("<<pid<<")::rtcpi_interval(): new interval: " << t << " -> " << round(t) << std::endl;
  #endif

  return double2int(t);
};


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiParticipant::reverseTimerReconsideration() {
  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::reverseTimerReconsinderation()" << std::endl;
  #endif

  // do we have a timer set
  if (initial_report_to_send) {
    #ifdef DEBUG2
    std::cout << "RtpiParticipant("<<pid<<")::reverseTimerReconsinderation(): No timer set." << std::endl;
    #endif
    return;
  }

  // retrieve init values
  int members = get_number_of_members();
  uint32 tc = (unixtime()).tv_sec;
  uint32 tp = last_report_sent.tv_sec;
  uint32 tn = next_report_sent.tv_sec;

  #ifdef DEBUG4
  std::cout << "tc: " << tc << std::endl;
  std::cout << "tp: " << tp << std::endl;
  std::cout << "tn: " << tn << std::endl;
  std::cout << "tc - tp: " << (tc - tp) << std::endl;
  std::cout << "members: " << members << std::endl;
  std::cout << "pmembers: " << pmembers << std::endl;
  #endif

  if (members >= pmembers) {
    // the number of members didn't change
    #ifdef DEBUG2
    std::cout << "RtpiParticipant("<<pid<<")::reverseTimerReconsinderation(): Number of participant unchanged or increased. Timer not modified. " << std::endl;
    #endif
    return;
  }

  if (tc >= tn) {
    // were over the time => let our participant report normally next time
    #ifdef DEBUG2
    std::cout << "RtpiParticipant("<<pid<<")::reverseTimerReconsinderation(): Report is due now. Timer not modified. " << std::endl;
    #endif
    return;
  }

  // now we can cancel the running timer
  rtcpi_timer->cancel();

  // calculate new interval (corresponds to tn - tc)
  int new_interval = double2int(((double)members/(double)pmembers) * (double)(tn - tc));

  #ifdef DEBUG4
  std::cout << "reconsidered interval: " << new_interval << std::endl;
  #endif

  // modify last_report_sent
  last_report_sent.tv_sec = tc - double2int(((double)members/(double)pmembers) * (double)(tc - tp));

  // re-schedule report timer
  schedule_rtcpi_timer(new_interval);
};


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//* Schedule next rtcpi transmission timer
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiParticipant::schedule_rtcpi_timer(int t) {
  #ifdef DEBUG2
  std::cout << "RtpiParticipant("<<pid<<")::schedule_rtcp_timer(" << t << ")" << std::endl;
  #endif
  rtcpi_timer->cancel();
  rtcpi_timer->ssched(t);
  next_report_sent = unixtime();
  next_report_sent.tv_sec += t;
  pmembers = get_number_of_members();
}


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//* Calculate various sdes items
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
string RTPI::RtpiParticipant::get_note() {
  string str = "";
  switch (type) {
  case RTPI_MASTER:
    str = prri->get_session_note();
    if (str.empty())
      str = note;
    break;
  case RTPI_LOCAL:
    str = (*prh->master_pos)->get_name();
    break;
  case RTPI_REMOTE:
    str = note;
    break;
  };
  return str;
};


string RTPI::RtpiParticipant::get_tool() {
  string str;
  switch (type) {
  case RTPI_MASTER:
    str = prri->get_session_tool();
    if (str.empty())
      str = tool;
    break;
  case RTPI_LOCAL: {
    // find master tool in tool
    string master_tool = (*prh->master_pos)->get_tool();
    int tmppos = master_tool.find(" ");
    if (tmppos != string::npos)
      master_tool = master_tool.substr(0, tmppos);
    if (tool.find(master_tool) != string::npos) {
      // already recorded...
      int toolpos = tool.find("(was: ");
      int toolpos_end = ((toolpos != string::npos) ? tool.find(")", toolpos) : string::npos);
      if ((toolpos != string::npos) && (toolpos_end != string::npos)){
	str = (*prh->master_pos)->get_tool() + " " +
	  tool.substr(toolpos, toolpos_end - toolpos) + ") (re-recorded)";
      } else {
	str = (*prh->master_pos)->get_tool() + " (re-recorded)";
      };
    } else {
      str = (*prh->master_pos)->get_tool();
      if (!tool.empty()) {
	str += " (was: " + tool + ")";
      };
    }
    break;
  }
  case RTPI_REMOTE:
    str = tool;
    break;
  }

  return str;
}


string RTPI::RtpiParticipant::get_loc() {
  string str;
  switch (type) {
  case RTPI_MASTER:
    str = prri->get_session_loc();
    if (str.empty())
      str = loc;
    break;
  case RTPI_LOCAL: {
    int tmppos = loc.find("(was: ");
    if (tmppos != string::npos) {
      // already recorded...
      str=(*prh->master_pos)->get_loc() + " " + loc.substr(tmppos);
    } else {
      if ((loc != (*prh->master_pos)->get_loc()) && (!loc.empty()))
	str = (*prh->master_pos)->get_loc() + " (was: " + loc + ")";
      else
	str = (*prh->master_pos)->get_loc();
    }
    break;
  }
  case RTPI_REMOTE:
    str=loc;
    break;
  }
  return str;
}

string RTPI::RtpiParticipant::get_email() {
  string str;
  switch (type) {
  case RTPI_MASTER:
    str = prri->get_session_email();
    if (str.empty())
      str = email;
    break;
  case RTPI_LOCAL: {
    int tmppos = email.find("(was: ");
    if (tmppos != string::npos) {
      // already recorded...
      str=(*prh->master_pos)->get_email() + " " + email.substr(tmppos);
    } else {
      if ((email != (*prh->master_pos)->get_email()) && (!email.empty()))
	str = (*prh->master_pos)->get_email() + " (was: " + email + ")";
      else
	str =(*prh->master_pos)->get_email();
    }
    break;
  }
  case RTPI_REMOTE:
    str=email;
    break;
  }
  return str;
}


string RTPI::RtpiParticipant::get_phone() {
  string str;
  switch (type) {
  case RTPI_MASTER:
    str = prri->get_session_phone();
    if (str.empty())
      str = phone;
    break;
  case RTPI_LOCAL: {
    int tmppos = phone.find("(was: ");
    if (tmppos != string::npos) {
      // already recorded...
      str=(*prh->master_pos)->get_phone() + " " + phone.substr(tmppos);
    } else {
      if ((phone != (*prh->master_pos)->get_phone()) && (!phone.empty()))
	str = (*prh->master_pos)->get_phone() + " (was: " + phone + ")";
      else
	str = (*prh->master_pos)->get_phone();
    }
    break;
  }
  case RTPI_REMOTE:
    str=phone;
    break;
  }
  return str;
}

string RTPI::RtpiParticipant::get_name() {
  string str;
  switch (type) {
  case RTPI_MASTER:
    str = prri->get_session_name();
    if (str.empty())
      str = name;
    break;
  case RTPI_LOCAL:
    if (name.find(" (recorded)") != string::npos) {
      // already recorded...
      str = name;
    } else {
      if (!name.empty()) {
	str = name + " (recorded)";
      } else {
	str = cname + " (recorded)";
      };
    }
    break;
  case RTPI_REMOTE:
    str = name;
  };
  return str;
};


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//* Helpers to implement bye backoff
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
int RTPI::RtpiParticipant::get_average_rtcpi_size() {
  if (bye_backoff_bypass) {
    return RTCPI_BYE_PACKET_SIZE;
  } else {
    return prh->get_avg_rtcpi_size();
  }
}

int RTPI::RtpiParticipant::get_number_of_members() {
  if (bye_backoff_bypass) {
    return prh->bye_counter - member_bypass;
  } else {
    return prh->get_members();
  }
}

void RTPI::RtpiParticipant::start_bye_backoff() {
  bye_backoff_bypass = true;
  member_bypass = prh->bye_counter - 1;
}


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//*
//*
//*
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void RTPI::RtpiParticipant::set_state(int s) {
  if (s == RTPI_REMOVE) {
    first_data = true;
    // tell the application about the participant to be removed
    if ((state != RTPI_BYED) && (prh->get_feedback_interface() != NULL) &&
	(type != RTPI_MASTER)) {
      prh->get_feedback_interface()->participant_removed(pid);
    }
  }
  if (s == RTPI_BYED) {
    // mark as byed and delete after several check intervals
    reported_bye = 0;
    // tell the application about the participant to be removed
    if ((prh->get_feedback_interface() != NULL) && (type != RTPI_MASTER)) {
      prh->get_feedback_interface()->participant_removed(pid);
    }
  }
  state=s;
}










