// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//
// Copyright (c) 2000 University of Mannheim. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for educational, research and non-profit purposes,
// without fee, and without a written agreement is hereby granted,
// provided that the above copyright notice, this paragraph and the
// following four paragraphs appear in all copies.
//
// Permission to incorporate this software into commercial products may
// be obtained from the following address:
// University of Mannheim
// Lehrstuhl fuer Praktische Informatik IV
// L15, 16
// 68161 Mannheim, Germany
//
// IN NO EVENT SHALL THE UNIVERSITY OF MANNHEIM BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
// ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF MANNHEIM HAS BEEN ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//
// THE UNIVERSITY OF MANNHEIM SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER
// IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF MANNHEIM HAS NO OBLIGATIONS
// TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The University of Mannheim will appreciate receiving any enhancements and
// modifications of this software made by others, and reserves the right to
// include them in future releases. The University also recommends that this
// software not be redistributed by third parties, but be obtained by all
// directly from the University's server, so that all users of the software
// may receive bug notifications and other relevant announcements.
//
// Authors:
//   Volker Hilt, Juergen Vogel, Martin Mauve
//   University of Mannheim
//   {hilt,vogel,mauve}@informatik.uni-mannheim.de
//
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#ifndef _RtpiParticipant_h_
#define _RtpiParticipant_h_

#ifdef WIN32
#  pragma warning (disable: 4786)
#endif


#ifdef WIN32
#include <winsock.h>
#else
#include <netinet/in.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif

#include <vector>
#include <string>

#include <util/types.h>
#include <util/random.h>
#include <util/tools.h>
#include <util/timerinterface.h>
#include <rtpi/rtpiPacket.h>
#include <rtpi/rtcpiPacket.h>
#include <rtpi/rtpiSeqNumber.h>
#include <rtpi/rtpiInterfaces.h>

/*
 * Minimum time between RTCPI packets from this site (in seconds).
 * This time prevents the reports from `clumping' when sessions are
 * small and the law of large numbers isn't helping to smooth out
 * the traffic.  it also keeps the report interval from becoming
 * ridiculously small during transient outages like a network
 * partition.
 * Since this value may change from profile to profile, the value
 * should be retrieved via a pt-info field from the playload.
 */
#define RTCPI_MIN_RPT_TIME (5.)

// states of a received RTPI participants
#define RTPI_INVALID	0  // site has timed out
#define RTPI_VALID	2  // site is ok
#define RTPI_BYED	3  // site has sent a bye
#define RTPI_REMOVE	4  // site will be deleted soon

// types of participants
#define RTPI_MASTER	0  // local participant itself
#define RTPI_LOCAL	1  // "virtual" local participants (e.g. recorded)
#define RTPI_REMOTE	2  // other participants (e.g. on remote hosts)

#define RTPI_SEQ_MOD (1<<16)            /* 2^16 needed for cycle counter */
/*
 * Parameters controling the RTCP report rate timer.
 */
/*
 * Fraction of the rtcp bandwidth to be shared among participants.
 */
#define RTCPI_SESSION_BW_FRACTION (0.05)

// how many rtcpi report intervals may pass until we remove a site if
// we don't receive data/reports from this site
#define RTPI_REMOVE_INTERVAL 5

/*
 * Bye rtcpi packet size used to reset the
 * average rtcp packet size estimate for bye backoff.
 * 56 roughly assumes a compound packet consisting of
 * one SDES item (CNAME) and one BYE item with a short
 * reason string.
 */
#define RTCPI_BYE_PACKET_SIZE 56


namespace RTPI {

  class RtpiHandler;
  class RtpiSubComp;
  class RtpiParticipant;

  class RtpiParticipant : public TimerInterface {
    friend class RtpiHandler;
    friend std::ostream& operator<<(std::ostream& os, RTPI::RtpiParticipant &p);

    class RtcpiReportParameters {
    public:
      RtcpiReportParameters(bool s_b, string b_m) :
	send_bye(s_b), bye_msg(b_m), sdes(0) { };
      RtcpiReportParameters(int s) :
	send_bye(false), bye_msg(""), sdes(s) { };
      // parameters needed for bye reports
      bool send_bye;
      string bye_msg;
      // parameters needed for sdes reports
      int sdes;
    };

  public:
    RtpiParticipant(RtcpiReportInterface*, RtpiHandler*, u_int32, int);
    // no copy constructor defined on purpose!!
    virtual ~RtpiParticipant();

    // get sdes infos
    string    get_cname()                      { return cname; };
    string    get_priv()                       { return ""; };     // ignore priv ;)
    string    get_name();
    string    get_email();
    string    get_phone();
    string    get_loc();
    string    get_tool();
    string    get_note();
    // set sdes infos
    // (RTCPI report interface function will override the data set here)
    inline void      set_cname(string s)         { cname=s; };
    inline void      set_name(string s)          { name=s; };
    inline void      set_email(string s)         { email=s; };
    inline void      set_phone(string s)         { phone=s; };
    inline void      set_loc(string s)	       { loc=s; };
    inline void      set_tool(string s)          { tool=s; };
    inline void      set_note(string s)	       { note=s; };
    inline void      set_priv(string s)	       { priv=s; };

    // get participant infos
    inline u_int32   get_pid()                  { return pid; };
    inline u_int32   get_addr()                 { return addr; };
    inline int       get_state()                { return state; };
    inline int       get_type()                 { return type; };
    inline u_int32   get_alt_pid()              { return alt_pid; };

    // the following function will typically not be needed by a typical application
    // however they might be useful to some

    // get transmission and reception times
    inline timeval   get_last_data_sent()       { return last_data_sent; };
    inline timeval   get_last_data_received()   { return last_data_received; };
    inline timeval   get_last_report_sent()     { return last_report_sent; };
    inline timeval   get_last_report_read()     { return last_report_read; };
    inline timeval   get_last_report_received() { return last_report_received; };
    inline timeval   get_next_report_sent()     { return next_report_sent; };

    // reset sync timestamps
    void reset_transmission();

    // send a report (usually triggered by the report timer)
    void send_report(int = 0, bool = false, string = "");

    // set or remove a flag for this participant
    // can be used freely by an application
    inline bool flag_participant() {
      bool tmp = flag;
      flag = true;
      return tmp;
    };
    inline bool deflag_participant() {
      bool tmp = flag;
      flag = false;
      return tmp;
    };

    // deny (=true) or allow (=false) reporting of sub-components
    void set_deny_subcomp_reporting(bool d) { deny_subcomp_reports = d; }

  private:
    // functions needed to implement bye backoff
    inline int get_average_rtcpi_size();
    inline int get_number_of_members();
    inline void start_bye_backoff();
    inline void stop_bye_backoff() { bye_backoff_bypass = false; }

    // get unmodified sdes infos
    inline string    get_plain_priv()      { return priv; };
    inline string    get_plain_cname()     { return cname; };
    inline string    get_plain_name()      { return name; };
    inline string    get_plain_email()     { return email; };
    inline string    get_plain_phone()     { return phone; };
    inline string    get_plain_loc()       { return loc; };
    inline string    get_plain_tool()      { return tool; };
    inline string    get_plain_note()      { return note; };

    // set participant infos
    inline void      set_addr(u_int32 a)         { addr=a; };
    void             set_state(int s);
    inline void      set_type(int t)             { type=t; };
    inline void      set_alt_pid(u_int32 p)      { alt_pid=p; };

    // handle rtpi adus
    void   	   check_incoming_rtpi(RefCount<RtpiAdu>);
    void   	   check_incoming_rtcpi(RtcpiPacketVector::iterator);
    void  	   check_outgoing_rtpi(RefCount<RtpiAdu>);
    void  	   check_outgoing_rtcpi(RtcpiPacketVector::iterator);

    // manage rtcpi report timer
    void init_report_timer(int = 0);
    void reverseTimerReconsideration();

    // send a bye packet
    void send_bye_and_remove(string);
    void send_bye_no_statechange(string);

    // implements TimerInterface
    void timerHandler(string);

    // updates participants state (if necessary)
    // called regularly by rtpihandler
    void update_state();

    // schedule next rtcpi transmission timer
    void schedule_rtcpi_timer(int);


    // the pointer to the RtcpiReportInterface to actually send a report
    RtcpiReportInterface* prri;
    // the pointer to the rtpihandler to get infos about other participants
    RtpiHandler* prh;

    // the timer for sending reports
    P2Timer	rtcpi_timer;

    // marked with true each time we start playing/recording,
    // must be set to false afterward we sent/recorded the first
    // packet sucessfully
    bool	first_data;
    // marked with 1 if we have not yet sent a report.
    bool        initial_report_to_send;

    // this sites current state (see above)
    int state;
    // this sites type (see above)
    int type;

    // the current payload type of the site
    char	payload_type;

    // this sites current pid
    u_int32	pid;
    // local sites recorded pid to keep track of packets...
    u_int32	alt_pid;
    // the transport address of the site as 32bit word
    u_int32 addr;

    // SDES items
    // Needed only for local and remote participants. The master obtains this
    // infos via RTCPI report interface from the application.
    string	cname;
    string	name;
    string	email;
    string	phone;
    string	loc;
    string	tool;
    string	note;
    string	priv;

    /* for validation check and statistics */
    bool	wait_for_rtcpi;

    // flag this participant
    // e.g. may indicate that packets were recorded for this participant
    bool	flag;

    // data needed to compute the rtcpi interval
    int           pmembers;   // members when the rtcpi report interval was computed

    // local unix time from last data packet sent
    timeval	last_data_sent;
    // local unix time from last data packet received
    timeval	last_data_received;
    // local unix time from last Report sent
    timeval	last_report_sent;
    timeval     last_report_read;
    // local unix time at which next report will be sent
    timeval     next_report_sent;
    // local unix time from last Report received
    timeval	last_report_received;

    // how many report intervals have passed since we have sent bye
    int         reported_bye;

    // use size of bye packet instead of real rtcpi size average
    // during bye backoff
    bool        bye_backoff_bypass;

    // deny the reporting of sub components for this participant
    bool        deny_subcomp_reports;

    // use own member counter during bye backoff
    unsigned int member_bypass;

    // needed by next_sdes_item() (see below)
    int		which_sdes_item;

    // calculate which SDES ITEM we will send in the next SDES packet
    int 	next_sdes_item();

    // build a sdes
    void	build_sdes(RtcpiCompoundPacketVector&, std::list<RtcpiReportParameters>&, bool = true);
    void	build_sdes(RtcpiCompoundPacketVector& rtcpi, bool automatic_sdes = true) {
			std::list<RtcpiReportParameters> dummy;
      build_sdes(rtcpi, dummy, automatic_sdes);
    };

    // build a sdes item
    RtcpiSourceDescriptionItem*   build_sdes_item(int);

    // build a bye packet
    void	build_bye(RtcpiCompoundPacketVector&, string);

    // build a subcomponent report packet
    void	build_subrep(RtcpiCompoundPacketVector&);

    // calculate the rtcpi interval
    int		rtcpi_interval();
    double      rtcpi_deterministic_interval(int);

    // list stores parameters for next report
		std::list<RtcpiReportParameters> ReportsToSend;
  };

#ifdef WIN32
  using std::list<RtpiParticipant::RtcpiReportParameters>;
#endif

  typedef RtpiParticipant* P2RtpiParticipant;
};

#endif






